/*
 * Copyright © 2020, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.phone;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableCellRenderer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.nexuswob.gui.TableModel3;
import org.nexuswob.util.Cell;
import org.nexuswob.util.Util;
import net.javacomm.client.resource.Resource;
import net.javacomm.multilingual.Babelfish;
import net.javacomm.multilingual.MultilingualLabel;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;
import net.javacomm.protocol.KonferenzraumUser;



/**
 * Ein grafischer Überblick über alle Anwender, die in einem Konferenzraum sind.
 * 
 * 
 * @author llange
 *
 */
public class JOnlineOfflinePanel extends JPanel implements Babelfish {

  /***************************************************************************/
  /*                                                                         */
  /* Innerclass ColorRenderer */
  /*                                                                         */
  /***************************************************************************/

  class ColorRenderer extends JLabel implements TableCellRenderer {

    private static final long serialVersionUID = 1483352116438614471L;

    public ColorRenderer() {
      setOpaque(true);
    }



    @Override
    public Component getTableCellRendererComponent(JTable table, Object str, boolean isSelected,
        boolean hasFocus, int row, int column) {

      // Wenn der chatuser anwesend ist, dann zeichne Mikro
      KonferenzraumUser user = konferenzraumUser.get(row);

      setBackground(new Color(user.getBackgroundColor()));
      setForeground(new Color(user.getForegroundColor()));
      setText((String) str);

      if (user.isAnwesend()) {
        if (user.isMute()) {
          try {
            BufferedImage image = JOnlineOfflinePanel.modifyRGBA(Resource.MUTE_26x26, getForeground());
            setIcon(new ImageIcon(image));
            setHorizontalAlignment(SwingConstants.LEFT);
            setIconTextGap(12);
          }
          catch (IOException e) {
            log.info(e.getMessage(), e);
          }
        }
        else {
          try {
            BufferedImage image = JOnlineOfflinePanel.modifyRGBA(Resource.UNMUTE_26x26, getForeground());
            setIcon(new ImageIcon(image));
            setHorizontalAlignment(SwingConstants.LEFT);
            setIconTextGap(12);
          }
          catch (IOException e) {
            log.info(e.getMessage(), e);
          }
        }
      }
      return this;
    }

  }

  static class CountUserPanel extends JPanel {

    private static final long serialVersionUID = 8123825495408532218L;
    private JLabel labelNumber = new JLabel();
    private GridBagLayout gridbag = new GridBagLayout();
    private GridBagConstraints con = new GridBagConstraints();
    private MultilingualLabel labelAnwesend = new MultilingualLabel(KEY.LABEL_ANWESEND);

    public CountUserPanel() {
      setLayout(gridbag);
      setBackground(Resource.JQUERY_ORANGE);

      con.insets = new Insets(2, 2, 0, 2);
      con.gridwidth = GridBagConstraints.REMAINDER;
      gridbag.setConstraints(labelAnwesend, con);
      add(labelAnwesend);
      labelAnwesend.setForeground(Resource.JQUERY_TEXTBRAUN);

      con.insets = new Insets(1, 2, 2, 2);
      gridbag.setConstraints(labelNumber, con);
      labelNumber.setForeground(Resource.JQUERY_TEXTBRAUN);
      add(labelNumber);
      setBorder(null);
    }



    public void setCount(int c) {
      labelNumber.setText(String.valueOf(c));
    }



    public void setTextBackground(Color color) {
      super.setBackground(color);
    }



    public void setTextForeground(Color color) {
      labelAnwesend.setForeground(color);
      labelNumber.setForeground(color);
    }

  }

  private class HeaderRenderer extends JLabel implements TableCellRenderer {
    private static final long serialVersionUID = 1483352116438614471L;

    public HeaderRenderer() {
      setOpaque(true);
    }



    @Override
    public Component getTableCellRendererComponent(JTable table, Object str, boolean isSelected,
        boolean hasFocus, int row, int column) {

      if (column == 0) {
        setBorder(new EmptyBorder(0, 0, 0, 0));
      }

      setBackground(Resource.JQUERY_ORANGE);

      setText((String) str);
      return this;
    }
  }

  private static final long serialVersionUID = -1258669005473361153L;
  private final static Logger log = LogManager.getLogger(JOnlineOfflinePanel.class);
  // private final static String names [] = {"", /*"First Name",
  // "Last Name", "UID"*/};
  private TableModel3 model = new TableModel3();
  private JTable table = new JTable();
  private JScrollPane scrollTable = new JScrollPane();
  private PropertyChangeSupport changes = new PropertyChangeSupport(this);
  private JTableHeader header;
  private ColorRenderer cRenderer = new ColorRenderer();
  private CountUserPanel countUserPanel = new CountUserPanel();
  private ArrayList<KonferenzraumUser> konferenzraumUser = new ArrayList<>();
  private HeaderRenderer headerrenderer = new HeaderRenderer();

  public JOnlineOfflinePanel() {
    super(true);
    model.setHeader("");
    setLayout(new BorderLayout());
    table.setModel(model);
    table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    // table.setAutoResizeMode (JTable.AUTO_RESIZE_OFF);
    header = table.getTableHeader();
    header.setReorderingAllowed(false);
    header.setDefaultRenderer(headerrenderer);
    headerrenderer.setHorizontalAlignment(SwingConstants.CENTER);

    table.setDefaultRenderer(String.class, cRenderer);
    cRenderer.setHorizontalAlignment(SwingConstants.CENTER);
    table.setRowHeight(28);

    scrollTable.setViewportView(table);
    scrollTable.getHorizontalScrollBar().setUnitIncrement(12);
    scrollTable.getVerticalScrollBar().setUnitIncrement(14);

    countUserPanel.setBorder(null);
    add(BorderLayout.NORTH, countUserPanel);
    add(BorderLayout.CENTER, scrollTable);

    scrollTable.setMinimumSize(new Dimension(84, getSize().height));
    scrollTable.setPreferredSize(new Dimension(84, getSize().height));
    scrollTable.setMaximumSize(new Dimension(84, getSize().height));

    scrollTable.setBorder(null);
    scrollTable.getViewport().setBackground(Resource.JQUERY_TEXTBRAUN);

  }



  /**
   * Ändere die Vordergrundfarbe und erhalte die Transparenz. Die übergebene
   * Resource ist ein monochromes PNG, JPEG oder GIF Bild. Der Resourcenstring ist
   * ein Pfad auf das Bild. Beachte, dass ein Resourcenstring mit einem Slash
   * {@code /} beginnt.
   * 
   * @param resource
   *                 dieser Resourcenstring
   * @param color
   *                 diese neue Vordergrundfarbe
   * @return aus einer Resource neu erstelltes BufferedImage
   * @throws IOException
   *                     die Resource konnte nicht verarbeitet und erstellt werden
   */
  public static BufferedImage modifyRGBA(String resource, Color color) throws IOException {

    BufferedImage bufferedimage;
    try(InputStream is = JOnlineOfflinePanel.class.getResourceAsStream(resource)) {
      bufferedimage = ImageIO.read(is);
      Util.modifyRGBA(bufferedimage, color);
    }
    return bufferedimage;
  }



  public void addChatInfoListener(PropertyChangeListener l) {
    changes.addPropertyChangeListener(l);
  }



  /**
   * Ein Konferenzteilnehmer wird mit seinem Nickname angezeigt.
   * 
   * 
   * @param user
   *             dieser Anwender
   */
  public void addUser(KonferenzraumUser user) {
    Util.Record record = new Util.Record();
    Cell<String> cellNickname = new Cell<>();
    cellNickname.setValue(user.getNickname());
    record.add(cellNickname);
    model.addRow(record);
    konferenzraumUser.add(user);
    countUserPanel.setCount(konferenzraumUser.size());
  }



  /**
   * Die Teilnehmerliste wird gelöscht.
   */
  public void clear() {
    model.removeAll();
    konferenzraumUser.clear();
    countUserPanel.setCount(konferenzraumUser.size());
  }



  /**
   * Gib alle Konferenzteilnehmer zurück.
   * 
   * @return diese Teilnehmer
   */
  public List<KonferenzraumUser> getKonferenzraumTeilnehmer() {
    return konferenzraumUser;
  }



  @Override
  public Dimension getPreferredSize() {
    return new Dimension(150, super.getHeight());
  }



  public void removeChatInfoListener(PropertyChangeListener l) {
    changes.removePropertyChangeListener(l);
  }



  @Override
  public void setLanguage(ISO639 code) {
    setLanguage(KEY.LABEL_ANWESEND, code);
  }



  public void setLanguage(KEY key, ISO639 code) {
    countUserPanel.labelAnwesend.setLanguage(key, code);
  }



  public void setTopBackground(Color color) {
    countUserPanel.setTextBackground(color);
  }



  public void setTopForeground(Color color) {
    countUserPanel.setTextForeground(color);
  }



  /**
   * Die Hintergrundfarbe wird gewechselt.
   * 
   * @param userid
   *               dieser Anwender
   * @param color
   *               diese neue Hintergrundfarbe
   */
  public void updateBackground(String userid, Color color) {
    // TODO
    // wird relevant, wenn sich der Nickname ändert
    for (int row = 0; row < model.getRowCount(); row++) {
      String value = model.getValueAt(row, 0).toString();
      if (value.equals(userid)) {
//      chatuser[row].setForegroundColor(color.getRGB());
        repaint();
        break;
      }
    }

  }



  /**
   * Der User hat seine Vordergrundfarbe gewechselt.
   * 
   * @param userid
   *               dieser Anwender
   * @param color
   *               diese neue Hintergrundfarbe
   */
  public void updateForeground(String userid, Color color) {
    // TODO
    // wird relevant, wenn sich der Nickname ändert
    for (int row = 0; row < model.getRowCount(); row++) {
      String value = model.getValueAt(row, 0).toString();
      if (value.equals(userid)) {
//        chatuser[row].setForegroundColor(color.getRGB());
        repaint();
        break;
      }
    }

  }

}
