/*
 * Copyright © 2025, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.phone;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.client.websocket.WebsocketClient;
import net.javacomm.protocol.CONFERENCEAUDIO;
import net.javacomm.protocol.Command;
import net.javacomm.protocol.HEADER;
import net.javacomm.protocol.Protocol;



/**
 * Der <b>Mikrofon Conference Consumer</b> versendet Sprachaufzeichnungen aus
 * einem Konferenzraum. Die Übertragung geschieht über einen Websocket an den
 * Server.
 *
 */
public class MicConferenceConsumer extends Thread {

  private final static Logger log = LogManager.getLogger(MicConferenceConsumer.class);
  private volatile String[] receiverSessions;
  private AtomicBoolean isMute = new AtomicBoolean(false);
  private ArrayBlockingQueue<byte[]> queue;
  private WebsocketClient client;
  private AtomicBoolean terminate = new AtomicBoolean(false);
  private ReentrantLock lock;
  private Condition condition;
  private String userid;

  /**
   * Der Mikrofon Conference Consumer wird eingerichtet.
   * 
   * @param queue
   *               diese Queue muss muss mit einem Producer verbunden werden
   * @param client
   *               die Mikrofonaufnahme wird über diesen Client an den Server
   *               gesendet
   */
  public MicConferenceConsumer(ArrayBlockingQueue<byte[]> queue, WebsocketClient client) {
    super("MicConferenceConsumer");
    this.queue = queue;
    this.client = client;
    isMute.set(false);
    receiverSessions = new String[0];
    lock = new ReentrantLock();
    condition = lock.newCondition();
  }



  /**
   * Bevor ein Datenpaket gesendet werden darf, muss eine Erlaubnis vorliegen.
   * Diese Methode muss von einem externen Thread in einer Art Endlos-Schleife
   * aufgerufen werden. Roundrobin ist so ein Szenario. Der Server erteilt dem
   * Consumer die Erlaubnis, ein Paket senden zu dürfen. Nachdem das Paket
   * gesendet wurde, wird der aufrufende Thread fortgesetzt. Der Consumer legt
   * sich schlafen. Nach einer unbestimmten Zeit wird der Server den Consumer
   * wieder aufwecken.
   * 
   * @param latch
   *              latch hat den Wert 1
   */
  public void enableHasPermission() {}



  /**
   * Das Mikrofon wurde auf stummschalten gestellt.
   * 
   */
  public void mute() {
    isMute.set(true);
  }



  @Override
  public void run() {
    while (!terminate.get()) {
      if (isMute.get()) {
        // schlafen legen
        lock.lock();
        try {
          condition.await();
        }
        catch (InterruptedException e) {
          log.info(e.getMessage(), e);
        }
        finally {
          lock.unlock();
        }
      }

      try {
        byte[] audiobytes = queue.take();
        CONFERENCEAUDIO conferenceaudio = new CONFERENCEAUDIO();
        conferenceaudio.setCommand(Command.CONFERENCEAUDIO);
        conferenceaudio.setHeader(HEADER.REQUEST);
        conferenceaudio.setDataset(Protocol.DATASET);
        conferenceaudio.setContent(audiobytes);
        conferenceaudio.setUserid(userid); // optional

        conferenceaudio.setReceiverSessions(receiverSessions);
        client.sendMessage(conferenceaudio);
      }
      catch (InterruptedException e) {
        terminate();
      }

    }
    log.info(MicConferenceConsumer.class.getSimpleName() + " hat terminiert.");
  }



  public void terminate() {
    terminate.set(true);
    unmute();
  }



  /**
   * Der Audiostream/Telefonat wird an diese Anwender gesendet. Jede Session wird
   * intern auf einen Anwender abgebildet.
   * 
   * @param sessions
   *                 alle Empfängersessions
   */
  public void setReceiverSessions(String[] sessions) {
    receiverSessions = sessions;
  }



  /**
   * Das Mikrofon wurde auf Aufnahme gestellt.
   */
  public void unmute() {
    isMute.set(false);
    lock.lock();
    condition.signal();
    lock.unlock();
  }



  /**
   * Die Sprachnachricht stammt von diesem Absender.
   * 
   * @param uid
   *            dieser Anwender
   */
  public void setUserid(String uid) {
    userid = uid;
  }

}
