/*
 * Copyright © 2025, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.phone;

import static net.javacomm.protocol.HEADER.REQUEST;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.client.websocket.WebsocketClient;
import net.javacomm.protocol.Command;
import net.javacomm.protocol.Protocol;
import net.javacomm.protocol.TCPAUDIO;



/**
 * Der <b>Mikrofon Consumer</b> versendet Sprachaufzeichnungen über einen
 * Websocket an einen Server.
 */
public class MicConsumer extends Thread {

  private final static Logger log = LogManager.getLogger(MicConsumer.class);
  private ArrayBlockingQueue<byte[]> queue;
  private String origin;
  private String target;
  private WebsocketClient client;
  private AtomicBoolean terminate = new AtomicBoolean(false);

  /**
   * Der Mikrofon Consumer wird eingerichtet.
   * 
   * @param queue
   *               diese Queue muss muss mit einem Producer verbunden werden
   * @param client
   *               die Mikrofonaufnahme wird über diesen Client an den Server
   *               gesendet
   */
  public MicConsumer(ArrayBlockingQueue<byte[]> queue, WebsocketClient client) {
    super("MicConsumer");
    this.queue = queue;
    this.client = client;
  }



  /**
   * Der Mikrofon Consumer wird eingerichtet.
   * 
   * @param queue
   *               diese Queue muss muss mit einem Producer verbunden werden
   * @param client
   *               die Mikrofonaufnahme wird über diesen Client an den Server
   *               gesendet
   * @param origin
   *               der Absender ist eine sessionid
   * @param target
   *               der Empfänger ist eine sessionid
   */
  public MicConsumer(ArrayBlockingQueue<byte[]> queue, WebsocketClient client, String origin, String target) {
    this(queue, client);
    this.origin = origin;
    this.target = target;
  }



  @Override
  public void run() {
    while (!terminate.get()) {

      try {
        byte[] audiobytes = queue.take();
        TCPAUDIO audio = new TCPAUDIO();
        audio.setCommand(Command.TCPAUDIO);
        audio.setHeader(REQUEST);
        audio.setDataset(Protocol.DATASET);
        audio.setOrigin(origin);
        audio.setTarget(target);
        audio.setContent(audiobytes);
        client.sendMessage(audio);
      }
      catch (InterruptedException e) {
        break;
      }

    }
    log.info(MicConsumer.class.getName() + " hat terminiert.");

  }



  public void terminate() {
    terminate.set(true);
  }



  /**
   * Der Absender der Nachricht.
   * 
   * @param absender
   *                 der Absender ist eine sessionid
   */
  public void setOrigin(String absender) {
    origin = absender;
  }



  /**
   * Der Empfänger der Nachricht.
   * 
   * @param receiver
   *                 der Empfänger ist eine sessionid
   */
  public void setTarget(String receiver) {
    target = receiver;
  }

}
