/*
 * Copyright © 2025, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.phone;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer.Info;
import javax.sound.sampled.TargetDataLine;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.multilingual.Babelfish;
import net.javacomm.multilingual.MultilingualString;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;
import io.nayuki.flac.app.EncodeWavToFlac;



/**
 * Diese Klasse ist ein Producer für Mikrofonaufnahmen.
 */
public class MicProducer extends Thread implements Babelfish {

  private final static Logger log = LogManager.getLogger(MicProducer.class);

  private AtomicBoolean terminate = new AtomicBoolean(false);
  private ArrayBlockingQueue<byte[]> queue;
  private Telefonformat telefonformat;
  private TargetDataLine targetDataline;
  private byte[] rawAudiobuffer;
  private MultilingualString mikrofonnameIstUnbekannt = new MultilingualString(
      KEY.STRING_DER_MIKROFONNAME_IST_UNBEKANNT
  );
  private LineListener lineListener = new LineListener() {

    @Override
    public void update(LineEvent event) {
      // wird vom JavaSoundEventDispatcher Thread ausgeführt
      if (event.getType().equals(LineEvent.Type.OPEN)) {}
      else if (event.getType().equals(LineEvent.Type.START)) {
        targetDataline.flush();
      }
      else if (event.getType().equals(LineEvent.Type.STOP)) {}
      else if (event.getType().equals(LineEvent.Type.CLOSE)) {
        targetDataline.removeLineListener(lineListener);
      }
    }
  };

  /**
   * Der Producer wird eingerichtet. Die Standardpuffergröße ist 2304 Bytes lang.
   * 
   * @param queue
   *              diese Queue muss mit einem Consumer gekoppelt werden
   */
  public MicProducer(ArrayBlockingQueue<byte[]> queue) {
    this(queue, 2304);// 3*564+576];
  }



  /**
   * Der Producer wird eingereichtet. Die Puffergröße sollte ein Vielfaches von
   * 576 sein. Ein zu großer Puffer erhöht die Latenzzeit.
   * 
   * @param queue
   *                        diese Queue muss mit einem Consumer gekoppelt werden
   * @param audioBufferSize
   *                        die Puffergröße ist anhängig vom Telefonformat
   */
  public MicProducer(ArrayBlockingQueue<byte[]> queue, int audioBufferSize) {
    super("MicProducer");
    this.queue = queue;
    rawAudiobuffer = new byte[audioBufferSize]; // 3*564+576];
  }



  /**
   * Öffnet eine Mikrofonleitung. Die Mikrofonleitung darf nicht belegt sein.
   * 
   * @param audioformat
   *                    dieses Audioformat
   * @throws MikrofonException
   *                           die Mikrofonleitung wird benutzt; schließe die
   *                           Mikrofonleitung
   */
  private void open(AudioFormat audioformat) throws MikrofonException {
    try {
      if (targetDataline == null) return;
      targetDataline.open(audioformat);
    }
    catch (LineUnavailableException e) {
      throw new MikrofonException("Diese Mikrofonleitung wird bereits benutzt.");
    }
  }



  @Override
  public void run() {
    int read;
    byte[] waveByteArray;
    byte[] flacArray;

    if (targetDataline == null) {
      log.warn(MicProducer.class.getName() + " - targetDataline is null");
      log.warn("Audioübertragungen sind nicht möglich.");
      log.warn("setTelefonformat(String, Telefonformat) muss vorher aufgerufen werden.");
      return;
    }
    targetDataline.start();
    while (!terminate.get()) {
      try {
        read = targetDataline.read(rawAudiobuffer, 0, rawAudiobuffer.length);
        waveByteArray = toWave(ByteBuffer.wrap(rawAudiobuffer, 0, read), telefonformat);
        flacArray = EncodeWavToFlac.entry(waveByteArray);
        queue.offer(flacArray);
      }
      catch (AudioFormatException | IOException e) {
        terminate();
      }
    }
    targetDataline.close();
    log.info(MicProducer.class.getSimpleName() + " hat terminiert.");
  }



  public void terminate() {
    terminate.set(true);
  }



  @Override
  public void setLanguage(ISO639 code) {
    mikrofonnameIstUnbekannt.setLanguage(code);
  }



  /**
   * Diese Methode muss aufgerufen werden, bevor der Producer gestartet wird.
   * 
   * Für ein Mikrofon wird das {@link net.javacomm.client.phone.Telefonformat}
   * festgelegt. Der übergebene Mikrofonname muss gültig sein. Der Mikrofonname
   * konnte über
   * {@link net.javacomm.client.phone.TelefonUtil#microMixer(Telefonformat) }
   * ermittelt werden.
   * 
   * @param mikroname
   *                  dieser Mikrofonname
   * @param value
   *                  dieses Audioformat
   * 
   * @throws MikrofonException
   *                            für das ausgewählte Mikrofon ist die zugehörige
   *                            Leitung belegt
   * 
   * @throws MicronameException
   *                            dieser Mikrofonname ist nicht gültig
   */
  public void setTelefonformat(String mikroname, Telefonformat value)
      throws MikrofonException, MicronameException {
    telefonformat = value;

    List<Info> mixer = TelefonUtil.microMixer(telefonformat);
    for (Info tmp : mixer) {
      if (tmp.getName().equals(mikroname)) {
        try {
          targetDataline = TelefonUtil.getTargetDataLine(tmp, telefonformat);
        }
        catch (LineUnavailableException e) {
          log.error(e.getMessage(), e.getCause());
          throw new MikrofonException("Für dieses Mikrofon ist die Leitung belegt. - " + mikroname);
        }
        targetDataline.addLineListener(lineListener);
        open(TelefonUtil.getAudioFormat(telefonformat));
        return;
      }
    }
    throw new MicronameException("<b>'" + mikroname + "'</b> - " + mikrofonnameIstUnbekannt.toString());

  }



  /**
   * 
   * Die gerade empfangenen Audiodaten werden übergeben. Die Audiodaten stammen
   * von einem Mikrofon.
   * 
   * 
   * @param micro
   *               aufgezeichnete Mikrofonbytes
   * @param format
   *               das Telefonformat der Mikrofonbytes
   * 
   * @return WAVE-Audiodaten als ByteArray
   * 
   * @throws AudioFormatException
   *                              das Audioformat wird nicht unterstützt
   * @throws IOException
   *                              encodieren ist fehlgeschlagen
   */
  protected byte[] toWave(ByteBuffer micro, Telefonformat format) throws AudioFormatException, IOException {
    Raw2Wave raw2wave = new Raw2Wave(format);
    return raw2wave.encode(micro);
  }

}
