/*
 * Copyright © 2020, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.phone;

import javax.sound.sampled.FloatControl;
import javax.sound.sampled.LineEvent.Type;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer.Info;
import javax.sound.sampled.SourceDataLine;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;



/**
 * Audiodaten werden über einen Lautsprecher abgespielt.
 * 
 * @author llange
 *
 */
public class Speaker {

  private final static Logger log = LogManager.getLogger(Speaker.class);
  private SourceDataLine sourceDataline;

  /**
   * Erzeuge einen Lautsprecher für {@code speakername}. {@code speakername} muss
   * das Telefonformat unterstützen. {@code speakername} ist in der
   * Konfigurationsdatei als Pflichtfeld hinterlegt.
   * 
   * 
   * @param speakername
   *                    der Lautsprechername als Klartext
   * @param format
   *                    das Telefonformat
   * 
   * @throws MixerNotFoundException
   *                                  es wurde kein Mixer für {@code speakername }
   *                                  gefunden
   * @throws LineUnavailableException
   *                                  der Mixer stellt keine Leitung für
   *                                  {@code speakername} zur Verfügung
   */
  public Speaker(String speakername, Telefonformat format)
      throws MixerNotFoundException, LineUnavailableException {
    Info info = TelefonUtil.toSpeakermixer(speakername);
    sourceDataline = TelefonUtil.getSourceDataLine(info, format);
    sourceDataline.addLineListener((event) -> {
      if (event.getType() == Type.OPEN) {
//        if (log.isDebugEnabled()) log.debug("OPEN");

        // Type.START wird nur angesprungen, wenn zuvor sourceDataline.write(...)
        // aufgerufen wird
        // siehe ORACLE Dokumentation
        synchronized (this) {
          sourceDataline.write(new byte[4], 0, 4);
          sourceDataline.start();
        }
      }
      else if (event.getType() == Type.START) {
        synchronized (this) {
          sourceDataline.flush();
        }
      }
      else if (event.getType() == Type.STOP) {}
      else if (event.getType() == Type.CLOSE) {}
    });
    sourceDataline.open(TelefonUtil.getAudioFormat(format));
  }



  /**
   * Die Lautstärke wird eingestellt.
   * 
   * 
   * @param value
   *              die Lautstärke zwischen 0 und 102
   */
  public void setVolume(int value) {
    if (sourceDataline == null) {
      if (log.isDebugEnabled()) log.debug("Eine Audioleitung ist nicht geöffnet.");
      return;
    }

    // (y2-y1)/(x2-x1) 6.0206--80 / 102-0

//    float changedVolume = 0.860206f * value - 80f; // bis 100
    float changedVolume = 0.843339215f * value - 80f;
    FloatControl volume = (FloatControl) sourceDataline.getControl(FloatControl.Type.MASTER_GAIN);
    float gain = Math.min(Math.max(changedVolume, volume.getMinimum()), volume.getMaximum());
    volume.setValue(gain);
  }



  /**
   * Der Lautsprecher wird deaktiviert.
   * 
   */
  public void release() {
    if (sourceDataline == null) return;
    sourceDataline.stop();
    sourceDataline.close();
  }



  /**
   * Spiele die übergebenen Audiodaten ab. Diese Methode wird prallel von mehreren
   * Threads aufgerufen.
   * 
   * 
   * @param audiodata
   *                  Die Byte-Kolonne sind Audiodaten, die abgespielt werden
   */
  public void play(byte[] audiodata) {
    {
      if (!sourceDataline.isActive()) return;
      sourceDataline.write(audiodata, 0, Math.min(audiodata.length, sourceDataline.available()));
//      sourceDataline.drain();
    }
  }



  public SourceDataLine getSourceDataline() {
    return sourceDataline;
  }

}
