/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright © 2020, Luis Andrés Lange <https://javacomm.net>
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License Version 3.0, as described below:
 *
 * This file is free software: you may copy, redistribute and/or modify
 * it under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3.0 of the License, or (at your
 * option) any later version.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 */
package net.javacomm.client.phone;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import io.nayuki.flac.decode.DataFormatException;



/**
 * Ein Kommunikationsendpunkt für Audiopakete, die mit UDP empfangen und
 * versendet werden. Der Empfang erfolgt über eine Task.
 * 
 * @author llange
 * @deprecated Übertragungen finden über TCP statt.
 *
 */
@Deprecated
public class UDPEndpoint {

  private final static Logger log = LogManager.getLogger(UDPEndpoint.class);

  private int port = 29477;
  private DatagramSocket socket;
  private Telefonformat telefonformat;
  private List<UDPListener> udplistenerList;
  private boolean isRunning = false;

  /**
   * Der lokale Telefonserver wird erstellt aber nicht gestartet.
   * 
   */
  public UDPEndpoint() {
    this.setTelefonformat(Telefonformat.MEDIUM);
    udplistenerList = Collections.synchronizedList(new LinkedList<UDPListener>());
  }



  /**
   * 
   * Der lokale Telefonserver wird erstellt aber nicht gestartet.
   * 
   * 
   * @param port
   *             der Serverport
   */
  public UDPEndpoint(int port) {
    this();
    this.setPort(port);
  }



  /**
   * Der Listener horcht eingehende Nachrichten auf dem UDP-Port ab.
   * 
   * @param listener
   *                 ein Listener
   */
  public synchronized void addUdpListener(UDPListener listener) {
    udplistenerList.add(listener);
  }



  /**
   * Der übergebene Listener wird entfernt.
   * 
   * @param listener
   *                 ein Listener
   */
  public synchronized void removeUdpListener(UDPListener listener) {
    udplistenerList.remove(listener);
  }



  /**
   * Alle Listener werden entfernt.
   */
  private synchronized void removeAllListener() {
    udplistenerList.clear();
  }



  /**
   * 
   * Eine Nachricht wird versendet.
   * 
   * @param toHost
   *               der Zielrechner
   * @param toPort
   *               der Zielport
   * @param data
   *               die gesendeten Daten
   * @throws IOException
   *                     die Daten konnten nicht gesendet werden
   * 
   */
  public void send(String toHost, int toPort, byte[] data) throws IOException {
    send(toHost, toPort, data, data.length);
  }



  /**
   * 
   * Eine Nachricht wird versendet.
   * 
   * @param toHost
   *               der Zielrechner
   * @param toPort
   *               der Zielport
   * @param data
   *               die gesendeten Daten
   * @param len
   *               die Anzahl der zu sendenden Bytes
   * @throws IOException
   *                     die Daten konnten nicht gesendet werden
   */
  public void send(String toHost, int toPort, byte[] data, int len) throws IOException {

    DatagramPacket packet = new DatagramPacket(data, len, InetAddress.getByName(toHost), toPort);
    if (socket == null) return; // send wird vom Mikrofonthread aufgerufen.
    // Eingerichtet wird die Task aber vom UDP-Thread.
    // Die UDP-Thread-Task kann null sein.
    socket.send(packet);
  }



  /**
   * Der lokale Socket wird geschlossen. Die Verbindung zum Gesprächspartner wird
   * unterbrochen.
   * 
   * 
   */
  public void shutdown() {
    synchronized (this) {
      if (socket == null) return;
      if (socket.isConnected()) socket.disconnect();
      socket.close();
    }
  }



  /**
   * Über das Telefonformat wird die Puffergröße des Empfängerports festgelegt.
   * Wenn der Port aktiv ist, kann das Telefonformat nicht nachträglich geändert
   * werden.
   * 
   * 
   * @param value
   *              Telefonformat
   * @throws UDPRuntimeException
   *                             der Port ist aktiv
   */
  public void setTelefonformat(Telefonformat value) throws UDPRuntimeException {
    if (isRunning) throw new UDPRuntimeException("Der Port ist aktiv");
    telefonformat = value;
  }



  public Telefonformat getTelefonformat() {
    return telefonformat;
  }



  /**
   * Ein Port zwischen 1025 und 65535. Ein einmal aktivierter Port kann nicht
   * nachträglich geändert werden.
   * 
   * @param value
   *              der Zielport
   * @throws PortException
   * @throws UDPRuntimeException
   *                             der Port ist aktiv
   */
  public void setPort(int value) throws PortException, UDPRuntimeException {
    if (isRunning) throw new UDPRuntimeException("Der Port ist aktiv");
    if (1025 > value || value > 65535)
      throw new PortException(String.valueOf(value) + " -----> valid port range [1025...65535]");
    port = value;
  }



  /**
   * Der entfernte Rechner mit seinem Port.
   * 
   * @return der Zielport
   */
  public int getPort() {
    return port;
  }



  /**
   * Die ReceiverTask läuft als Daemon. Einmal gestartet kann sie nur über
   * {@link #shutdown()} abgebrochen werden. Die empfangenen Audiopakete werden an
   * den Lautsprecher weitergegeben. Bevor die Task terminiert, werden alle
   * UDPListener entfernt.
   * 
   * @return ReceiverTask für eingehende Audiopakete
   */
  public Runnable udptask() {
    Runnable runnable = () -> {
      isRunning = true;
      try {
        socket = new DatagramSocket(this.getPort());
        byte[] buffer = new byte[RecordMic.UDPPUFFER];
        while (true) {
          DatagramPacket flac = new DatagramPacket(buffer, buffer.length);
          if (socket == null) break;
          socket.receive(flac); // wird geblockt
          // neue Task
          // Die Pakete werden an den Speaker geschickt
          // Die Pakete müssen von FLAC nach RAW frcodiert werden

          try {

//            byte[] wavebyteArray = DecodeFlacToWav.entry(flac.getData());
//            log.info("(flac(wave) (" + flac.getData().length + "/" + wavebyteArray + ")");
//            UDPEvent udpevent = new UDPEvent(this, wavebyteArray);

            UDPEvent udpevent = new UDPEvent(this, flac.getData());
            for (UDPListener tmp : udplistenerList) {
              tmp.onReceive(udpevent);
            }
          }
          catch (DataFormatException e) {
            log.info(e.getMessage(), e);
          }
        }
      }
      catch (IOException e) {
        log.debug(e.getMessage(), e.getCause());
        log.info("Der UDP-Port wurde geschlossen. Die Task wird geschlossen.");
      }
      finally {
        UDPEvent udpevent = new UDPEvent(this);
        for (UDPListener tmp : udplistenerList) {
          tmp.onClose(udpevent);
        }
        removeAllListener();
      }
    };
    return runnable;
  }

}
