/**
 * Copyright © 2022, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.screencaster;

import jakarta.ws.rs.client.Client;
import jakarta.ws.rs.client.ClientBuilder;
import jakarta.ws.rs.client.Entity;
import jakarta.ws.rs.core.MediaType;
import java.awt.AWTException;
import java.awt.EventQueue;
import java.awt.Graphics2D;
import java.awt.MouseInfo;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Robot;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.concurrent.ArrayBlockingQueue;
import javax.imageio.ImageIO;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import org.glassfish.jersey.media.multipart.MultiPartFeature;
import net.javacomm.client.resource.Resource;
import net.javacomm.share.Constants;



/**
 * Der ImageCaster ist ein Producer-Consumer. Er erstellt Screenshots(Producer)
 * und sendet sie an den Webserver(Consumer) weiter. Der Producer ist ein
 * eigenständiger Thread und der Consumer ist eigenständiger Thread.
 */
public class ImageCaster {

  private final static Logger log = LogManager.getLogger(ImageCaster.class);
  private final static int COUNT_THREADS = 2;
  private Robot robot;
  private Rectangle screenSize;
  private ArrayBlockingQueue<byte[]> queue = new ArrayBlockingQueue<>(COUNT_THREADS);
  private Producer p1;
  private Consumer c1;

  private Thread producer_1;
  private Thread consumer_1;

  private static ImageCaster caster;

  private ImageCaster() throws AWTException {
    robot = new Robot();
    screenSize = new Rectangle(Toolkit.getDefaultToolkit().getScreenSize());
  }

  static {
    try {
      caster = new ImageCaster();
    }
    catch (AWTException e) {
      log.fatal(e.getMessage(), e);
    }
  }

  public static synchronized ImageCaster getInstance() {
    return caster;
  }



  /**
   * Die Aufzeichnung von Screenshots beginnt.
   * 
   * 
   * @param senderUserid
   *                     die Screenshots stammen von diesem User
   */
  public void recordScreenshots(String senderUserid) {
    stopRecording();

    p1 = new Producer(queue, robot, screenSize);
    c1 = new Consumer(queue, senderUserid); // userid ist der Sender

    producer_1 = new Thread(Thread.currentThread().getThreadGroup(), p1, "screenshot");
    consumer_1 = new Thread(Thread.currentThread().getThreadGroup(), c1, "sendToDatabase");

    consumer_1.start();
    producer_1.start();

  }



  /**
   * 
   * Screenshots werden nicht mehr aufgezeichnet und versendet. Ein mehrmaliger
   * Aufruf bewirkt nichts Neues.
   * 
   */
  public void stopRecording() {
    if (consumer_1 != null) consumer_1.interrupt();
    if (producer_1 != null) producer_1.interrupt();
    try {
      if (consumer_1 != null) consumer_1.join();
    }
    catch (InterruptedException e) {
      log.error(e.getMessage(), e);
    }
    try {
      if (producer_1 != null) producer_1.join();
    }
    catch (InterruptedException e) {
      log.error(e.getMessage(), e);
    }
  }

  /**
   * Über den Consumer werden die Screenshots an den Server gesendet.
   * 
   * 
   * @author llange
   *
   */
  private static class Consumer implements Runnable {

    private ArrayBlockingQueue<byte[]> queue;
    private String userid;

    /**
     * Der Consumer wird initialisiert.
     * 
     * @param queue
     *               aus dieser Queue werden die Bilder gelesen.
     * @param userid
     *               an diesen Anwender werden die Bilder gesendet
     */
    public Consumer(ArrayBlockingQueue<byte[]> queue, String userid) {
      this.queue = queue;
      this.userid = userid;
    }



    @Override
    public void run() {
      StringBuilder url = new StringBuilder(Constants.HTTP).append(Constants.DOMAIN)
          .append("/javacommserver/screenshot/save");
      Client client = null;
      while (!Thread.currentThread().isInterrupted()) {
        try {
          client = ClientBuilder.newClient();
          client.register(MultiPartFeature.class);
          try(FormDataMultiPart multipart = new FormDataMultiPart()) {
            byte[] picture = queue.take(); // aus der Queue lesen
            multipart.field("userid", userid);
            multipart.field("image", picture, MediaType.APPLICATION_OCTET_STREAM_TYPE);
            client.target(url.toString()).request(MediaType.TEXT_PLAIN)
                .post(Entity.entity(multipart, multipart.getMediaType()), Boolean.class);

          }
          catch (IOException | InterruptedException e) {
            break;
          }
        }
        finally {
          client.close();
        }
      }
    }

  }

  /**
   * Der Producer erzeugt Screenshots im WebP-Format.
   * 
   * 
   * @author llange
   *
   */
  private static class Producer implements Runnable {

    private Robot robot;
    private Rectangle screenSize;
    private ArrayBlockingQueue<byte[]> queue;
    private BufferedImage duke = null;
    BufferedImage image;
    ByteArrayOutputStream outputStream;

    /**
     * Der Producer wird initialisiert.
     * 
     * 
     * @param queue
     *                   in diese Queue werden die Screenshots geschrieben
     * @param robot
     * @param screenSize
     *                   die Bildschirmgröße
     */
    public Producer(ArrayBlockingQueue<byte[]> queue, Robot robot, Rectangle screenSize) {
      this.queue = queue;
      this.robot = robot;
      this.screenSize = screenSize;
      outputStream = new ByteArrayOutputStream();
      ImageIO.setUseCache(false);
      try {
        duke = ImageIO.read(getClass().getResource(Resource.DUKEMAUS));
      }
      catch (IOException e) {
        log.warn("Mauscursor konnte nicht gelesen werden");
      }

    }



    public static DataBufferByte convertIntToByte(DataBufferInt intBuffer) {
      int size = intBuffer.getSize();
      byte[] byteArray = new byte[size * 4]; // 4 Bytes pro int

      for (int i = 0; i < size; i++) {
        int value = intBuffer.getElem(i);
        byteArray[i * 4] = (byte) (value >> 24); // Höchstes Byte
        byteArray[i * 4 + 1] = (byte) (value >> 16);
        byteArray[i * 4 + 2] = (byte) (value >> 8);
        byteArray[i * 4 + 3] = (byte) value; // Niedrigstes Byte
      }

      return new DataBufferByte(byteArray, byteArray.length);
    }



    // Helper method to extract image bytes
    private byte[] extractImageBytes(BufferedImage image, boolean hasAlpha) {
      int width = image.getWidth();
      int height = image.getHeight();
      int bytesPerPixel = hasAlpha ? 4 : 3;

      byte[] imageBytes = null;

      DataBuffer databuffer = image.getRaster().getDataBuffer();
      if (databuffer.getDataType() == DataBuffer.TYPE_INT) {
        imageBytes = convertIntToByte((DataBufferInt) databuffer).getData();
      }
      else if (databuffer.getDataType() == DataBuffer.TYPE_BYTE) {
        imageBytes = ((DataBufferByte) image.getRaster().getDataBuffer()).getData();
      }
      else {
        throw new UnsupportedOperationException();
      }

//      byte[] imageBytes = ((DataBufferByte) image.getRaster().getDataBuffer()).getData();

      // Convert RGB(A) to BGR(A)
      byte[] output = new byte[width * height * bytesPerPixel];
      for (int i = 0; i < width * height; i++) {
        int index = i * bytesPerPixel;
        output[index] = imageBytes[index + 2]; // R
        output[index + 1] = imageBytes[index + 1]; // G
        output[index + 2] = imageBytes[index]; // B
        if (hasAlpha) {
          output[index + 3] = imageBytes[index + 3]; // A
        }
      }

      return output;
    }



    @Override
    public void run() {
      if (duke == null) return;
      // Instantiate WebP4j
//      WebP4j webP4j = new WebP4j();
      while (!Thread.currentThread().isInterrupted()) {

        try {
          EventQueue.invokeAndWait(() -> {
            image = robot.createScreenCapture(screenSize);
          });
        }
        catch (InvocationTargetException | InterruptedException e1) {
          return;
        }

        Graphics2D g2 = image.createGraphics();
        Point mousepointer = MouseInfo.getPointerInfo().getLocation();
        g2.drawImage(duke, mousepointer.x, mousepointer.y, null);
        g2.dispose();

//        // Get image width, height, and byte array
//        int width = image.getWidth();
//        int height = image.getHeight();
//        byte[] imageBytes = extractImageBytes(image, false);
//
//        // Encode image to WebP
//        float quality = 90.0f;
//        byte[] encodedWebP = webP4j.encodeRGB(imageBytes, width, height, width * 3, quality);

        try {
          outputStream.reset();
          ImageIO.write(image, "png", outputStream);
          image.flush();
          queue.put(outputStream.toByteArray());
//          queue.put(encodedWebP);
        }
        catch (IOException e) {
          log.error("ImageIO - " + e.getMessage());
        }
        catch (InterruptedException e) {
          return;
        }
      }
    }
  }

  public Rectangle getScreenSize() {
    return screenSize;
  }

}
