/*
 * Copyright © 2025, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.screencaster;

import jakarta.ws.rs.ProcessingException;
import jakarta.ws.rs.client.Client;
import jakarta.ws.rs.client.ClientBuilder;
import jakarta.ws.rs.client.Entity;
import jakarta.ws.rs.core.MediaType;
import jakarta.ws.rs.core.Response;
import java.io.IOException;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.stream.Stream;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import org.glassfish.jersey.media.multipart.MultiPartFeature;
import net.javacomm.restserver.Wrapper;



/**
 * Der NetworkConsumer wird von einem Producer mit Daten gespeist. Diese Daten
 * sendet er an einen Server weiter.
 */
public class NetworkConsumer {

  private final static Logger log = LogManager.getLogger(NetworkConsumer.class);
  private ArrayBlockingQueue<byte[]> queue;

  /**
   * Der Consumer wird initialisiert.
   * 
   * @param queue
   *               aus dieser Queue werden die Bilder gelesen.
   * @param userid
   *               an diese Anwender werden die Bilder gesendet
   * @param url
   *               an diese Serveradresse werden die Bilder gesendet
   */
  public NetworkConsumer(ArrayBlockingQueue<byte[]> queue, String userid[], String url) {
    this.queue = queue;

    Client client = null;
    while (!Thread.currentThread().isInterrupted()) {
      client = ClientBuilder.newClient();
      client.register(MultiPartFeature.class);
      try(FormDataMultiPart formMultipart = new FormDataMultiPart()) {

        byte[] picture = queue.take(); // aus der Queue lesen

        formMultipart.field("userid", Wrapper.toString(Stream.of(userid).toList()));
        formMultipart.field("image", picture, MediaType.APPLICATION_OCTET_STREAM_TYPE);

//        StringBuilder url = new StringBuilder(Constants.HTTP).append(Constants.DOMAIN)
//            .append("/javacommserver/screenshot/save");

        /**
         * Das Bild wird asynchron an den Server gesendet.
         * 
         */
        Future<Response> response = client.target(url).request(MediaType.APPLICATION_JSON).async()
            .post(Entity.entity(formMultipart, formMultipart.getMediaType()));

        response.get(200, TimeUnit.MILLISECONDS); // 5 Bilder pro Sekunde

      }
      catch (ProcessingException | IOException e) {
        // Connection kann verloren gehen
        log.warn(e.getMessage());
      }
      catch (InterruptedException e) {
        break;
      }
      catch (ExecutionException e) {
        log.warn(e.getMessage(), e.getCause());
      }
      catch (TimeoutException e) {
//        log.error(e.getMessage(), e.getCause());
      }
      finally {
        client.close();
      }
    }
  }

}
