/*
 * Copyright © 2020, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.websocket;

import jakarta.websocket.ClientEndpoint;
import jakarta.websocket.CloseReason;
import jakarta.websocket.ContainerProvider;
import jakarta.websocket.DeploymentException;
import jakarta.websocket.OnClose;
import jakarta.websocket.OnError;
import jakarta.websocket.OnMessage;
import jakarta.websocket.OnOpen;
import jakarta.websocket.Session;
import jakarta.websocket.WebSocketContainer;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.client.websocket.WebsocketEvent.Event;
import net.javacomm.protocol.MESSAGE;
import net.javacomm.protocol.MessageDecoder;
import net.javacomm.protocol.MessageEncoder;



@ClientEndpoint(encoders = {MessageEncoder.class}, decoders = {MessageDecoder.class})
public class WebsocketClient {

  private final Logger log = LogManager.getLogger(WebsocketClient.class);
  private List<WebsocketListener> websocketListenerList = Collections
      .synchronizedList(new LinkedList<WebsocketListener>());
  private Session activeSession;
  private ReentrantLock lock;

  public WebsocketClient() {
    lock = new ReentrantLock(true);
  }



  /**
   * Ein Listener wird registriert, um über Ereignisse zu informieren.
   * 
   * @param listener
   *                 ein Listener
   */
  public synchronized void addWebsocketListener(WebsocketListener listener) {
    websocketListenerList.add(listener);
  }



  @OnClose
  public void close(Session session, CloseReason reason) {
    WebsocketEvent event = new WebsocketEvent(this, Event.ON_CLOSE);
    for (WebsocketListener listener : websocketListenerList) {
      listener.onClose(event);
    }
  }



  /**
   * Der Client möchte sich mit dem Server verbinden.
   * 
   * 
   * @param url
   *            eine Verbindungsroute zum Server. Die {@code url} kommt aus der
   *            Domänenliste.
   * @return das permanente Verbindungsobjekt zum Server. Die Lebensdauer ist
   *         theoretisch unbegrenzt bis eine Seite die Verbindung abbaut.
   * 
   * @throws DeploymentException
   *                             ein generelles Verbindungsproblem. Die
   *                             {@code Exception} enthält eine genauere
   *                             Fehlerbeschreibung.
   * @throws IOException
   *                             ein generischer Fehler
   * @throws URISyntaxException
   *                             der übergebene Parameter enthält Fehler, der zu
   *                             einer {@code URISyntaxException} führt.
   */
  public Session connectToServer(String url) throws DeploymentException, IOException, URISyntaxException {
    WebSocketContainer container = ContainerProvider.getWebSocketContainer();
    container.setDefaultMaxTextMessageBufferSize(32768);
    container.setDefaultMaxBinaryMessageBufferSize(32768);
    container.setAsyncSendTimeout(2000);
    Session sessionToSend = container.connectToServer(this, new URI(url));
    setSessionToSend(sessionToSend);
    return sessionToSend;
  }



  @OnError
  public void error(Session session, Throwable throwable) {
    WebsocketEvent event = new WebsocketEvent(this, Event.ON_ERROR);
    for (WebsocketListener listener : websocketListenerList) {
      listener.onError(event);
    }
  }



  @OnOpen
  public void onOpen(Session value) {
    WebsocketEvent event = new WebsocketEvent(this, Event.ON_OPEN);
    for (WebsocketListener listener : websocketListenerList) {
      listener.onOpen(event);
    }
  }



  /**
   * Der Client empfängt über diese Methode Textnachrichten. Die Nachricht wird an
   * einen Listener weitergegebn.
   * 
   * @param peer
   *                hält die Serververbindung
   * @param message
   *                diese Nachricht wurde empfangen
   */
  @OnMessage
  public void receiveMessage(Session peer, MESSAGE message) {
    WebsocketEvent event = new WebsocketEvent(this, message);
    for (WebsocketListener listener : websocketListenerList) {
      listener.onMessage(event);
    }
  }



  /**
   * Entferne einen Listener.
   * 
   * @param listener
   *                 der zu entfernende Listener
   */
  public void removeWebsocketListener(WebsocketListener listener) {
    websocketListenerList.remove(listener);
  }



  /**
   * Alle Anfragen sendet der Client über diese Methode. Diese Methode ist
   * threadsafe. Die Reihenfolge eingehender Nachrichten wird berücksichtigt.
   * 
   * @param message
   *                diese Botschaft wird gesendet
   * 
   * @return {@code true} die Botschaft wurde gesendet, {@code false} nicht
   *         gesendet
   * 
   */
  public boolean sendMessage(MESSAGE message) {
    if (log.isDebugEnabled()) log.debug("Command=" + message.getCommand() + "|" + message.getHeader());
    try {
      lock.lock();
      activeSession.getAsyncRemote().sendObject(message);
      return true;
    }
    catch (Exception e) {
      // IllegalStateException
      // Command=CALLPRIVATECHAT|REQUEST wsSession.closed
      // Der Fehler tritt auf, wenn der Anwender angemeldet ist und
      // Server zwischenzeitlich neu gestartet wurde.
      // Der Server ist für den Client nicht mehr vorhanden.
      // Der Client aber, er ist nach wie vor verbunden
      log.error("Command=" + message.getCommand() + "|" + message.getHeader());
      log.error(e.fillInStackTrace());
      return false;
    }
    finally {
      lock.unlock();
    }
  }



  void setSessionToSend(Session value) {
    lock.lock();
    activeSession = value;
    lock.unlock();
  }



  /**
   * Schliesst die aktuelle Session zum Server.
   * 
   */
  public void shutdown() {
    lock.lock();
    try {
      if (activeSession == null) return;
      try {
        activeSession.close();
      }
      catch (IOException e) {
        log.error(e.getMessage(), e.getCause());
      }
    }
    finally {
      lock.unlock();
    }
  }

}
