/**
 * Copyright © 2020, Luis Andrés Lange llange@javacomm.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.restserver;

import static net.javacomm.protocol.HEADER.CONFIRM;
import static net.javacomm.protocol.HEADER.ERROR;
import static net.javacomm.protocol.HEADER.REQUEST;
import static net.javacomm.protocol.HEADER.RESPONSE;
import jakarta.mail.Message;
import jakarta.mail.MessagingException;
import jakarta.mail.Session;
import jakarta.mail.Transport;
import jakarta.mail.internet.InternetAddress;
import jakarta.mail.internet.MimeMessage;
import jakarta.ws.rs.DefaultValue;
import jakarta.ws.rs.FormParam;
import jakarta.ws.rs.POST;
import jakarta.ws.rs.Path;
import jakarta.ws.rs.Produces;
import jakarta.ws.rs.WebApplicationException;
import jakarta.ws.rs.core.MediaType;
import java.sql.SQLException;
import javax.naming.NamingException;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.database.DatabaseException;
import net.javacomm.database.WebdatabaseImpl;
import net.javacomm.database.entity.Config;
import net.javacomm.facade.DatabaseService;
import net.javacomm.multilingual.MultilingualString;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;
import net.javacomm.protocol.Command;
import net.javacomm.protocol.Protocol;
import net.javacomm.protocol.SIGNIN;
import net.javacomm.share.Constants;



@Path("/signin")
public class Signinrs {

  private final Logger log = LogManager.getLogger(Signinrs.class);
  final String USER_ID = "userid";
  final String CONFIRMATION_KEY = "confirmationKey";
  private Config config;
  private WebdatabaseImpl databaseservice;

  public Signinrs() {
    databaseservice = (WebdatabaseImpl) WebdatabaseImpl.getInstance();
  }



  /**
   * 
   * Gesendet wird über HTTPS.
   * 
   * @param recipientMail
   *                         an diese Mailadresse wird gesendet
   * 
   * @param uid
   *                         dieser Anwender wird angeschrieben. <b>uid</b> ist
   *                         ein Teil der Bestätigungs-URL.
   * 
   * @param password
   *                         das neue Passwort. <b>Passwort</b> ist ein Teil der
   *                         Bestätigungs-URL.
   * @param lang
   *                         die Sprachkennung ist ein Teil der Bestätigungs-URL.
   * @param confirmation_key
   *                         der Bestätigungsschlüssel wird benötigt, wenn ein
   *                         altes Passort durch ein neues ersetzt wird.
   *                         <b>confirmation_key</b> darf {@code null} sein.
   * @throws DatabaseException
   *                           der SQL-Befehl ist nicht ausführbar
   */
  void sendMail(String recipientMail, String uid, String password, String lang, String confirmation_key)
      throws DatabaseException {
    try {
      config = databaseservice.fetchConfig();
      Session mailSession = databaseservice.getMailsession(config);
      Message message = new MimeMessage(mailSession);
      InternetAddress from = new InternetAddress(config.getMailSmtpUser());
      message.setFrom(from);
      message.setRecipient(Message.RecipientType.TO, new InternetAddress(recipientMail));
      MultilingualString betreff = new MultilingualString(KEY.SERVER_MAIL_BETREFF, ISO639.fromValue(lang));
      message.setSubject(betreff.toString());
      // richtige UID und password setzen

      MultilingualString mail = new MultilingualString(KEY.SERVER_MAIL_ERSTANMELDUNG, ISO639.fromValue(lang));

      String neuerwert = mail.toString().replace("XXX", uid);
      String neuerwert2 = neuerwert.replace("YYY", password);

      if (confirmation_key == null) {
        message.setText(neuerwert2);
        Transport.send(message);
        return;
      }
      StringBuilder buffer = new StringBuilder(neuerwert2);

      buffer.append("\r\n");
      buffer.append("\r\n\r\n");
      MultilingualString mailDasNeuePasswort = new MultilingualString(
          KEY.SERVER_MAIL_DAS_NEUE_PASSWORT, ISO639.fromValue(lang)
      );
      buffer.append(mailDasNeuePasswort.toString());
      buffer.append("\r\n\r\n");
      buffer.append(Constants.PROTOCOL);
      buffer.append(config.getDomain());
      buffer.append("/javacommserver/reset/password/");
      buffer.append(uid);
      buffer.append("/");
      buffer.append(confirmation_key);
      buffer.append("/");
      buffer.append(lang);
      buffer.append("\r\n\r\n");
      MultilingualString mailDasAltePassword = new MultilingualString(
          KEY.SERVER_MAIL_DAS_ALTE_PASSWORT, ISO639.fromValue(lang)
      );
      buffer.append(mailDasAltePassword);
      message.setText(buffer.toString());
      Transport.send(message);
      return;
    }
    catch (NamingException | MessagingException e) {
      log.error(e.getMessage(), e);
    }

  }



  void sendMail(String recipientMail, String uid, String password, String lang) throws DatabaseException {
    sendMail(recipientMail, uid, password, lang, null);
  }



  /**
   * Hier wird ohne IP gesendet
   * 
   * @param email
   *                   E-Mail vom Benutzer
   * @param lang
   *                   die Sprache
   * @param webconfirm
   *                   {@code true}, wenn die Anfrage von der Homepage kommt
   * 
   * @return SIGNIN-Response
   * 
   * @throws DatabaseException
   *                           der SQL-Befehl konnte nicht verarbeitet werden
   */
  private SIGNIN signIn(String email, String lang, boolean webconfirm) throws DatabaseException {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    SIGNIN request = new SIGNIN();
    request.setCommand(Command.SIGNIN);
    request.setHeader(REQUEST);
    request.setDataset(Protocol.DATASET);
    request.setEmail(email == null ? "" : email);
    SIGNIN response = databaseservice.signinUser(request.getEmail(), webconfirm);
    if (CONFIRM == response.getHeader()) {
      // Eine EMail wird gesendet
      sendMail(response.getEmail(), response.getUserid(), response.getPassword(), lang);
      MultilingualString text = webconfirm
          ? new MultilingualString(KEY.SERVER_WEBCONFIRM_PASSWORT, ISO639.fromValue(lang))
          : new MultilingualString(KEY.SERVER_CONFIRM_PASSWORT, ISO639.fromValue(lang));
      response.setText(text.toString());
    }
    else if (RESPONSE == response.getHeader()) {
      // ein neues PASSWORD wird angefordert
      SIGNIN signinChangePassword = databaseservice.createConfirmationKey(response.getUserid());
      signinChangePassword.setNickname(response.getNickname());
      signinChangePassword.setPassword(response.getPassword());
      signinChangePassword.setForegroundColor(response.getForegroundColor());
      signinChangePassword.setBackgroundColor(response.getBackgroundColor());
      signinChangePassword.setEmail(response.getEmail());
      signinChangePassword.setMultilingualkey(KEY.SERVER_NEUES_PASSWORT);
      sendMail(
          response.getEmail(), signinChangePassword.getUserid(), signinChangePassword.getNeuesPassword(),
          lang, signinChangePassword.getConfirmationKey()
      );
      databaseservice
          .updateConfirmationFlag(signinChangePassword.getUserid(), signinChangePassword.getEmail(), 1);
      // TODO response setText mit der richtigen Sprache setzen, den KEY habe ich
      // bereits in response
      MultilingualString text = new MultilingualString(response.getMultilingualkey(), ISO639.fromValue(lang));
      response.setText(text.toString());
    }
    else if (ERROR == response.getHeader()) {
      MultilingualString text = new MultilingualString(response.getMultilingualkey(), ISO639.fromValue(lang));
      response.setText(text.toString());
    }
    if (log.isDebugEnabled()) log.debug(response.toString());

    return response;
  }



  /**
   * Ein neuer Anwender möchte sich einschreiben.
   * 
   * @param email
   *                   der Anmelder
   * @param lang
   *                   die Spracheinstellung vom Anwender. Sie ist als
   *                   Pfadparameter in der Antwortmail enthalten
   * @param webconfirm
   *                   {@code true}, die Anfrage kommt von der Homepage, sonst vom
   *                   Client
   * 
   * @return JSON-String als encoded SIGNIN
   * 
   * @throws SQLException
   */
  @POST
  @Produces(MediaType.APPLICATION_JSON)
  public String subscribe(@FormParam("email") String email,
      @DefaultValue("en") @FormParam("lang") String lang,
      @DefaultValue("false") @FormParam("webconfirm") boolean webconfirm) {

    try {
      return signIn(email, lang, webconfirm).toString();
    }
    catch (DatabaseException e) {
      log.error(e.getMessage(), e);
      throw new WebApplicationException();
    }
  }

}
