/**
 * Copyright © 2021, Luis Andrés Lange llange@javacomm.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.restserver;

import jakarta.ws.rs.FormParam;
import jakarta.ws.rs.GET;
import jakarta.ws.rs.HEAD;
import jakarta.ws.rs.POST;
import jakarta.ws.rs.PUT;
import jakarta.ws.rs.Path;
import jakarta.ws.rs.PathParam;
import jakarta.ws.rs.Produces;
import jakarta.ws.rs.QueryParam;
import jakarta.ws.rs.core.MediaType;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.database.WebdatabaseImpl;
import net.javacomm.database.entity.EntityUser;
import net.javacomm.database.entity.UserModule;
import net.javacomm.facade.DatabaseService;
import net.javacomm.share.Constants;
import net.javacomm.transfer.Helado;
import net.javacomm.transfer.Modul;
import net.javacomm.transfer.Sprachguete;
import net.javacomm.transfer.TransferUser;



@Produces(MediaType.TEXT_PLAIN)
@Path("/user")
public class User {

  private final static Logger log = LogManager.getLogger(User.class);
  private DatabaseService database;

  public User() {
    database = WebdatabaseImpl.getInstance();
  }



  @Path("/volume")
  @POST
  public void writeTelefonvolume(@FormParam(value = "userid") String userid,
      @FormParam(value = "volume") int volume) {

    database.updatePhoneVolume(userid, volume);
  }



  @Path("/write/playervolume")
  @POST
  public boolean writePlayervolume(@FormParam(value = "userid") String userid,
      @FormParam(value = "playervolume") int playervolume) {

    if (0 > playervolume || playervolume > 102) return false;
    return database.updatePlayervolume(userid, playervolume);
  }



  @Path("/read/speaker")
  @GET
  public String readSpeaker(@QueryParam(value = "email") String email) {

    if (email == null) return "";
    return database.fetchSpeaker(email);
  }



  @Path("/read/playervolume")
  @POST
  public int readPlayervolume(@FormParam(value = "userid") String userid) {
    return database.fetchPlayervolume(userid);
  }



  /**
   * Sollen Chatanfragen abgelehnt werden?
   *
   *
   * @param userid
   *                   die Benutzer-ID vom Schreiber
   * @param chatdenied
   *                   {@code true}, Chatanfragen werden abgelehnt
   */
  @Path("/write/chatdenied")
  @POST
  public void writeChatdenied(@FormParam(value = "userid") String userid,
      @FormParam(value = "chatdenied") boolean chatdenied) {
    database.updateChatdenied(userid, chatdenied);
  }



  /**
   * Nur die Benutzerdaten für das Telefon werden gespeichert.
   *
   *
   * @param telefon
   *                die Telefondaten wie Telefonport, Sprachgüte und Userid
   * @return {@code true}, die Daten konnten gespeichert werden.
   */
  @Path("/write/telefon")
  @POST
  public boolean writeTelefon(@FormParam(value = "telefon") TransferUser telefon) {

    net.javacomm.database.entity.EntityUser user = new net.javacomm.database.entity.EntityUser();
    user.setUid(telefon.getUserid());
    user.setTelefonpuffer(telefon.getTelefonpuffer().getPuffer());
    user.setMikrofon(telefon.getMikrofon());
    user.setSpeaker(telefon.getSpeaker());
    boolean done = database.updateTelefon(user);
    return done;
  }



  /**
   * Der Anwender teilt mit, ob er Bildübertragungen annimmt.
   *
   * @param rekorder
   * @return {@code true}, der Anwender nimmt Bildübertragungen entgegen
   */
  @Path("/write/onrekorder")
  @POST
  public boolean writeOnrekorder(@FormParam(value = "rekorder") TransferUser rekorder) {

    EntityUser user = new EntityUser();
    user.setUid(rekorder.getUserid());
    user.setOnrekord(rekorder.getOnrekord());
    boolean done = database.updateOnrekorder(user);
    return done;
  }



  /**
   * Der Anwender teilt mit, dass sein Projektor läuft
   *
   * @param projektor
   * @return {@code true}, der Projektor ist in Betrieb
   */
  @Path("/write/onprojektor")
  @POST
  public boolean writeOnprojektor(@FormParam(value = "projektor") TransferUser projektor) {

    EntityUser user = new EntityUser();
    user.setUid(projektor.getUserid());
    user.setOnprojektor(projektor.getOnprojektor());
    boolean done = database.updateOnprojektor(user);
    return done;
  }



  /**
   * Der Anwender teilt seine Einwilligung oder Ablehnung für den
   * Bildschirmempfang mit.
   *
   * @param rekorder
   * @return {@code true}, die Daten konnten gespeichert werden.
   */
  @Path("/write/rekorder")
  @POST
  public boolean writeRekorder(@FormParam(value = "rekorder") TransferUser rekorder) {

    EntityUser user = new EntityUser();
    user.setUid(rekorder.getUserid());
    user.setOnrekord(rekorder.getOnrekord());
    boolean done = database.updateRekorder(user);
    return done;
  }



  @Path("/write/chattooltip")
  @HEAD
  public void writeChattooltip(@QueryParam(value = "chattooltip") boolean tooltip,
      @QueryParam(value = "userid") String userid) {
    database.updateChattooltip(tooltip, userid);
  }



  /**
   * Der Anwender teilt seine Einwilligung oder Ablehnung für den
   * Bildschirmempfang mit.
   *
   * @param rekorder
   * @return {@code true}, die Daten konnten gespeichert werden.
   */
  @Path("/write/rekorderport")
  @POST
  @Deprecated
  public boolean writeRekorderport(@FormParam(value = "rekorder") TransferUser rekorder) {

    if (
      Constants.MIN_PORT_RANGE > rekorder.getRekorderport()
          || rekorder.getRekorderport() > Constants.MAX_PORT_RANGE
    ) return false;
    boolean done = database.updateRekorderport(rekorder.getUserid(), rekorder.getRekorderport());
    return done;
  }



  @Path("/write/eis")
  @POST
  public void writeEis(@FormParam(value = "userid") String userid, @FormParam(value = "eis") Helado helado) {

    database.updateEis(userid, helado.getHelado());
  }



  /**
   * Globale Benutzerattribute werden gespeichert.
   *
   *
   */
  @Path("/write/global/attributes")
  @POST
  public void writeGlobalAttributes(@FormParam(value = "userid") String userid,
      @FormParam(value = "playervolume") int playervolume, @FormParam(value = "iptvvolume") int iptvvolume,
      @FormParam(value = "eis") Helado helado, @FormParam(value = "language") String language,
      @FormParam(value = "chatDownloadDir") String chatDownloadDir,
      @FormParam(value = "chatUploadDir") String chatUploadDir) {

    database.updateGlobalAttributes(
        userid, playervolume, iptvvolume, helado.getHelado(), language, chatDownloadDir, chatUploadDir
    );

  }



  /**
   * Alle Attribute für einen Anwender aus der Tabelle TB_USER werden gelesen. Der
   * Rückgabewert kann leer sein, wenn kein Anwender gefunden wurde.
   *
   * @param userid
   *               ein Anwender
   * @return die Attribute sind in genau einem Listenelement
   */
  @Path("/read/data/{userid}")
  @GET
  @Produces(MediaType.APPLICATION_JSON)
  public List<TransferUser> readData(@PathParam(value = "userid") String userid) {

    ArrayList<TransferUser> list = new ArrayList<>();
    net.javacomm.database.entity.EntityUser entityUser = database.fetchUser(userid);
    if (entityUser == null) return list;
    TransferUser transferuser = new TransferUser();
    transferuser.setUserid(entityUser.getUid());
    transferuser.setBenutzerstatus(entityUser.getBenutzerstatus());
    transferuser.setMail(entityUser.getMail());
    transferuser.setChatdenied(entityUser.isChatdenied());
    transferuser.setChattooltip(entityUser.isChattooltip());
    transferuser.setVolume(entityUser.getVolume());
    transferuser.setPlayervolume(entityUser.getPlayervolume());
    transferuser.setOncall(entityUser.isOncall());
    transferuser.setTelefonpuffer(Sprachguete.toSprachguete(entityUser.getTelefonpuffer()));
    transferuser.setHelado(Helado.toHelado(entityUser.getEis()));
    transferuser.setSpeaker(entityUser.getSpeaker());
    transferuser.setMikrofon(entityUser.getMikrofon());
    transferuser.setOnrekord(entityUser.isOnrekord());
    transferuser.setIptvvolume(entityUser.getIptvvolume());
    transferuser.setLanguage(entityUser.getLanguage());
    transferuser.setChatdownloaddir(entityUser.getChatDownloadDir());
    transferuser.setChatuploaddir(entityUser.getChatUploadDir());
    transferuser.setLdelete(entityUser.isLdelete());
    transferuser.setNickname(entityUser.getNickname());
//    if (log.isDebugEnabled()) log.debug(transferuser.toString());
    log.info(transferuser.toString());
    list.add(transferuser);
    return list;
  }



  /**
   * Hat der Anwender eine Zugriffsberechtigung für ODX?
   *
   *
   * @param userid
   *               ein Anwender
   *
   * @return {@code true}, der Anwender ist berechtigt
   *
   */
  @GET
  @Path("/read/modul/odx")
  @Produces(MediaType.TEXT_PLAIN)
  public boolean isODX(@QueryParam(value = "userid") String userid) {
    List<UserModule> list = database.fetchModules(userid);
    for (UserModule usermodule : list) {
      Modul module = Modul.toModule(usermodule.getModule());
      if (module == Modul.ODX) return true;
    }
    return false;
  }



  /**
   * KEEPALIVE ist eine Zeitscheibe und wird als Referenz für das Schließen der
   * Clientverbindung herangezogen.
   * 
   * @param userid
   *               dieser Anwender
   * @return {@code true}, der Zeitstempel konnte für den Anwender aktualisiert
   *         werden
   * 
   * @throws SQLException
   *                      generischer SQL-Fehler
   */
  @PUT
  @Path("/write/keepalive/{userid}")
  @Produces(MediaType.TEXT_PLAIN)
  public boolean keepAlive(@PathParam(value = "userid") String userid) throws SQLException {
    return database.updateKEEPALIVE(userid);
  }

}
