/**
 * Copyright © 2020, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.server;

import static net.javacomm.protocol.HEADER.CONFIRM;
import static net.javacomm.protocol.HEADER.ERROR;
import static net.javacomm.protocol.HEADER.REQUEST;
import static net.javacomm.protocol.HEADER.RESPONSE;
import jakarta.websocket.CloseReason;
import jakarta.websocket.CloseReason.CloseCodes;
import jakarta.websocket.EncodeException;
import jakarta.websocket.Endpoint;
import jakarta.websocket.EndpointConfig;
import jakarta.websocket.MessageHandler;
import jakarta.websocket.Session;
import jakarta.websocket.server.ServerEndpoint;
import java.io.IOException;
import java.sql.SQLException;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Stream;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.database.DatabaseException;
import net.javacomm.database.WebdatabaseImpl;
import net.javacomm.database.entity.EntityUser;
import net.javacomm.database.entity.Online;
import net.javacomm.database.entity.PrivateChatfile;
import net.javacomm.database.entity.Record;
import net.javacomm.database.entity.Screencast;
import net.javacomm.facade.DatabaseService;
import net.javacomm.multilingual.schema.KEY;
import net.javacomm.protocol.BEAMEROFF;
import net.javacomm.protocol.BREAKROOMS;
import net.javacomm.protocol.BlacklistTypes;
import net.javacomm.protocol.CALLPRIVATECHAT;
import net.javacomm.protocol.CALLREMOTEUSER;
import net.javacomm.protocol.CANDIDATETOPICMEMBER;
import net.javacomm.protocol.CHANGETOPICMEMBER;
import net.javacomm.protocol.CHATMESSAGE;
import net.javacomm.protocol.CHATONLINELIST;
import net.javacomm.protocol.CHATUSERLIST;
import net.javacomm.protocol.CONFERENCE;
import net.javacomm.protocol.CONFERENCEAUDIO;
import net.javacomm.protocol.CONFERENCEMUTE;
import net.javacomm.protocol.CONFERENCEVIDEO;
import net.javacomm.protocol.CREATETEMPROOM;
import net.javacomm.protocol.Command;
import net.javacomm.protocol.DELETEROOM;
import net.javacomm.protocol.DELETEUPLOADFILES;
import net.javacomm.protocol.DIAL;
import net.javacomm.protocol.DOWNLOAD;
import net.javacomm.protocol.ENTERROOM;
import net.javacomm.protocol.FILETYPES;
import net.javacomm.protocol.HISTORYMESSAGE;
import net.javacomm.protocol.IMAGE;
import net.javacomm.protocol.INCOMINGCALL;
import net.javacomm.protocol.LEAVEPRIVATECHAT;
import net.javacomm.protocol.LEAVEROOM;
import net.javacomm.protocol.MESSAGE;
import net.javacomm.protocol.MICROERROR;
import net.javacomm.protocol.MessageDecoder;
import net.javacomm.protocol.MessageEncoder;
import net.javacomm.protocol.ONCALL;
import net.javacomm.protocol.ONHOOK;
import net.javacomm.protocol.PRIVATECHATFILE;
import net.javacomm.protocol.PRIVATEMESSAGE;
import net.javacomm.protocol.PROYECTORCLOSING;
import net.javacomm.protocol.Protocol;
import net.javacomm.protocol.READROOMS;
import net.javacomm.protocol.READTOPICROOMOWNER;
import net.javacomm.protocol.ROOMLIST;
import net.javacomm.protocol.SCREENCAST;
import net.javacomm.protocol.SEARCHFILES;
import net.javacomm.protocol.Searchfile;
import net.javacomm.protocol.TCPAUDIO;
import net.javacomm.protocol.TOPICMEMBER;
import net.javacomm.protocol.UPDATEFILETYPES;
import net.javacomm.protocol.UPDATEPHONE;
import net.javacomm.protocol.UPDATEPORT;
import net.javacomm.protocol.UPDATEUSER;
import net.javacomm.protocol.UPLOADFILES;
import net.javacomm.protocol.USERONLINELIST;
import net.javacomm.protocol.USRLOGIN;



@ServerEndpoint(value = "/portal", encoders = {MessageEncoder.class}, decoders = {MessageDecoder.class})
public class Webserver extends Endpoint {

  private final static Logger log = LogManager.getLogger(Webserver.class);

  private static ReentrantLock lock = new ReentrantLock(true);

  /**
   * key/value ---> (sessionId/Session) sessionId ist ein Datenbankwert
   */
  private volatile static ConcurrentHashMap<String, Session> sessionmap;

  private static ScheduledExecutorService executorKeepAlive;

  static Runnable keepalvieRunnable = () -> {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();

    // Lösche alle Screenshots
//      impl.deleteScreenshots();
    int minutos = databaseservice.forumCycle();
    int result = databaseservice.changeToForum(minutos); // in ein Forum wandeln 150 Minuten

    if (result > 0) {
      // wenn gewandelt, dann Raumliste aktualisieren
      // kann auch wieder einen Zähler einbauen
      // Für Testzwecke auf 5 Minuten eingestellt
      // Roomliste übergeben
      try {
        ROOMLIST roomlist = databaseservice.fetchRoomlist(); // Alle vorhandenen Chaträume werden ermittelt
        // sende eine vollständige Raumliste an alle Anwender
        // ROMMLIST beschreibt die Räume(Pausenraum, Forum etc.) und nnicht die Anzahl
        // der User und deren Attribute JLauncher Alle Räume anzeigen
        broadcastAll(roomlist);
      }
      catch (DatabaseException e2) {
        log.warn(e2.getMessage(), e2.getCause());
        return;
      }
    }

    try {
      DateTimeFormatter formatter = DateTimeFormatter.ofLocalizedDateTime(FormatStyle.LONG)
          .withLocale(Locale.getDefault());

      CloseReason reason = new CloseReason(CloseCodes.NORMAL_CLOSURE, "KeepAlive exprired");

      List<Online> online = Collections.synchronizedList(databaseservice.deleteOnlineKeepalive());
      online.forEach(user -> {

        ZonedDateTime zdtKeepalive = user.getKeepalive().withZoneSameInstant(ZoneId.systemDefault());
        String datumKeepAlive = zdtKeepalive.format(formatter);

        log.info(user.getUid());
        log.info(datumKeepAlive);
        log.info("keepalive Zeitüberschreitung vom user mit der Session = " + user.getSession());

        databaseservice.deleteOnline(user.getUid()); // Lösche user aus der TB_ONLINE, wegen Patrick
        databaseservice.deletePhoneUser(user.getUid()); // Lösche aus TB_PHONE

        Session closeSession = sessionmap.get(user.getSession());
        try {
          closeSession.close(reason);
        }
        catch (IOException e) {
          log.info(e.getMessage(), e);
        }
      });
    }
    catch (SQLException e) {
      log.error(e);
    }
  };

  /**
   * Sende eine Nachricht an alle Anwender.
   * 
   * @param message
   *                diese Nachricht wird gesendet
   */
  public static void broadcastAll(MESSAGE message) {
    try {
      lock.lock();
      sessionmap.values().forEach(client -> {
        try {
          client.getAsyncRemote().sendObject(message);
        }
        catch (Exception e) {
          log.error(e.getMessage(), e.getCause());
        }

      });
    }
    finally {
      lock.unlock();
    }
  }



  /**
   * Der Server sendet eine einzelne Botschaft an den Client. Diese Methode ist
   * threadsafe.
   * 
   * @param message
   *                      diese Botschaft
   * @param sessionmapKey
   *                      an diesen Client wird gesendet
   */
  public static void sendMessage(MESSAGE message, String sessionmapKey) {
    try {
      lock.lock();
      Session client = sessionmap.get(sessionmapKey);
      try {
        client.getAsyncRemote().sendObject(message);
      }
      catch (Exception e) {
        log.error(e.getMessage(), e.getCause());
      }
    }
    finally {
      lock.unlock();
    }
  }



  public static void sendMessage(MESSAGE message, Session websocketSession) {
    try {
      lock.lock();
      try {
        websocketSession.getAsyncRemote().sendObject(message);
      }
      catch (Exception e) {
        log.error(e.getMessage(), e.getCause());
      }
    }
    finally {
      lock.unlock();
    }
  }



  public static void sendMessage(MESSAGE message, String[] receiverSessions) {
    try {
      lock.lock();
      Stream.of(receiverSessions).parallel().forEach(target -> {
        try {
          Session client = sessionmap.get(target);
          if (client == null) return;
          client.getAsyncRemote().sendObject(message);
        }
        catch (Exception e) {
          log.error("CONFERENCEAUDIO - " + target);
        }
      });
    }
    finally {
      lock.unlock();
    }
  }



  /**
   * Die Application wird gerade gestartet und als erstes werden die
   * Hintergrundthreads gestartet.
   */
  public static void startup() {
    executorKeepAlive = Executors.newSingleThreadScheduledExecutor(runnable -> {
      return new Thread(Thread.currentThread().getThreadGroup(), runnable, "keepalive");
    });
    sessionmap = new ConcurrentHashMap<String, Session>();
    executorKeepAlive.scheduleWithFixedDelay(keepalvieRunnable, 20, 60, TimeUnit.SECONDS);
  }



  /**
   * Jeder Login startet eine neue Instance.
   */
  public Webserver() {
    log.info("new client instance has started...");
  }



  /**
   * 
   * Bekommt eine confirm response, wenn die gegenstelle bereit ist Bekommt eine
   * positiv response, wenn die Gegenstelle nicht bereit ist
   * Rufnummerunterdrückung Bekommt eine error response, wenn die gegenstelle
   * offline ist
   * 
   * @param request
   * @return
   */
  private CALLPRIVATECHAT callPrivatechat(CALLREMOTEUSER remoteuser) {
    if (CONFIRM == remoteuser.getHeader()) {
      CALLPRIVATECHAT call = new CALLPRIVATECHAT();
      call.setCommand(Command.CALLPRIVATECHAT);
      call.setHeader(CONFIRM);
      call.setDataset(Protocol.DATASET);
      call.setLocalNickname(remoteuser.getLocalNickname());
      call.setLocalSessionid(remoteuser.getLocalSessionid());
      call.setRemoteNickname(remoteuser.getRemoteNickname());
      call.setRemoteSessionid(remoteuser.getRemoteSessionid());
      return call;
    }
    else if (RESPONSE == remoteuser.getHeader()) {
      CALLPRIVATECHAT call = new CALLPRIVATECHAT();
      call.setCommand(Command.CALLPRIVATECHAT);
      call.setHeader(RESPONSE);
      call.setDataset(Protocol.DATASET);
      call.setLocalNickname(remoteuser.getLocalNickname());
      call.setLocalSessionid(remoteuser.getLocalSessionid());
      call.setRemoteNickname(remoteuser.getRemoteNickname());
      call.setRemoteSessionid(remoteuser.getRemoteSessionid());
      call.setMultilingualkey(KEY.SERVER_LEHNT_CHATANFRAGEN_AB);
      return call;
    }
    CALLPRIVATECHAT call = new CALLPRIVATECHAT();
    return call;

  }



  private CANDIDATETOPICMEMBER candidateMember(CANDIDATETOPICMEMBER request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    CANDIDATETOPICMEMBER response = databaseservice.fetchCandidatemember(request);
    return response;
  }



  private CHANGETOPICMEMBER changeTopicmember(CHANGETOPICMEMBER request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    CHANGETOPICMEMBER response = databaseservice.changeMitglieder(
        request.getUserid(), request.getRoom(), request.getRoomtype(), request.getChatUser()
    );
    return response;
  }



  private CHATMESSAGE chatmessage(CHATMESSAGE request) {
    CHATMESSAGE response = new CHATMESSAGE();
    if (REQUEST == request.getHeader()) {
      DatabaseService databaseservice = WebdatabaseImpl.getInstance();
      try {
        Record entity = databaseservice.insertChatmessage(
            request.getRoom(), request.getChatUser().getUserid(), request.getAttachment(),
            request.getMessage()
        );

        response.setCommand(Command.CHATMESSAGE);
        response.setHeader(RESPONSE);
        response.setDataset(Protocol.DATASET);
        response.setFilename(request.getFilename()); // Protocollaänderunge nicht übertragen
        response.setFilesize(request.getFilesize()); // Protocollaänderunge nicht übertragen
        response.setChatUser(request.getChatUser());
        response.setRoom(entity.getChatid());
        response.setAttachment(entity.getAnlage());
        response.setMessage(entity.getMessage());

        // Zeit ist nicht aus der Datenbank
        ZonedDateTime tz = ZonedDateTime.now(ZoneId.of(DatabaseService.TIMEZONE));
        long value = tz.toInstant().toEpochMilli();
        response.setDatetime(value);

        if (response.getAttachment() != null) databaseservice.updateDeletable(true, response.getAttachment());
        return response;
      }
      catch (DatabaseException e) {
        log.info(e.getMessage());
        response = new CHATMESSAGE();
        response.setCommand(Command.CHATMESSAGE);
        response.setHeader(ERROR);
        response.setDataset(Protocol.DATASET);
        response.setErrorMessage(e.getMessage());
      }
    }
    else {
      response.setCommand(Command.CHATMESSAGE);
      response.setHeader(ERROR);
      response.setDataset(Protocol.DATASET);
      response.setErrorMessage("CHATMESSAGE ist kein REQUEST");
    }
    return response;
  }



  CREATETEMPROOM createTemproom(CREATETEMPROOM request) {
    if (REQUEST == request.getHeader()) {
      DatabaseService databaseservice = WebdatabaseImpl.getInstance();
      CREATETEMPROOM response = databaseservice.createRoom(request);
      return response;
    }
    return new CREATETEMPROOM();
  }



  private DELETEROOM deleteRoom(DELETEROOM request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    boolean done = databaseservice.deleteRoom(request.getRoom(), request.getUserid());
    DELETEROOM response = new DELETEROOM();
    if (done) {
      response.setCommand(Command.DELETEROOM);
      response.setHeader(RESPONSE);
      response.setDataset(Protocol.DATASET);
      response.setRoom(request.getRoom());
      response.setUserid(request.getUserid());
      StringBuffer buffer = new StringBuffer();
      buffer.append("Der Besprechungsraum <b><span style=\"color:red\">");
      buffer.append(request.getRoom());
      buffer.append("</b></span> wurde gelöscht.");
      response.setText(buffer.toString());
    }
    else {
      response.setCommand(Command.DELETEROOM);
      response.setHeader(ERROR);
      response.setDataset(Protocol.DATASET);
      response.setRoom(request.getRoom());
      response.setUserid(request.getUserid());
      StringBuffer buffer = new StringBuffer();
      buffer.append("Diesen Besprechungsraum <b><span style=\"color:red\">");
      buffer.append(request.getRoom());
      buffer.append("</b></span> gibt es nicht.");
      response.setText(buffer.toString());
    }
    return response;
  }



  private DELETEUPLOADFILES deleteuploadfiles(DELETEUPLOADFILES request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    try {
      databaseservice.deleteUploadfiles(request.getUserid(), request.getIp());
      DELETEUPLOADFILES deletefiles = new DELETEUPLOADFILES();
      deletefiles.setCommand(Command.DELETEUPLOADFILES);
      deletefiles.setHeader(CONFIRM);
      return deletefiles;
    }
    catch (SQLException e) {
      log.warn(e.getMessage(), e.getCause());
      DELETEUPLOADFILES deletefiles = new DELETEUPLOADFILES();
      deletefiles.setCommand(Command.DELETEUPLOADFILES);
      deletefiles.setHeader(ERROR);
      deletefiles.setDataset(Protocol.DATASET);
      deletefiles.setErrorCode(e.getErrorCode());
      deletefiles.setErrorMessage(e.getMessage());
      deletefiles.setText("A database error has occured.");
      return deletefiles;
    }
  }



  /**
   * Der Anwender mit der {@code sessionid} ist gegangen.
   * 
   * @param sessionid
   *                  ein eindeutiger Onlineschlüssel
   */
  private void disconnect(String sessionid) {
    try {
      lock.lock();
      sessionmap.remove(sessionid);
    }
    finally {
      lock.unlock();
    }
    // in welchen Räumen war der User vor dem Löschen?
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    List<String> roomlist = databaseservice.fetchRoomlist(sessionid);
    removeUser(sessionid); // löscht den User aus allen Onlinelisten
    broadcast(roomlist);
  }



  private ENTERROOM enterroom(ENTERROOM request) {
    if (request.getHeader() == REQUEST) {
      DatabaseService databaseservice = WebdatabaseImpl.getInstance();
      ENTERROOM response = databaseservice.enterRoom(request);
      return response;
    }
    ENTERROOM error = new ENTERROOM();
    error.setCommand(Command.ENTERROOM);
    error.setHeader(ERROR);
    return error;
  }



  private HISTORYMESSAGE historymessage(HISTORYMESSAGE request) {
    if (REQUEST == request.getHeader()) {
      DatabaseService databaseservice = WebdatabaseImpl.getInstance();
      HISTORYMESSAGE response = databaseservice.fetchHistorymessage(request.getRoom());
      return response;
    }
    HISTORYMESSAGE history = new HISTORYMESSAGE();
    return history;
  }



  private LEAVEPRIVATECHAT leaveprivatechat(Session mySession, LEAVEPRIVATECHAT request) {

    // log.info("mySession=" + mySession.getId()); // Sender=localsession
    // log.info("gegangen=" + request.getGoneSessionid()); // remoteSession -->
    // response muss an diese ID zurückgesendet werden

    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    databaseservice.deletePrivateChatSingle(mySession.getId(), request.getGoneSessionid());

    LEAVEPRIVATECHAT chat = new LEAVEPRIVATECHAT();
    chat.setCommand(Command.LEAVEPRIVATECHAT);
    chat.setHeader(RESPONSE);
    chat.setDataset(Protocol.DATASET);
    chat.setGoneSessionid(mySession.getId());
    chat.setRemotesession(request.getGoneSessionid());

    return chat;
  }



  private ONCALL onCall(ONCALL request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    databaseservice.userOnCall(request.getUserid(), request.isOnCall());

    ONCALL response = new ONCALL();
    response.setCommand(request.getCommand());
    response.setDataset(request.getDataset());
    response.setHeader(RESPONSE);
    response.setOnCall(request.isOnCall());
    response.setUserid(request.getUserid());
    return response;
  }



  /**
   * Die Methode wird aufgerufen, wenn die Server-Client-Verbindung gerade
   * geschlossen wird.
   * 
   * @param session
   *                diese Session wird geschlossen.
   * @param reason
   *                der Grund für den Verbindungsabbruch
   */
  @Override
  public void onClose(Session session, CloseReason reason) {
    log.info("!!!!!!!!!!!!!!!! löschen !!!!!!!!!!!!!!!!!!!!!!!");
    String sessionid = session.getId();
    log.info(sessionid);
    disconnect(sessionid);
  }



  @Override
  public void onError(Session session, Throwable throwable) {
    log.fatal(throwable.getMessage());
    String sessionid = session.getId();
    log.info("sessioID=" + sessionid);
    disconnect(sessionid);
  }



  /**
   * Eine Server-Client-Verbindung wird in diesem Moment hergestellt.
   * 
   * @param session
   *                die gerade geöffnete Session
   * @param config
   *                Websocket Endpoint
   */
  @Override
  public void onOpen(Session session, EndpointConfig config) {
    try {
      lock.lock();
      sessionmap.put(session.getId(), session);
    }
    finally {
      lock.unlock();
    }

    /**
     * Der Server empfängt in diesem Moment eine Botschaft.
     * 
     * @param session
     *                die aktuelle Verbindung
     * @param message
     *                eine Botschaft
     */

    session.addMessageHandler(new MessageHandler.Whole<MESSAGE>() {
      @Override
      public void onMessage(MESSAGE message) {

        if (message.getCommand().equals(Command.TCPAUDIO)) {

        }
        else if (message.getCommand().equals(Command.CONFERENCEAUDIO)) {

        }
        else if (message.getCommand().equals(Command.CONFERENCEVIDEO)) {

        }
        else if (message.getCommand().equals(Command.CONFERENCE)) {
          log.info("Command=" + message.getCommand() + "|" + message.getHeader().toString());
        }
        else {
          log.info("Command=" + message.getCommand() + "|" + message.getHeader());
        }

        if (message instanceof CHATMESSAGE) {
          CHATMESSAGE chatmessage = (CHATMESSAGE) message;
          // Die Nachricht wird noch nicht dauerhaft in der Datenbank abgelegt */

          CHATMESSAGE response = chatmessage(chatmessage);
          if (ERROR == response.getHeader()) {
            log.info("Antwort wird nicht gesendet.");
            return;
          }
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          List<String> sessionlist = databaseservice.fetchSessionlist(chatmessage.getRoom());
          sendMessage(response, sessionlist.toArray(String[]::new));
          return;
        }
        else if (message instanceof TCPAUDIO) {
          TCPAUDIO audio = (TCPAUDIO) message;
          sendMessage(audio, audio.getTarget());
          return;
        }
        else if (message instanceof CONFERENCEAUDIO) {
          CONFERENCEAUDIO conferenceaudio = (CONFERENCEAUDIO) message;
          // der Anwender sendet seinen Stream an alle
          // target sind Empfänger
          sendMessage(conferenceaudio, conferenceaudio.getReceiverSessions());
          return;
        }
        else if (message instanceof CONFERENCEVIDEO) {
          CONFERENCEVIDEO conferenceVideo = (CONFERENCEVIDEO) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          try {
            databaseservice.updateVideocall(
                conferenceVideo.getOrganisator(), conferenceVideo.getKonferenzraum(),
                conferenceVideo.isShared()
            );
            sendMessage(conferenceVideo, conferenceVideo.getReceiverSessions());
          }
          catch (DatabaseException e) {
            log.info(e.getMessage(), e);
          }
          return;
        }
        else if (message instanceof CONFERENCE) {
          CONFERENCE conference = (CONFERENCE) message;
          if (conference.getHeader() == REQUEST) {

            // Lies den Status der Konferenzteilnehmer aus.
            // Wer ist online und wer ist offline

            String konferenzname = conference.getKonferenzname();
            String organisatorUid = conference.getOrganisatorUid();
            String userid = conference.getUserid(); // er hat die Anfrage gestellt
            boolean offhook = conference.isOffhook();
            DatabaseService databaseservice = WebdatabaseImpl.getInstance();
            databaseservice.updateKonferenzraumOnlinestatus(konferenzname, organisatorUid, userid, offhook);
            // in CONFERENCE sind aller Telkomitglieder mit ihrem Status anwesend/abwesend
            // enthalten.
            CONFERENCE response = databaseservice.readKonferenzuser(konferenzname, organisatorUid);
            response.setOffhook(offhook);
            response.setUserid(userid);
            // an alle Sessions eine Antwort senden
            List<String> sessions = databaseservice.fetchKonferenzraumSessions(konferenzname, organisatorUid);
            sendMessage(response, sessions.toArray(String[]::new));
          }
          return;
        }
        else if (message instanceof CONFERENCEMUTE) {
          CONFERENCEMUTE mute = (CONFERENCEMUTE) message;
          switch(mute.getHeader()) {
            case CONFIRM:
              break;
            case ERROR:
              break;
            case REQUEST:
              try {
                DatabaseService databaseservice = WebdatabaseImpl.getInstance();
                databaseservice.updateMute(mute.isMute(), mute.getUserid());
                sendMessage(mute, mute.getReceiverSessions());
              }
              catch (SQLException e) {
                log.error(e.getMessage(), e);
              }
              break;
            case RESPONSE:
              break;
            default:
              break;
          }
          return;
        }
        else if (message instanceof PROYECTORCLOSING) {
          PROYECTORCLOSING projector = (PROYECTORCLOSING) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          switch(projector.getHeader()) {
            case REQUEST:
              // Teilenframe geschlossen
              String callerSession = databaseservice.getSessionidByUserid(projector.getUserid());
              Screencast screencast = databaseservice.getScreencastByCallerSession(callerSession);
              log.info("receiverSession=" + screencast.getReceiverSession());
              log.info("Die Projektorverbindung wurde gekappt.");
              log.info("Der Anwender(Receiver) ist gerade offline gegangen");
              if (screencast.getReceiverSession() == null) return;
              databaseservice.deleteScreencastUser(screencast.getReceiverSession());
              sendMessage(projector, screencast.getReceiverSession());
              return;
            default:
              return;
          }
        }
        else if (message instanceof BEAMEROFF) {
          BEAMEROFF beamer = (BEAMEROFF) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          switch(beamer.getHeader()) {
            case REQUEST:
              // aus der TB_SCREENCAST löschen
              String sessionid = session.getId();
              Screencast screencast = databaseservice.getScreencastByReceiverSession(sessionid);
              // die screencast.getSenderSession kann null sein, weil zuvor ein RESPONSE
              // eingegenagen ist.
              // der Sender hatte den Abbrechen-Knopft betätigt.
              // Später hat dann der Empfänger den Projektorframe geschlossen.
              // die sessionid ist in der Tabelle nicht mehr enthalten.
              if (screencast.getSenderSession() == null) return;
              databaseservice.deleteScreencastUser(screencast.getReceiverSession());
              // weiterleiten an den Sender
              sendMessage(beamer, screencast.getSenderSession());
              break;
            case RESPONSE:
              break;
            case CONFIRM:
              break;
            case ERROR:
              break;
            default:
              break;
          }
          return;
        }
        else if (message instanceof IMAGE) {
          IMAGE image = (IMAGE) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          Session item;
          switch(image.getHeader()) {
            case REQUEST:
              // an den Empfänger weiterleiten
              String receiverSessionId = databaseservice.getSessionidByNickname(image.getReceiverNickname());
              if (receiverSessionId == null) {
                // Der Empfänger 'Nickname' ist verschwunden.
                IMAGE error = new IMAGE();
                error.setCommand(Command.IMAGE);
                error.setHeader(ERROR);
                error.setDataset(Protocol.DATASET);
                error.setSenderUid(image.getSenderUid());
                error.setSenderNickname(image.getSenderNickname());
                error.setReceiverNickname(image.getReceiverNickname());
                error.setMultilingualkey(KEY.SERVER_DER_EMPFAENGER_IST_VERSCHWUNDEN);
                sendMessage(error, session);
                return;
              }
              item = sessionmap.get(receiverSessionId);
              if (item == null) {
                IMAGE error = new IMAGE();
                error.setCommand(Command.IMAGE);
                error.setHeader(ERROR);
                error.setDataset(Protocol.DATASET);
                error.setSenderUid(image.getSenderUid());
                error.setSenderNickname(image.getSenderNickname());
                error.setReceiverNickname(image.getReceiverNickname());
                error.setMultilingualkey(KEY.SERVER_DER_EMPFAENGER_IST_VERSCHWUNDEN);
                sendMessage(error, session);
                return;
              }
              // 2. Nimmt der Empfänger Übertragungen an?
              if (!databaseservice.isOnrekord(receiverSessionId)) {
                IMAGE error = new IMAGE();
                error.setCommand(Command.IMAGE);
                error.setHeader(ERROR);
                error.setDataset(Protocol.DATASET);
                error.setSenderUid(image.getSenderUid());
                error.setSenderNickname(image.getSenderNickname());
                error.setReceiverNickname(image.getReceiverNickname());
                error.setMultilingualkey(KEY.SERVER_LEHNT_UEBERTRAGUNGEN_AB);
                sendMessage(error, session);
                return;
              }
              // 3. Empfänger muss noch seinen Projektor aussschalten, weil er in einer
              // Sitzung ist.
              if (databaseservice.isOnProjektor(receiverSessionId)) {
                IMAGE error = new IMAGE();
                error.setCommand(Command.IMAGE);
                error.setHeader(ERROR);
                error.setDataset(Protocol.DATASET);
                error.setSenderUid(image.getSenderUid());
                error.setSenderNickname(image.getSenderNickname());
                error.setReceiverNickname(image.getReceiverNickname());
                error.setMultilingualkey(KEY.SERVER_MUSS_NOCH_SEINEN_PROJEKTOR_AUSSCHALTEN);
                sendMessage(error, session);
                return;
              }
              // 4. Anfrage wird an den Empfänger weitergeleitet
              sendMessage(image, receiverSessionId);
              break;
            case RESPONSE:
              String senderSessionId = databaseservice.getSessionidByUserid(image.getSenderUid());
              databaseservice.setProjektorSession(session.getId(), senderSessionId);
              sendMessage(image, senderSessionId);
              break;
            case ERROR:
              break;
            case CONFIRM:
              break;
          }
          return;
        }
        else if (message instanceof SCREENCAST) {
          SCREENCAST screencast = (SCREENCAST) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          switch(screencast.getHeader()) {
            case CONFIRM:
              break;
            case ERROR:
              break;
            case REQUEST:
              databaseservice.updateRekorderport(screencast.getCallerUserid(), screencast.getCallerPort());
              String receiverSessionid = databaseservice
                  .getSessionidByNickname(screencast.getReceiverNickname());
              // 1. Ist der Empfänger noch online?
              if (receiverSessionid == null) {
                // Der Empfänger ist nicht mehr online
                // Antwort mit Error senden
                SCREENCAST response = new SCREENCAST();
                response.setCommand(Command.SCREENCAST);
                response.setHeader(ERROR);
                response.setDataset(Protocol.DATASET);
                response.setReceiverNickname(screencast.getReceiverNickname());
                response.setErrorMessage(
                    "<html><b>" + screencast.getReceiverNickname() + "</b> ist nicht mehr online.</html>"
                );
                try {
                  session.getAsyncRemote().sendObject(response);
                }
                catch (Exception e) {
                  log.error(e.getMessage(), e.getCause());
                }
                return;
              }

              // ich muss die
              // ic habe nur den ReceiverNickname
              if (databaseservice.isOnProjektor(receiverSessionid)) {
                log.info("Projektor ist eingeschaltet, Error zurücksenden");
                SCREENCAST response = new SCREENCAST();
                response.setCommand(Command.SCREENCAST);
                response.setHeader(ERROR);
                response.setDataset(Protocol.DATASET);
                response.setReceiverNickname(screencast.getReceiverNickname());
                response.setErrorMessage("<html>Der Projektor ist in Betrieb.</html>");
                try {
                  session.getAsyncRemote().sendObject(response);
                }
                catch (Exception e) {
                  log.error(e.getMessage(), e.getCause());
                }
                return;
              }

              // 2. Nimmt der Empfänger Übertragungen an?

              if (!databaseservice.isOnrekord(receiverSessionid)) {
                SCREENCAST response = new SCREENCAST();
                response.setCommand(Command.SCREENCAST);
                response.setHeader(ERROR);
                response.setDataset(Protocol.DATASET);
                response.setReceiverNickname(screencast.getReceiverNickname());
                response.setErrorMessage(
                    "<html><b>" + screencast.getReceiverNickname()
                        + "</b> nimmt keine Bildschirmübertragungen entgegen.</html>"
                );
                try {
                  session.getAsyncRemote().sendObject(response);
                }
                catch (Exception e) {
                  log.error(e.getMessage(), e.getCause());
                }
                return;
              }
              // weiterleiten an den Empfänger, damit er ein Fenster öffnet
              sendMessage(screencast, receiverSessionid);
              break;
            case RESPONSE:
              // Antwort weiterleiten an den Sender
              String callerSession = databaseservice.getSessionidByUserid(screencast.getCallerUserid());
              // eine Projektorverbindung wird eingetragen
              databaseservice.setProjektorSession(session.getId(), callerSession);
              sendMessage(screencast, callerSession);
              break;
            default:
              break;
          }
          return;
        }
        else if (message instanceof BREAKROOMS) {
          BREAKROOMS breakrooms = (BREAKROOMS) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          List<String> rooms = databaseservice.fetchBreakrooms(breakrooms.getUserid());
          BREAKROOMS response = new BREAKROOMS();
          response.setCommand(Command.BREAKROOMS);
          response.setHeader(RESPONSE);
          response.setDataset(Protocol.DATASET);
          response.setUserid(breakrooms.getUserid());
          response.setRooms(rooms.toArray(new String[rooms.size()]));
          sendMessage(response, session);
          return;
        }
        else if (message instanceof READROOMS) {
          READROOMS readrooms = (READROOMS) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          List<String> rooms = databaseservice.fetchRooms(readrooms.getUserid(), readrooms.getRoomtype());
          READROOMS response = new READROOMS();
          response.setCommand(Command.READROOMS);
          response.setHeader(RESPONSE);
          response.setDataset(Protocol.DATASET);
          response.setUserid(readrooms.getUserid());
          response.setRoomtype(readrooms.getRoomtype());
          response.setRooms(rooms.toArray(new String[rooms.size()]));
          sendMessage(response, session);
          return;
        }
        else if (message instanceof UPDATEFILETYPES) {
          UPDATEFILETYPES types = (UPDATEFILETYPES) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          databaseservice.updateUserBlacklist(types.getUserid(), types.getFiletypes());
          // Confirm senden
          UPDATEFILETYPES response = new UPDATEFILETYPES();
          response.setCommand(Command.UPDATEFILETYPES);
          response.setHeader(CONFIRM);
          response.setDataset(Protocol.DATASET);
          response.setUserid(types.getUserid());
          sendMessage(response, session);
          return;
        }
        else if (message instanceof FILETYPES) {
          FILETYPES filetypes = (FILETYPES) message;
          if (REQUEST != filetypes.getHeader()) return;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          databaseservice.createUserBlacklist(filetypes.getUserid());
          List<BlacklistTypes> sperrliste = databaseservice.fetchFiletypes(filetypes.getUserid());
          FILETYPES response = new FILETYPES();
          response.setCommand(Command.FILETYPES);
          response.setHeader(RESPONSE);
          response.setDataset(Protocol.DATASET);
          response.setUserid(filetypes.getUserid());
          response.setFiletypes(sperrliste.toArray(new BlacklistTypes[sperrliste.size()]));
          sendMessage(response, session);
          return;
        }
        else if (message instanceof SEARCHFILES) {
          // jetzt überträgt er alles
          SEARCHFILES searchfiles = (SEARCHFILES) message;
          List<Searchfile> foundFiles = searchfiles(searchfiles);

          final int blocksize = 64;
          int maxRound = foundFiles.size() / blocksize;
          ArrayList<Searchfile> part = new ArrayList<>();
          int round = 0;
          while (round <= maxRound) {
            part.clear();
            int m = 0;
            while (!foundFiles.isEmpty()) {
              if (m > blocksize) break;
              part.add(foundFiles.remove(0));
              m++;
            }
            // part senden // bestimmte Dateien filtern
            // part enthält bis 64 Dateien
            SEARCHFILES response = new SEARCHFILES();
            response.setCommand(Command.SEARCHFILES);
            response.setHeader(RESPONSE);
            response.setDataset(Protocol.DATASET);
            response.setSearchfiles(part.toArray(new Searchfile[part.size()]));
            response.setStarted(round == 0 ? true : false);
            response.setUserid(searchfiles.getUserid());
            sendMessage(response, session);
            round++;
          }
          return;
        }
        else if (message instanceof DOWNLOAD) {
          DOWNLOAD download = (DOWNLOAD) message;
          DOWNLOAD response = selectDOWNLOAD(download);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof UPLOADFILES) {
          UPLOADFILES uploadfiles = (UPLOADFILES) message;
          UPLOADFILES response = uploadfiles(uploadfiles);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof DELETEUPLOADFILES) {
          DELETEUPLOADFILES deletefiles = (DELETEUPLOADFILES) message;
          DELETEUPLOADFILES response = deleteuploadfiles(deletefiles);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof PRIVATECHATFILE) {
          PRIVATECHATFILE privatechatfile = (PRIVATECHATFILE) message;
          if (REQUEST == privatechatfile.getHeader()) {
            DatabaseService databaseservice = WebdatabaseImpl.getInstance();
            String receiver_uid = databaseservice.getUserBySession(privatechatfile.getRemoteSessionId());
            try {
              PrivateChatfile entity = databaseservice.uploadPrivateChatfileAttributes(
                  privatechatfile.getUserid(), receiver_uid, privatechatfile.getAbsoluteFilename(),
                  privatechatfile.getFilename(), privatechatfile.getFilesize(), privatechatfile.getMessage(),
                  privatechatfile.getRemoteNickname()
              );
              PRIVATECHATFILE response = new PRIVATECHATFILE();
              response.setCommand(Command.PRIVATECHATFILE);
              response.setHeader(RESPONSE);
              response.setDataset(Protocol.DATASET);
              response.setFilename(entity.getFilename());
              response.setAbsoluteFilename(privatechatfile.getAbsoluteFilename());
              response.setFilesize(entity.getFilesize());
              response.setMessage(entity.getMessage());
              response.setNumber(entity.getAnlage());
              response.setRemoteNickname(entity.getRemoteNickname());
              response.setRemoteSessionId(privatechatfile.getRemoteSessionId());
              response.setUserid(entity.getUid());
              sendMessage(response, session);
            }
            catch (DatabaseException e) {
              log.info(e.getMessage(), e);
            }
          }
          return;
        }
        else if (message instanceof PRIVATEMESSAGE) {
          PRIVATEMESSAGE privatemessage = (PRIVATEMESSAGE) message;
          if (REQUEST == privatemessage.getHeader()) {
            privatemessage.setDatetime(System.currentTimeMillis());
            privatemessage.setLocalSessionid(session.getId());
            sendMessage(privatemessage, privatemessage.getRemoteSessionid());
          }
          else if (RESPONSE == privatemessage.getHeader()) {
            sendMessage(privatemessage, privatemessage.getLocalSessionid());
          }
          return;
        }

        else if (message instanceof CHATUSERLIST) {
          // CHATUSERLIST response = chatuserlist((CHATUSERLIST)message);
          // Kann noch verbessert werden. Nur an die User senden, die
          // mit iherer SessionID im Raum sind
          // Aktuell sende ich an alle
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          List<String> userInRooms = databaseservice.fetchRoomlist(session.getId());
          // alle infos werden für den User aktualisiert
          broadcast(userInRooms);
          return;
        }
        else if (message instanceof HISTORYMESSAGE) {
          HISTORYMESSAGE response = historymessage((HISTORYMESSAGE) message);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof ROOMLIST) {
          if (message instanceof READTOPICROOMOWNER) return; // READTOPICROOMOWNER erweitert ROOMLIST
          // Eine Liste aller Chaträume wird angefordert
          // Der Server broadcastet eine Raumbelegung
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          List<String> roomlist = databaseservice.fetchRoomlist(session.getId());

          log.info("eine Raumliste wird angefordert=" + roomlist.size());

          broadcast(roomlist);
          return;
        }
        else if (message instanceof READTOPICROOMOWNER) {
          READTOPICROOMOWNER response;
          try {
            response = readTopicroom((READTOPICROOMOWNER) message);
            session.getAsyncRemote().sendObject(response);
          }
          catch (Exception e) {
            log.error(e.getMessage(), e.getCause());
          }
          return;
        }
        else if (message instanceof LEAVEROOM) {
          LEAVEROOM leaveroom = (LEAVEROOM) message;
          // Nach LEAVEROMM ist der User mit der SESSION_ID nicht mehr vorhanden */
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          // Der User mit der session.getID ist schon nicht mehr in dem Raum
          // Auch nicht so effizient, weil alle Räume ermittelt werden
          List<String> roomlist = databaseservice.fetchRoomlist(session.getId());

          // Die Raumliste muss vielleicht gekürzt werden, wenn der Raum
          // ein Tempraum war und leer ist

          // Ist der Chat ein Pausenraum oder ein Gruppenraum?

          try {
            switch(databaseservice.getLifetime(leaveroom.getRoom())) {
              case PERMANENT:
                databaseservice.remove(leaveroom.getUserid(), leaveroom.getRoom());
                break;
              case TEMPORARY:
                databaseservice.remove(leaveroom.getUserid(), leaveroom.getRoom());
                int count = databaseservice.countTemproomUser(leaveroom.getRoom());
                if (count == 0) {
                  // Temproom löschen
                  // und raumliste kürzen
                  // log.info(response.getRoom());
                  roomlist.remove(leaveroom.getRoom());
                  // Beim Löschen eines Temprooms, spielt die Userid keine Rolle.
                  // Der Raum muss gelöscht werden, wenn niemand anwesend ist
                  databaseservice.deleteTemproom(leaveroom.getRoom());
                }
                break;
              default:
                break;
            }

            // log.info("muss wieder angepasst werden !!!!!!!!!!!!!!!!!!!!!!");
            // Kann noch verbessert werden. Nur an die Senden, die
            // mit iherer SessionID im Raum sind
            // Aktuell sende ich an alle
            broadcast(roomlist);
          }
          catch (DatabaseException e) {
            log.fatal(e.getMessage());
          }
          return;
        }
        else if (message instanceof ENTERROOM) {
          ENTERROOM response = enterroom((ENTERROOM) message);
          try {
            session.getAsyncRemote().sendObject(response);
          }
          catch (Exception e) {
            log.error(e.getMessage(), e.getCause());
          }
          return;
        }
        else if (message instanceof CREATETEMPROOM) {
          CREATETEMPROOM temproom = (CREATETEMPROOM) message;
          CREATETEMPROOM response = createTemproom(temproom);

          try {
            session.getAsyncRemote().sendObject(response);
          }
          catch (Exception e) {
            log.error(e.getMessage(), e.getCause());
          }
          return;
        }
        else if (message instanceof LEAVEPRIVATECHAT) {
          // REQUEST kommt rein
          LEAVEPRIVATECHAT leavechat = (LEAVEPRIVATECHAT) message;
          LEAVEPRIVATECHAT response = leaveprivatechat(session, leavechat);

          // log.info("sessionid=" + session.getId());
          // log.info("responseid=" + response.getRemotesession());
          sendMessage(response, response.getRemotesession());
          return;
        }
        else if (message instanceof CALLREMOTEUSER) {
          CALLREMOTEUSER remoteuser = (CALLREMOTEUSER) message;

          if (CONFIRM == remoteuser.getHeader()) {
            // in die Datenbank eintragen, wenn möglich
            /* log.info("localNickname=" + remoteuser.getLocalNickname());
             * log.info("localSessionid=" + remoteuser.getLocalSessionid());
             * log.info("remoteNickname=" + remoteuser.getRemoteNickname());
             * log.info("remoteSessionid=" + remoteuser.getRemoteSessionid()); */
            DatabaseService databaseservice = WebdatabaseImpl.getInstance();
            if (
              !databaseservice
                  .insertPrivateChat(remoteuser.getRemoteSessionid(), remoteuser.getLocalSessionid())
            ) {
              // Verbindung steht schon
              return;
            }
          }
          else if (RESPONSE == remoteuser.getHeader()) {
            DatabaseService databaseservice = WebdatabaseImpl.getInstance();
            databaseservice
                .deletePrivateChatSingle(remoteuser.getLocalSessionid(), remoteuser.getRemoteSessionid());
          }
          else {
            log.error("unbekannter HEADER, " + remoteuser.getHeader());
          }
          // hier ist bereits der OK Fall
          CALLPRIVATECHAT callprivatechat = callPrivatechat(remoteuser);
          // an die LocalSessionID senden
          sendMessage(callprivatechat, callprivatechat.getLocalSessionid());
          return;
        }
        else if (message instanceof CALLPRIVATECHAT) {
          CALLPRIVATECHAT callprivate = (CALLPRIVATECHAT) message;
          log.info(callprivate.toString());
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          CALLREMOTEUSER remoteuser = databaseservice.callRemoteuser(callprivate, session.getId());
          if (ERROR == remoteuser.getHeader()) {
            sendMessage(remoteuser, session);
            return;
          }

          if (log.isInfoEnabled()) {
            /* log.info("localNickname=" + remoteuser.getLocalNickname());
             * log.info("localSessionid=" + remoteuser.getLocalSessionid());
             * log.info("remoteNickname=" + remoteuser.getRemoteNickname());
             * log.info("remoteSessionid=" + remoteuser.getRemoteSessionid()); */
          }

          // log.info("remoteUserFehlt==fehler");

          // der 1:1 Chat oder PrivateChat wird durch CALLPRIVATECHAT eingeleitet
          // localSessionID und remoteSessionid müssen verschieden sein

          if (remoteuser.getLocalSessionid().equals(remoteuser.getRemoteSessionid())) {
            // Sender und Empfänger sind identisch
            CALLPRIVATECHAT call = new CALLPRIVATECHAT();
            call.setCommand(Command.CALLPRIVATECHAT);
            call.setHeader(ERROR);
            call.setDataset(Protocol.DATASET);
            call.setLocalNickname(remoteuser.getRemoteNickname());
            call.setLocalSessionid(remoteuser.getRemoteSessionid());
            call.setRemoteNickname(remoteuser.getLocalNickname());
            call.setRemoteSessionid(remoteuser.getLocalSessionid());
            call.setMultilingualkey(KEY.STRING_EINSAME_SEELEN);
            sendMessage(call, remoteuser.getLocalSessionid());
            return;
          }

          // private chat eintragen
          // remote SessionID und LocalSessionID

          // remote und local sind vertauscht!
          // der Anrufer trägt sich ein
          if (
            !databaseservice
                .insertPrivateChat(remoteuser.getRemoteSessionid(), remoteuser.getLocalSessionid())
          ) {
            // Verbindung steht schon
            CALLPRIVATECHAT call = new CALLPRIVATECHAT();
            call.setCommand(Command.CALLPRIVATECHAT);
            call.setHeader(ERROR);
            call.setDataset(Protocol.DATASET);
            call.setLocalNickname(remoteuser.getLocalNickname());
            call.setLocalSessionid(remoteuser.getLocalSessionid());
            call.setRemoteNickname(remoteuser.getRemoteNickname());
            call.setRemoteSessionid(remoteuser.getRemoteSessionid());
            call.setErrorMessage("Der Chat läuft bereits.");
            call.setMultilingualkey(KEY.SERVER_DER_CHAT_LAEUFT_BEREITS);
            sendMessage(call, remoteuser.getRemoteSessionid());
            return;
          }
          // sende an local session id
          sendMessage(remoteuser, remoteuser.getLocalSessionid());
          return;
        }
        else if (message instanceof USERONLINELIST) {
          USERONLINELIST response = useronlinelist((USERONLINELIST) message);
          broadcastAll(response);
          return;
        }
        else if (message instanceof DIAL) {
          DIAL request = (DIAL) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          DIAL dialResponse = databaseservice.dial(request);
          // War der RESPONSE positiv?
          // Falls ja
          // andere Seite benachrichtigen
          if (ERROR == dialResponse.getHeader()) {
            sendMessage(dialResponse, session);
            return;
          }
          else if (RESPONSE == dialResponse.getHeader()) {
            // Nachricht an dem Empfänger schicken, dass gleich eine Nachricht kommt
            // ich benötige die SessionID, damit die Nachricht einen Adressaten findet
            String sessionid = databaseservice.getSessionidByUserid(dialResponse.getReceiverUserid());
            if (sessionid == null) {
              // der Anrufempfänger ist nicht mehr online
              DIAL dial = new DIAL();
              dial.setHeader(ERROR);
              dial.setCommand(Command.DIAL);
              dial.setDataset(Protocol.DATASET);
              dial.setReceiverNickname(dialResponse.getReceiverNickname());
              dial.setMultilingualkey(KEY.STRING_IST_NICHT_MEHR_ONLINE);
              sendMessage(dial, session);
              return;
            }
            // RESPONSE ist eine positive Antwort
            // SESSIONID wird für TCP-Übertragung optional hinzugepackt.
            dialResponse.setReceiverSession(sessionid); // die Sessionid vom Empfänger
            dialResponse.setCallerSession(session.getId()); // die Sessionid vom Caller
            sendMessage(dialResponse, session);
          }
          return;
        }
        else if (message instanceof INCOMINGCALL) {
          INCOMINGCALL incoming = (INCOMINGCALL) message;
          if (RESPONSE == incoming.getHeader()) {
            DatabaseService databaseservice = WebdatabaseImpl.getInstance();
            final String sessionid = databaseservice.getSessionidByUserid(incoming.getCallerUserid());
            if (sessionid == null) {
              // exakte Fehlermeldung zurück an den Absender, Caller ist nicht mehr online
              // Protocoll anpassen error auch für CALLER_NICKNAME? oder RECEIVER_NICKNAME
              // komplett streichen
              // Fehler simulieren
              // Emepfänger hat den Hörer noch nicht abgenommen
              // und der Anrufer legt auf
              INCOMINGCALL error = new INCOMINGCALL();
              error.setCommand(Command.INCOMINGCALL);
              error.setHeader(ERROR);
              error.setDataset(Protocol.DATASET);
              error.setReceiverNickname(incoming.getCallerNickname()); // receiver und caller vertauschen
                                                                       // wegen response oder receiverNickname
                                                                       // optional
              error.setMultilingualkey(KEY.STRING_IST_NICHT_MEHR_ONLINE);
              // error senden
              sendMessage(error, session);
              return;
            }
            // response ist der CALLER telefonisch noch erreichbar?

            if (!databaseservice.isOnBusyByUid(incoming.getCallerUserid())) {
              // Gegenseite hat aufgelegt
              INCOMINGCALL error = new INCOMINGCALL();
              error.setCommand(Command.INCOMINGCALL);
              error.setHeader(ERROR);
              error.setDataset(Protocol.DATASET);
              error.setReceiverNickname(incoming.getCallerNickname()); // receiver und caller vertauschen
                                                                       // wegen response oder receiverNickname
                                                                       // optional
              error.setMultilingualkey(KEY.STRING_HAT_AUFGELEGT);
              // error senden
              sendMessage(error, session);
              return;
            }
            // an den Receiver CONFIRM senden, damit er den Lautsprecher starten kann
            INCOMINGCALL confirm = new INCOMINGCALL();
            confirm.setCommand(Command.INCOMINGCALL);
            confirm.setHeader(CONFIRM);
            confirm.setDataset(Protocol.DATASET);
            confirm.setCallerIp(incoming.getCallerIp());
            confirm.setCallerNickname(incoming.getCallerNickname());
            confirm.setCallerPort(incoming.getCallerPort());
            confirm.setCallerUserid(incoming.getCallerUserid());
            confirm.setCallerVoice(incoming.getCallerVoice());
            confirm.setCallerSession(sessionid);
            confirm.setReceiverIp(incoming.getReceiverIp());
            confirm.setReceiverNickname(incoming.getReceiverNickname());
            confirm.setReceiverPort(incoming.getReceiverPort());
            confirm.setReceiverUserid(incoming.getReceiverUserid());
            confirm.setReceiverSession(incoming.getReceiverSession());
            sendMessage(confirm, session);
            // RESPONSE wird an den CALLER weitergereicht

            databaseservice.setOnBusy(incoming.getReceiverUserid(), true);
            sendMessage(incoming, sessionid);
            // !!!!!!!!!!!!!!!! RESPONSE wird an den CALLER durchgereicht !!!!!!!!!!
            // Beide Seiten als telefonierend eintragen
            databaseservice.insertToPhone(incoming.getCallerUserid(), incoming.getReceiverUserid());
            return;
          }
          else if (REQUEST == incoming.getHeader()) {
            // RECEIVER_SESSIONID habe ich aus incoming
            DatabaseService databaseservice = WebdatabaseImpl.getInstance();
            final String sessionid = databaseservice.getSessionidByUserid(incoming.getReceiverUserid());
            if (sessionid == null) {
              INCOMINGCALL error = new INCOMINGCALL();
              error.setCommand(Command.INCOMINGCALL);
              error.setHeader(ERROR);
              error.setDataset(Protocol.DATASET);
              error.setReceiverNickname(incoming.getReceiverNickname());
              error.setMultilingualkey(KEY.STRING_IST_NICHT_MEHR_ONLINE);
              sendMessage(error, session);
              return;
            }
            // Der Anwender telefoniert
            databaseservice.setOnBusy(incoming.getReceiverUserid(), true);
            // über den HEADER_RESPONSE sagt er, ich telefoniere
            sendMessage(incoming, sessionid);
            return;
          }
          return;
        }
        else if (message instanceof MICROERROR) {
          MICROERROR microerror = (MICROERROR) message;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          final String sessionid = databaseservice.getSessionidByUserid(microerror.getToUserid());
          if (sessionid == null) return;
          sendMessage(microerror, sessionid);
          return;
        }
        else if (message instanceof ONCALL) {
          ONCALL response = onCall((ONCALL) message);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof ONHOOK) {
          ONHOOK onhook = (ONHOOK) message;
          final String sessionid;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          switch(onhook.getHeader()) {
            case REQUEST:
              if (databaseservice.findPartnerUserid(onhook.getCallerUserid()) == null) {
//                Clientseite
//                1. Anrufen
//                2. Abnehmen
//                3. Wählen
//                4. Auflegen

                // An den Receiver trotzdem durchstellen, aber nicht klingeln
                onhook.setSignal(false);
                if (onhook.isTo()) {
                  // Caller hat aufgelegt an Receiver senden
                  sessionid = databaseservice.getSessionidByUserid(onhook.getReceiverUserid());
                }
                else {
                  // Receiver hat aufgeleg an Caller senden
                  sessionid = databaseservice.getSessionidByUserid(onhook.getCallerUserid());
                }
                if (sessionid == null) return;
                sendMessage(onhook, sessionid);
                return;
              }
              if (onhook.isTo()) {
                // Caller hat aufgelegt
//                log.info("der Caller hat aufgelegt");
                // Pärchen aus der Telefonierentabelle löschen
                databaseservice.deletePartner(onhook.getCallerUserid()); // TB_TOPHONE

                databaseservice.setOnBusy(onhook.getCallerUserid(), false); // TB_PHONE
                sessionid = databaseservice.getSessionidByUserid(onhook.getReceiverUserid());

                // Telefoniert der Receiver?
                if (!databaseservice.isOnBusyByUid(onhook.getReceiverUserid())) { // TB_PHONE
                  // TODO RECEIVER ist nicht am telefonieren
                  // ONHOOK muss trotzdem durchgestellt werden
                  // SIGNAL auf 1
                  log.info(onhook.toString());
                  return;
                }

              }
              else {

                // Receiver hat aufgelegt
//                log.info("der receiver hat aufgelegt");

                // Pärchen aus der Telefonierentabelle löschen
                databaseservice.deletePartner(onhook.getReceiverUserid());

                databaseservice.setOnBusy(onhook.getReceiverUserid(), false);
                sessionid = databaseservice.getSessionidByUserid(onhook.getCallerUserid());

                // Telefoniert der Caller?
                if (!databaseservice.isOnBusyByUid(onhook.getCallerUserid())) {
                  log.info("444444444444444444444"); // OK wird aufgerufen
                  return;
                }

              }
              if (sessionid == null) return;
              sendMessage(onhook, sessionid);
              log.info("ONHOOK - gesendet");
              break;
            default:
              break;
          }
          return;
        }
        else if (message instanceof UPDATEPHONE) {
          UPDATEPHONE response = updatePhone((UPDATEPHONE) message); // TB_PHONE
          sendMessage(response, session);
          return;
        }
        else if (message instanceof UPDATEPORT) {
          UPDATEPORT response = updateport((UPDATEPORT) message);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof USRLOGIN) {
          USRLOGIN response = userlogin((USRLOGIN) message, session);
          sendMessage(response, session);
          if (response.getHeader() == ERROR) return;
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          USERONLINELIST userOnlinelist = databaseservice.fetchUseronlinelist(response.getUserid());
          broadcastAll(userOnlinelist);
          try {
            databaseservice.insertScreenshot(response.getUserid());
          }
          catch (DatabaseException e) {
            log.error(e.getMessage());
          }
          return;
        }
        else if (message instanceof CHANGETOPICMEMBER) {
          CHANGETOPICMEMBER member = (CHANGETOPICMEMBER) message;
          CHANGETOPICMEMBER response = changeTopicmember(member);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof CANDIDATETOPICMEMBER) {
          CANDIDATETOPICMEMBER member = (CANDIDATETOPICMEMBER) message;
          CANDIDATETOPICMEMBER response = candidateMember(member);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof TOPICMEMBER) {
          TOPICMEMBER member = (TOPICMEMBER) message;
          TOPICMEMBER response = topicMember(member);
          sendMessage(response, session);
          return;
        }
        else if (message instanceof DELETEROOM) {
          DELETEROOM member = (DELETEROOM) message;
          DELETEROOM response = deleteRoom(member);
          sendMessage(response, session);
          DatabaseService databaseservice = WebdatabaseImpl.getInstance();
          List<String> roomlist = databaseservice.fetchRoomlist(session.getId());
          broadcast(roomlist);
          return;
        }
        else if (message instanceof UPDATEUSER) {
          UPDATEUSER response = updateuser((UPDATEUSER) message);
          sendMessage(response, session);
          // nickname chaträume anpassen
          if (CONFIRM == response.getHeader()) {
            DatabaseService databaseservice = WebdatabaseImpl.getInstance();
            boolean done = databaseservice.updateOWNER(response.getUserid(), response.getNickname());
            if (done) {
              // broadcast roomlist
              List<String> roomlist = databaseservice.fetchRoomlist(session.getId());
              broadcast(roomlist);
            }
          }
          return;
        }
        MessageEncoder encoder = new MessageEncoder();
        try {
          String unknown = encoder.encode(message);
          log.warn(unknown);
        }
        catch (EncodeException e) {
          log.error(e.getMessage(), e.getCause());
        }
        log.error("Verbindung schließen wegen unbekanntem Befehl");
      }
    });
  }



  private READTOPICROOMOWNER readTopicroom(READTOPICROOMOWNER request) throws DatabaseException {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    READTOPICROOMOWNER response = databaseservice.fetchTopicroom(request);
    return response;
  }



  /**
   * Löscht den User aus allen Onlinelisten.
   * 
   * @param sessionID
   *                  der Onlineschlüssel identifiziert einen Anwender
   */
  private void removeUser(String sessionID) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    String uid = databaseservice.logoutUser(sessionID);

    // War der der der letzte in einem Temproom
    // Durchsuche alle Temprooms, ob sie leer sind
    // Wenn ja, lösche den Temproom und alle Messages

    // select all TEMPROOMS
    try {
      List<String> temproom = databaseservice.fetchTemproomlist();
      for (String room : temproom) {
        int countUser = databaseservice.countTemproomUser(room);
        // deleteChatroom
        if (countUser == 0) {
          databaseservice.deleteTemproom(room);
        }
        // User wird aus einem Pausenraum/Group herausgenommen
        databaseservice.remove(uid, room);
      }

      // die privatechats müssen auch gelöscht werden!
      // die Gegenseite muss benachrichtigt werden
      // input sessionID dieser User ist gegangen
      List<String> counterIds = databaseservice.deletePrivateChatAll(sessionID);
      LEAVEPRIVATECHAT leavePrivatechat = new LEAVEPRIVATECHAT();
      leavePrivatechat.setCommand(Command.LEAVEPRIVATECHAT);
      leavePrivatechat.setHeader(REQUEST);
      leavePrivatechat.setDataset(Protocol.DATASET);
      leavePrivatechat.setGoneSessionid(sessionID);
      sendMessage(leavePrivatechat, counterIds.toArray(String[]::new));

//      Telefonate
      // Hole den Partner von uid
      String partneruid = databaseservice.findPartnerUserid(uid);
      if (partneruid != null) {
        databaseservice.setOnBusy(partneruid, false);
        log.debug("partneruid benachrichtigen, dass die Verbindung weg ist");

        // uid hat aufgelegt
        // uid!=partneruid
        boolean isCaller = databaseservice.isPartnerCaller(partneruid); // true partneruid ist calleruid
        ONHOOK onhook = new ONHOOK();
        onhook.setCommand(Command.ONHOOK);
        onhook.setHeader(REQUEST);
        onhook.setDataset(Protocol.DATASET);
        onhook.setCallerUserid(isCaller ? partneruid : uid); // Wer ist Anrufer?
        onhook.setReceiverUserid(isCaller ? uid : partneruid); // Wer ist Empfänger
        onhook.setTo(isCaller ? false : true); // Wer hat aufgelegt? // true der Caller hat aufgelegt
        onhook.setSignal(true);
        String key = databaseservice.getSessionidByUserid(partneruid);
        sendMessage(onhook, key);
      }

      databaseservice.deleteOnline(uid); // Lösche user aus der TB_ONLINE
      databaseservice.deletePhoneUser(uid); // Lösche aus TB_PHONE
      databaseservice.deletePartner(uid); // TB_TOPHONE
      // Der Gegenüber müsste benachrichtigt werden, dass ein plötzlicher Ausfall
      // vorliegt. Die Gegenseite muss benachrichtig werden, dass der Gesprächspartner
      // nicht mehr online ist

      // TODO benachrichtige alle in der Telefonkonferenz; reagiere wie bei Telefon
      // auflegen

      databaseservice.deleteScreencastUser(sessionID);
      databaseservice.deleteScreenshots(uid);
      databaseservice.unmute(uid);

      USERONLINELIST userOnlinelist = databaseservice.fetchUseronlinelist(uid);
      broadcastAll(userOnlinelist);

    }
    catch (DatabaseException e) {
      log.error(e.getMessage());
    }

    // Alle Einträge in transfer_lib vom User löschen
    databaseservice.deleteUserChatfiles(uid);
  }



  // TODO aktuell werden maximal 256 zurückgegeben.
  private List<Searchfile> searchfiles(SEARCHFILES request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    final int max_files = 256;
    List<Searchfile> foundFiles = databaseservice.selectSearchfiles(request.getPatterns(), max_files);
    return foundFiles;
  }



  private DOWNLOAD selectDOWNLOAD(DOWNLOAD request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    DOWNLOAD response = databaseservice.selectDOWNLOAD(request);
    return response;
  }



  private TOPICMEMBER topicMember(TOPICMEMBER request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    TOPICMEMBER response = databaseservice.fetchTopicmember(request);
    return response;
  }



  private UPDATEPHONE updatePhone(UPDATEPHONE request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    databaseservice.setOnBusy(request.getUserid(), request.isOnBusy());
    UPDATEPHONE response = new UPDATEPHONE();
    response.setCommand(Command.UPDATEPHONE);
    response.setHeader(RESPONSE);
    response.setDataset(Protocol.DATASET);
    response.setUserid(request.getUserid());
    response.setOnBusy(request.isOnBusy());
    return response;
  }



  private UPDATEPORT updateport(UPDATEPORT request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    databaseservice.updateUpdateport(request.getIp(), request.getPort());
    UPDATEPORT response = new UPDATEPORT();
    response.setCommand(Command.UPDATEPORT);
    response.setHeader(RESPONSE);
    response.setDataset(Protocol.DATASET);
    response.setUserid(request.getUserid());
    response.setIp(request.getIp());
    response.setPort(request.getPort());
    return response;
  }



  private UPDATEUSER updateuser(UPDATEUSER request) {
    if (request.getHeader() == REQUEST) {
      if (request.getDataset().equals(Protocol.DATASET)) {
        DatabaseService databaseservice = WebdatabaseImpl.getInstance();
        EntityUser user = databaseservice.updateUser(request);
        if (user != null) {
          UPDATEUSER response = new UPDATEUSER();
          response.setCommand(Command.UPDATE_USER);
          response.setHeader(CONFIRM);
          response.setDataset(Protocol.DATASET);
          response.setUserid(user.getUid());
          response.setEmail(user.getMail());
          response.setNickname(user.getNickname());
          response.setForegroundColor(user.getForeground());
          response.setBackgroundColor(user.getBackground());
          response.setVolume(user.getVolume());
          response.setOncall(user.isOncall());
          response.setPassword(user.getPassword());
          response.setIp(request.getIp());
          return response;
        }
      }
    }
    UPDATEUSER error = new UPDATEUSER();
    error.setCommand(Command.UPDATE_USER);
    error.setHeader(ERROR);
    return error;
  }



  private UPLOADFILES uploadfiles(UPLOADFILES request) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    try {
      if (request.hasStarted()) databaseservice.deleteUploadfiles(request.getUserid(), request.getIp());

      databaseservice.insertUploadfiles(
          request.getUserid(), request.getIp(), request.getPort(), request.getUploadfile()
      );

      UPLOADFILES uploadfiles = new UPLOADFILES();
      uploadfiles.setCommand(Command.UPLOADFILES);
      uploadfiles.setHeader(RESPONSE);
      uploadfiles.setDataset(Protocol.DATASET);
      uploadfiles.setUserid(request.getUserid());
      uploadfiles.setIp(request.getIp());
      uploadfiles.setStarted(request.hasStarted());
      return uploadfiles;
    }
    catch (SQLException e) {
      log.error(e.getMessage(), e.getCause());
      UPLOADFILES uploadfiles = new UPLOADFILES();
      uploadfiles.setCommand(Command.UPLOADFILES);
      uploadfiles.setHeader(ERROR);
      uploadfiles.setDataset(Protocol.DATASET);
      uploadfiles.setErrorCode(e.getErrorCode());
      uploadfiles.setErrorMessage(e.getMessage());
      return uploadfiles;
    }
  }



  /**
   * Ist die Botschaft formal korrekt.
   * 
   * @return
   */
  private USRLOGIN userlogin(USRLOGIN request, Session session) {
    if (request.getHeader() == REQUEST) {
      if (request.getDataset().equals(Protocol.DATASET)) {
        if (log.isDebugEnabled()) log.debug("USRLOGIN=" + request.getEmail());
        DatabaseService databaseservice = WebdatabaseImpl.getInstance();
        USRLOGIN response = databaseservice.loginUser(request, session.getId());
        if (response.getHeader() == CONFIRM) {
          databaseservice.updateConfirmationFlag(request.getUserid(), request.getEmail(), 0);
        }
        return response;
      }
    }
    USRLOGIN error = new USRLOGIN();
    error.setCommand(Command.USRLOGIN);
    error.setHeader(ERROR);
    return error;

  }



  /**
   * 
   * Alle Ueser, die gearda online sind
   * 
   * @param request
   * @return
   */
  private USERONLINELIST useronlinelist(USERONLINELIST request) {
    if (request.getHeader() == REQUEST) {
      DatabaseService databaseservice = WebdatabaseImpl.getInstance();
      USERONLINELIST response = databaseservice.fetchUseronlinelist(request.getUserid());
      return response;

    }
    USERONLINELIST useronlinelist = new USERONLINELIST();
    useronlinelist.setCommand(Command.USERONLINELIST);
    useronlinelist.setHeader(ERROR);
    useronlinelist.setDataset(Protocol.DATASET);
    useronlinelist.setErrorMessage("general failure");
    return useronlinelist;
  }



  /**
   * Chatuserliste wird gesendet;CHATUSERLIST fetchRoomlist wird gesende;ROOMLIST
   * Onlineliste wird gesendet;ONLINELIST
   * 
   * @param chatRaume
   *                  diese Räume werden aktualisiert
   */
  private void broadcast(List<String> chatRaume) {
    DatabaseService databaseservice = WebdatabaseImpl.getInstance();
    final Logger log = LogManager.getLogger(Webserver.class);

    // an alle Anwender wird gesendet und nicht nur die Anwender im Raum
    // die Benutzerlsiten(Nicknames) werden aktualisert
    // die Aktualisierung bezieht sich nur auf bestimmte Räume

    chatRaume.forEach(room -> {
      CHATUSERLIST chatuserlist = databaseservice.fetchChatuserlist(room);
      // muss noch optimiert werden, weil er an alle Anwender sendet
      // er soll nur an die Anwender in einem Raum senden
      // TODO wenn die user/Sessions ermittelt wurden, ist alles vorhanden
      broadcastAll(chatuserlist);
    });

    // Roomliste übergeben
    ROOMLIST roomlist;
    try {
      roomlist = databaseservice.fetchRoomlist(); // Alle vorhandenen Chaträume werden ermittelt
    }
    catch (DatabaseException e2) {
      log.warn(e2.getMessage(), e2.getCause());
      return;
    }
    // sende eine vollständige Raumliste an alle Anwender
    // ROMMLIST beschreibt die Räume(Pausenraum, Forum etc.) und nicht die Anzahl
    // der User und deren Attribute JLauncher Alle Räume anzeigen
    broadcastAll(roomlist);
    // Onlinelist übergeben Onlinelist hat einen Request
    // Ermittle die Anzehl aller User in einem Raum
    // Zurückgegeben werden der Raum und die Anzahl der Anwender in dem Raum
    CHATONLINELIST onlinelist;
    try {
      onlinelist = databaseservice.fetchOnlinelist();
    }
    catch (DatabaseException e1) {
      log.warn(e1.getMessage(), e1.getCause());
      return;
    }
    // Sende an alle Anwender
    broadcastAll(onlinelist);
  }



  /**
   * Der Server wird heruntergefahren und alle offenen Threads beendet.
   */
  public static void shutdown() {
    executorKeepAlive.shutdown();
    try {
      boolean done = executorKeepAlive.awaitTermination(1, TimeUnit.SECONDS);
      if (!done) {
        executorKeepAlive.shutdownNow();
      }
    }
    catch (InterruptedException e) {
      log.error(e.getMessage(), e.getCause());
    }
    log.info("KEEPALIVE is down");
  }

}
