/*
 * The MIT License
 * Copyright © 2023 Thorsten Schulz <https://javacomm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.javacomm.mastermind.gui;

import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.border.EtchedBorder;
import net.javacomm.mastermind.audio.SoundClip;
import net.javacomm.mastermind.control.Status;
import net.javacomm.mastermind.control.StatusEvent;
import net.javacomm.mastermind.control.StatusListener;
import net.javacomm.mastermind.tool.DefColors;
import net.javacomm.multilingual.Babelfish;
import net.javacomm.multilingual.MultilingualLabel;
import net.javacomm.multilingual.MultilingualString;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;
import net.javacomm.window.manager.Control;



/**
 * Der Anwender tifft für das Suchmuster eine Farbvorauswahl.
 */
@SuppressWarnings("serial")
public class JColorSelectPanel extends JPanel implements Babelfish, StatusListener {

//  private final static Logger log = LogManager.getLogger(JColorSelectPanel.class);
  private EtchedBorder etched = new EtchedBorder(EtchedBorder.LOWERED);
  private GridBagLayout gridbag = new GridBagLayout();
  private GridBagConstraints c = new GridBagConstraints();
  private ButtonGroup chooseButtons = new ButtonGroup();
  private Color background = Constants.OLIV;
  private JColorRadioButton purpurButton = new JColorRadioButton(DefColors.PURPUR.colorValue(), background);
  private JColorRadioButton blueButton = new JColorRadioButton(DefColors.BLUE.colorValue(), background);
  private JColorRadioButton yellowButton = new JColorRadioButton(DefColors.YELLOW.colorValue(), background);
  private JColorRadioButton orangeButton = new JColorRadioButton(DefColors.ORANGE.colorValue(), background);
  private JColorRadioButton greenButton = new JColorRadioButton(DefColors.GREEN.colorValue(), background);
  private JColorRadioButton redButton = new JColorRadioButton(DefColors.RED.colorValue(), background);
  private DefColors chosenColor;
  private PropertyChangeSupport changes = new PropertyChangeSupport(this);
  private MultilingualString benutzerfehler = new MultilingualString(KEY.STRING_BENUTZERFEHLER);
  private MultilingualString spielstart = new MultilingualString(KEY.STRING_ES_WURDE_NICHT);
  private ImageIcon logo = new ImageIcon(getClass().getResource(Constants.LOGO));
  protected MultilingualLabel verloren = new MultilingualLabel(KEY.LABEL_DU_HAST_VERLOREN);
  protected ImageIcon sadman = new ImageIcon(getClass().getResource(Constants.SADMAN));
  protected ImageIcon hurraman = new ImageIcon(getClass().getResource(Constants.HURRAMAN));
  protected MultilingualLabel gewonnen = new MultilingualLabel(KEY.LABEL_DU_HAST_GEWONNEN);

  private Status status;

  public JColorSelectPanel() {
    setBackground(background);
    setLayout(gridbag);

    chooseButtons.add(yellowButton);
    chooseButtons.add(orangeButton);
    chooseButtons.add(purpurButton);
    chooseButtons.add(blueButton);
    chooseButtons.add(greenButton);
    chooseButtons.add(redButton);

    yellowButton.addActionListener(event -> {
      checkStatus(DefColors.YELLOW, yellowButton);
    });
    orangeButton.addActionListener(event -> {
      checkStatus(DefColors.ORANGE, orangeButton);
    });
    greenButton.addActionListener(event -> {
      checkStatus(DefColors.GREEN, greenButton);
    });
    purpurButton.addActionListener(event -> {
      checkStatus(DefColors.PURPUR, purpurButton);
    });
    redButton.addActionListener(event -> {
      checkStatus(DefColors.RED, redButton);
    });
    blueButton.addActionListener(event -> {
      checkStatus(DefColors.BLUE, blueButton);
    });

    c.gridy = 0;
    c.weightx = 1.0;
    c.gridx = 0;
    gridbag.setConstraints(yellowButton, c);
    add(yellowButton);

    c.gridx = 1;
    gridbag.setConstraints(orangeButton, c);
    add(orangeButton);

    c.gridx = 2;
    gridbag.setConstraints(redButton, c);
    add(redButton);

    c.gridx = 3;
    gridbag.setConstraints(purpurButton, c);
    add(purpurButton);

    c.gridx = 4;
    gridbag.setConstraints(blueButton, c);
    add(blueButton);

    c.gridx = 5;
    gridbag.setConstraints(greenButton, c);
    add(greenButton);

    setBorder(etched);

  }



  private void errorClick() {
    try {
      SoundClip.playClickError();
    }
    catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
  }



  public void addColorListener(PropertyChangeListener listener) {
    changes.addPropertyChangeListener(listener);

  }



  public void removeColorListener(PropertyChangeListener listener) {
    changes.removePropertyChangeListener(listener);
  }



  public void removeAllListener() {
    for (PropertyChangeListener pl : changes.getPropertyChangeListeners()) {
      removeColorListener(pl);
    }
  }



  /**
   * Der Anwender hat diese Farbe ausgewählt.
   * 
   * @return eine Farbe
   */
  public DefColors getColor() {
    return chosenColor;
  }



  @Override
  public void setLanguage(ISO639 code) {
    benutzerfehler.setLanguage(code);
    spielstart.setLanguage(code);
    verloren.setLanguage(code);
    gewonnen.setLanguage(code);
  }



  /**
   * Die aktuelle Farbe wird auf der Leiste hervorgehoben.
   * 
   * @param color
   *              eine Farbe
   */
  void setFarbvorauswahl(DefColors color) {
    switch(color) {
      case BLACK:
        break;
      case BLUE:
        blueButton.setSelected(true);
        break;
      case BROWN:
        break;
      case GRAY:
        break;
      case GREEN:
        greenButton.setSelected(true);
        break;
      case ORANGE:
        orangeButton.setSelected(true);
        break;
      case PURPUR:
        purpurButton.setSelected(true);
        break;
      case RED:
        redButton.setSelected(true);
        break;
      case WHITE:
        break;
      case YELLOW:
        yellowButton.setSelected(true);
        break;
      default:
        throw new UnsupportedOperationException(
            color + " - die Farbe ist nicht auf der Farbauswahlleiste entahlten"
        );
    }
  }



  /**
   * Alle Knöpfe werden auf nicht gedrückt eingestellt.
   */
  private void deselect() {
    chooseButtons.clearSelection();
  }



  /**
   * Spielt Verloren-Audio.
   */
  protected void gameOver() {
    try {
      SoundClip.playCartoonGossip();
    }
    catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
  }



  private void checkStatus(DefColors color, JComponent component) {
    switch(status) {
      case Init:
        deselect();
        errorClick();
        JOptionPane.showInternalMessageDialog(
            component, spielstart.toString(), benutzerfehler.toString(), JOptionPane.PLAIN_MESSAGE, logo
        );
        break;
      case Lost:
        deselect();
        gameOver();
        JOptionPane
            .showInternalMessageDialog(component, verloren, "Game Over", JOptionPane.PLAIN_MESSAGE, sadman);
        SoundClip.stopCartoonGossipClip();
        break;
      case Running:
        chosenColor = color;
        changes.firePropertyChange(JColorSelectPanel.this.getName(), Control.NULL, Control.SELECTED);
        break;
      case Won:
        deselect();
        try {
          SoundClip.playVictory();
        }
        catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
        JOptionPane
            .showInternalMessageDialog(component, gewonnen, "Game Over", JOptionPane.PLAIN_MESSAGE, hurraman);
        break;
      default:
        break;
    }

  }



  @Override
  public void stateChanged(StatusEvent event) {
    status = event.hasStatus();
    switch(status) {
      case Init:
        // TODO auf deselect
        deselect();
        break;
      case Lost:
        break;
      case Running:
        break;
      case Won:
        break;
      default:
        break;

    }
  }

}
