/*
 * The MIT License
 * Copyright © 2023 Thorsten Schulz <https://javacomm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.javacomm.mastermind.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyVetoException;
import java.io.IOException;
import java.util.SortedMap;
import java.util.TreeMap;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.ImageIcon;
import javax.swing.JInternalFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.Timer;
import javax.swing.WindowConstants;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.mastermind.audio.SoundClip;
import net.javacomm.mastermind.control.Status;
import net.javacomm.mastermind.control.StatusController;
import net.javacomm.mastermind.control.StatusEvent;
import net.javacomm.mastermind.tool.CodeGenerator;
import net.javacomm.mastermind.tool.DefColors;
import net.javacomm.multilingual.Babelfish;
import net.javacomm.multilingual.MultilingualLabel;
import net.javacomm.multilingual.MultilingualString;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;
import net.javacomm.window.manager.Control;



/**
 * Das GamePanel ist die linke Bildschirmhälfte im Mastermind-InternalFrame.
 * 
 */
public class JGamePanel extends JPanel implements Babelfish {

  private final static Logger log = LogManager.getLogger(JGamePanel.class);
  private static final long serialVersionUID = -4865068845402460309L;

  private JColorSelectPanel colorSelectPanel = new JColorSelectPanel();
  private JDifficultyPanel difficultyPanel = new JDifficultyPanel();
  private ControlAndResultView controlAndResultView = new ControlAndResultView();
  private BorderLayout border = new BorderLayout();
  private ImageIcon runningPicture = new ImageIcon(getClass().getResource(Constants.GAME_IS_STARTING));
  private MultilingualString spielStarten = new MultilingualString(KEY.BUTTON_SPIEL_STARTEN);
  private Timer timer;
  private MultilingualString weiterSpielenOderAbbrechen = new MultilingualString(
      KEY.STRING_WEITERSPIELEN_ODER_ABBRECHEN
  );
  private MultilingualString spiel = new MultilingualString(KEY.STRING_SPIEL);
  private MultilingualString abortar = new MultilingualString(KEY.STRING_ABORT);
  private MultilingualString fortsetzen = new MultilingualString(KEY.STRING_FORTSETZEN);
  private MultilingualString afoot = new MultilingualString(KEY.STRING_AFOOT);
  private boolean confirmDialog = false;
  private Cardpanel cardpanel2;
  private JSpielfeld8 acht = new JSpielfeld8();
  private JSpielfeld12 zwoelf = new JSpielfeld12();
  private JSpielfeld16 sechzehn = new JSpielfeld16();
  private SortedMap<Integer, ASpielfeld> mapLevelSpielfeld = new TreeMap<>();
  private Status status;
  private MultilingualString benutzerfehler = new MultilingualString(KEY.STRING_BENUTZERFEHLER);
  private MultilingualString spielstart = new MultilingualString(KEY.STRING_ES_WURDE_NICHT);
  private ImageIcon logo = new ImageIcon(getClass().getResource(Constants.LOGO));
  private ASpielfeld arena; // auf diesem Spielfeld wird gerade gespielt
  private StatusController controller = new StatusController();
  protected ImageIcon hurraman = new ImageIcon(getClass().getResource(Constants.HURRAMAN));
  protected MultilingualLabel gewonnen = new MultilingualLabel(KEY.LABEL_DU_HAST_GEWONNEN);
  protected MultilingualLabel verloren = new MultilingualLabel(KEY.LABEL_DU_HAST_VERLOREN);
  protected ImageIcon sadman = new ImageIcon(getClass().getResource(Constants.SADMAN));

  public JGamePanel() {
    controller.addStatusListener(difficultyPanel);
    controller.addStatusListener(colorSelectPanel);
    controller.addStatusListener(acht);
    controller.addStatusListener(zwoelf);
    controller.addStatusListener(sechzehn);
    controller.addStatusListener((event) -> {
      status = event.hasStatus();
    });
    controller.trigger(new StatusEvent(this, Status.Init));
    cardpanel2 = new Cardpanel();

    mapLevelSpielfeld.put(1, sechzehn);
    mapLevelSpielfeld.put(2, zwoelf);
    mapLevelSpielfeld.put(3, acht);

    sechzehn.addSpielfeldListener(event -> {
      Control value = (Control) event.getNewValue();
      switch(value) {
        case COLOR:
          colorSelectPanel.setFarbvorauswahl(sechzehn.getAktuelleFarbe());
          return;
        case GEWONNEN:
          controller.trigger(new StatusEvent(sechzehn, Status.Won));
          controlAndResultView.setSolutionPattern(sechzehn.getSolutionPattern());
          return;
        case VERLOREN:
          controller.trigger(new StatusEvent(sechzehn, Status.Lost));
          controlAndResultView.setSolutionPattern(sechzehn.getSolutionPattern());
          try {
            SoundClip.playGamersFail();
          }
          catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
            log.info(e.getMessage(), e);
          }
//          JOptionPane
//              .showInternalMessageDialog(sechzehn, verloren, "Game Over", JOptionPane.PLAIN_MESSAGE, sadman);
//
          // TODO einblenden Verlust wegen Rundenende
          return;
        default:
          throw new UnsupportedOperationException();
      }

    });
    acht.addSpielfeldListener(event -> {
      Control value = (Control) event.getNewValue();
      switch(value) {
        case COLOR:
          colorSelectPanel.setFarbvorauswahl(acht.getAktuelleFarbe());
          return;
        case GEWONNEN:
          controller.trigger(new StatusEvent(acht, Status.Won));
          controlAndResultView.setSolutionPattern(acht.getSolutionPattern());
          return;
        case VERLOREN:
          controller.trigger(new StatusEvent(acht, Status.Lost));
          controlAndResultView.setSolutionPattern(acht.getSolutionPattern());
          try {
            SoundClip.playGamersFail();
          }
          catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
            log.info(e.getMessage(), e);
          }
//          cartoonGossip();
//          JOptionPane
//              .showInternalMessageDialog(acht, verloren, "Game Over", JOptionPane.PLAIN_MESSAGE, sadman);
//          SoundClip.stopCartoonGossipClip();

          return;
        default:
          throw new UnsupportedOperationException();
      }

    });
    zwoelf.addSpielfeldListener(event -> {
      Control value = (Control) event.getNewValue();
      switch(value) {
        case COLOR:
          colorSelectPanel.setFarbvorauswahl(zwoelf.getAktuelleFarbe());
          return;
        case GEWONNEN:
          controller.trigger(new StatusEvent(zwoelf, Status.Won));
          controlAndResultView.setSolutionPattern(zwoelf.getSolutionPattern());
          return;
        case VERLOREN:
          controller.trigger(new StatusEvent(zwoelf, Status.Lost));
          controlAndResultView.setSolutionPattern(zwoelf.getSolutionPattern());
          try {
            SoundClip.playGamersFail();
          }
          catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
            log.info(e.getMessage(), e);
          }

//          cartoonGossip();
//          JOptionPane
//              .showInternalMessageDialog(zwoelf, verloren, "Game Over", JOptionPane.PLAIN_MESSAGE, sadman);
//          SoundClip.stopCartoonGossipClip();

          return;
        default:
          throw new UnsupportedOperationException();
      }

    });

    cardpanel2.addSpielfelder(mapLevelSpielfeld);

    difficultyPanel.addDifficultListener(event -> {
      cardpanel2.setSchwierigkeitsgrad(difficultyPanel.getLevel());
    });
    controlAndResultView.addVetoableResultControlListener(event -> {

      Control value = (Control) event.getNewValue();
      switch(value) {
        case START_GAME:
          switch(status) {
            case Init:
              infoGameCountdown();
              return;
            case Lost:
              cartoonGossip();
              JOptionPane.showInternalMessageDialog(
                  controlAndResultView.getComponentStartGame(), verloren, "Game Over",
                  JOptionPane.PLAIN_MESSAGE, sadman
              );
              SoundClip.stopCartoonGossipClip();
              return;
            case Running:
              throw new PropertyVetoException("das Spiel läuft", event);
            case Won:
              try {
                SoundClip.playMonkeyApplause();
              }
              catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
              JOptionPane.showInternalMessageDialog(
                  controlAndResultView.getComponentStartGame(), gewonnen, "Game Over",
                  JOptionPane.PLAIN_MESSAGE, hurraman
              );

              return;
            default:
              throw new UnsupportedOperationException();
          }
        case NEW_GAME:

          switch(status) {
            case Init:
              log.info("init");
              // klick Fehlermeldung
              // INit bleiben
//              infoGameCountdown();
              try {
                SoundClip.playClickError();
              }
              catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
                log.info(e.getMessage(), e);
              }
              JOptionPane.showInternalMessageDialog(
                  controlAndResultView.getComponentNewGame(), spielstart.toString(),
                  benutzerfehler.toString(), JOptionPane.PLAIN_MESSAGE, logo
              );

              break;
            case Lost:
              // Löschen, auf init
              log.info("lost");
              arena = mapLevelSpielfeld.get(difficultyPanel.getLevel());
              arena.reset();
              controlAndResultView.setSolutionPattern(
                  String.valueOf(
                      new char[] {DefColors.GRAY.charValue(), DefColors.GRAY.charValue(),
                          DefColors.GRAY.charValue(), DefColors.GRAY.charValue()}
                  )
              );
              controller.trigger(new StatusEvent(this, Status.Init));
              try {
                SoundClip.playPlayerRecharging();
              }
              catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
              break;
            case Running:
              questionFortsetzenOderAufgeben();
              break;
            case Won:
              log.info("won, neues Spiel neues männchen");
              try {
                SoundClip.playLongGameover();
              }
              catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
              arena = mapLevelSpielfeld.get(difficultyPanel.getLevel());
              arena.reset();
              controlAndResultView.setSolutionPattern(
                  String.valueOf(
                      new char[] {DefColors.GRAY.charValue(), DefColors.GRAY.charValue(),
                          DefColors.GRAY.charValue(), DefColors.GRAY.charValue()}
                  )
              );
              controller.trigger(new StatusEvent(this, Status.Init));
              break;
            default:
              break;
          }
          break;
        default:
          throw new UnsupportedOperationException();
      }

    });
    colorSelectPanel.addColorListener(event -> {

      // Die Farbe muss JCardPanel -> JSpielfeld -> JPlayRoundPanel ->
      // JRoundAttemptPanel mitgeteilt werden
      // Spiel läuft
      Control control = (Control) event.getNewValue();
      switch(control) {
        case SELECTED:
          ASpielfeld result = mapLevelSpielfeld.get(difficultyPanel.getLevel());
          result.setAktuelleFarbe(colorSelectPanel.getColor());
          break;
        default:
          throw new UnsupportedOperationException("nicht unterstützt");
      }

    });

    setBackground(Constants.OLIV);
    setLayout(border);

    this.add(colorSelectPanel, BorderLayout.SOUTH);
    this.add(difficultyPanel, BorderLayout.WEST);
    this.add(controlAndResultView, BorderLayout.NORTH);
    this.add(cardpanel2, BorderLayout.CENTER);
  }



  /**
   * Spielt Verloren-Audio.
   */
  private void cartoonGossip() {
    try {
      SoundClip.playCartoonGossip();
    }
    catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
  }



  private void questionFortsetzenOderAufgeben() {
    if (confirmDialog) return;
    confirmDialog = true;
    JOptionPane pane = new JOptionPane(
        weiterSpielenOderAbbrechen.toString(), JOptionPane.QUESTION_MESSAGE, JOptionPane.OK_CANCEL_OPTION,
        runningPicture, new String[] {fortsetzen.toString(), abortar.toString()}
    );

    JInternalFrame question = pane
        .createInternalFrame(controlAndResultView.getComponentNewGame(), spiel.toString());
    question.setMaximizable(false);
    question.setResizable(false);
    question.setIconifiable(false);
    question.setClosable(false);
    question.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
    question.show();
    question.addInternalFrameListener(new InternalFrameAdapter() {
      @Override
      public void internalFrameClosed(InternalFrameEvent event) {
        confirmDialog = false;
        if (pane.getValue().toString().equals(abortar.toString())) {
          try {
            SoundClip.playGamersFail();
          }
          catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
            log.info(e.getMessage(), e);
          }
          controller.trigger(new StatusEvent(this, Status.Lost));
          arena = mapLevelSpielfeld.get(difficultyPanel.getLevel());
          controlAndResultView.setSolutionPattern(arena.getSolutionPattern());
        }
        else {
          try {
            SoundClip.playJumpingInVideogame();
          }
          catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
            log.info(e.getMessage(), e);
          }
        }
      }
    });
  }



  /**
   * Eine Infomeldung, dass ein Spiel gestartet wird.
   * 
   */
  private void infoGameCountdown() {
    ASpielfeld result = mapLevelSpielfeld.get(difficultyPanel.getLevel());
    controller.trigger(new StatusEvent(this, Status.Running));

    result.setSolutionPattern(
        difficultyPanel.isSameColorsAreAllowed() ? CodeGenerator.getColorCode()
            : CodeGenerator.getColorCodeUnique()
    );

    try {
      SoundClip.playSimpleGameCountdown();
    }
    catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
      log.info(e.getMessage(), e);
    }

    JOptionPane pane = new JOptionPane(
        afoot.toString(), JOptionPane.PLAIN_MESSAGE, JOptionPane.DEFAULT_OPTION, runningPicture, new Object[0]
    );
    JInternalFrame question = pane
        .createInternalFrame(controlAndResultView.getComponentNewGame(), spiel.toString());
    question.setMaximizable(false);
    question.setResizable(false);
    question.setIconifiable(false);
    question.setClosable(false);
    question.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
    question.show();

    timer = new Timer(4000, new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent event) {
        try {
          question.setClosed(true);
        }
        catch (PropertyVetoException e1) {}
      }

    });
    timer.setRepeats(false);
    timer.start();
  }



  /**
   * Entferne alle Listener und gib den Speicher frei.
   */
  void removeAllListener() {
    colorSelectPanel.removeAllListener();
    difficultyPanel.removeAllListener();
    controller.removeAllListener();
  }



  @Override
  public void setLanguage(ISO639 code) {
    controlAndResultView.setLanguage(code);
    difficultyPanel.setLanguage(code);
    colorSelectPanel.setLanguage(code);
    spielStarten.setLanguage(code);
    weiterSpielenOderAbbrechen.setLanguage(code);
    spiel.setLanguage(code);
    abortar.setLanguage(code);
    fortsetzen.setLanguage(code);
    afoot.setLanguage(code);
    benutzerfehler.setLanguage(code);
    spielstart.setLanguage(code);

  }



  /**
   * Die rechte untere Ecke im {@code JScrollPane} erhält die richtige Farbe und
   * der Rahmen wird dabei übermalt.
   * 
   * @param color
   *              Eckfarbe
   */
  void corner(Color color) {
    ASpielfeld tmp = mapLevelSpielfeld.get(difficultyPanel.getLevel());
    tmp.corner(color);
  }



  /**
   * In der Grundeinstellung für jedes Spiel soll der vertikale Scrollbalken ganz
   * unten sein, so dass die 1-te Spielrunde angezeigt wird.
   * 
   */
  void scrollToFirstRow() {
    cardpanel2.setSchwierigkeitsgrad(difficultyPanel.getLevel());
  }

  /**
   * Diese Klasse existiert nur als Platzhalter für ein CardLayout. Im CardLayout
   * werden drei unterschiedliche Spielfelder abgelegt.
   */
  @SuppressWarnings("serial")
  private static class Cardpanel extends JPanel {
    private final static Logger log = LogManager.getLogger(Cardpanel.class);
    private CardLayout card = new CardLayout();
    private SortedMap<Integer, ASpielfeld> spielfelder = new TreeMap<>();
    private boolean flag = false;

    private Cardpanel() {
      setLayout(card);
    }



    private void setSchwierigkeitsgrad(int level) {
      card.show(this, spielfelder.get(level).getDescription());
      spielfelder.get(level).scrollToFirstRow();
    }



    /**
     * Das Cardpanel bekommt seine Spielfelder übergeben. Alle Spielfelder werden
     * gesammelt und können nur einmal übergeben werden. Ein abermaliger Aufruf
     * bewirkt nichts.
     * 
     * @param map
     *            alle Spielfelder
     */
    private void addSpielfelder(SortedMap<Integer, ASpielfeld> map) {
      if (flag) {
        log.warn("Diese Methode wurde bereits aufgerufen. Alle Spielfelder werden auf einmal übergeben.");
        return;
      }
      flag = true;
      spielfelder = map;
      spielfelder.keySet().forEach(key -> {
        add(spielfelder.get(key), spielfelder.get(key).getDescription());
      });
    }

  }

}
