/*
 * The MIT License
 * Copyright © 2023 Thorsten Schulz <https://javacomm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.javacomm.mastermind.gui;

import java.awt.GridBagConstraints;
import java.io.IOException;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JOptionPane;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.mastermind.audio.SoundClip;
import net.javacomm.mastermind.control.MastermindException;
import net.javacomm.mastermind.control.Status;
import net.javacomm.mastermind.control.StatusEvent;
import net.javacomm.mastermind.tool.DefColors;
import net.javacomm.mastermind.tool.Knopf;
import net.javacomm.mastermind.tool.PatternChecker;
import net.javacomm.window.manager.Control;



@SuppressWarnings("serial")
public class JSpielfeld12 extends ASpielfeld {

  private final static Logger log = LogManager.getLogger(JSpielfeld12.class);
  private final static int DOZE = 12;

  private RundeSuchmusterErgebnis[] rundeSuchmusterErgebnis;

  private int aktuelleSpielrunde; // control Spiel
  private String attemptPattern; // control Spiel
  private Status status;
  private DefColors aktuelleFarbe;

  private KnopfUndFarbe alt, neu;

  public JSpielfeld12() {
    super();

    rundeSuchmusterErgebnis = new RundeSuchmusterErgebnis[DOZE]; // Zahl 4felder ergebnis
    for (int index = 0; index < DOZE; index++) {
      rundeSuchmusterErgebnis[index] = new RundeSuchmusterErgebnis(DOZE - index);
      rundeSuchmusterErgebnis[index].addRundenListener(event -> {
        // Wer war der Auslöser?
        for (RundeSuchmusterErgebnis element : rundeSuchmusterErgebnis) {
          if (event.getSource().equals(element)) {
            Control value = (Control) event.getNewValue();

            switch(hasStatus()) {
              case Lost:
                if (value == Control.KNOPF_AUF_RUNDENZAHL) {
                  gameOver();
                  JOptionPane.showInternalMessageDialog(
                      element.getKnopfAufRundenzahl(), verloren, "Game Over", JOptionPane.PLAIN_MESSAGE,
                      sadman
                  );
                  SoundClip.stopCartoonGossipClip();

                }
                else {
                  gameOver();
                  JOptionPane.showInternalMessageDialog(
                      element.getKnopfleiste(), verloren, "Game Over", JOptionPane.PLAIN_MESSAGE, sadman
                  );
                  SoundClip.stopCartoonGossipClip();

                }
                return;
              case Won:
                if (value == Control.KNOPF_AUF_RUNDENZAHL) {
                  try {
                    SoundClip.playVictory();
                  }
                  catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
                  JOptionPane.showInternalMessageDialog(
                      element.getKnopfAufRundenzahl(), gewonnen, "Game Over", JOptionPane.PLAIN_MESSAGE,
                      hurraman
                  );

                }
                else {
                  try {
                    SoundClip.playVictory();
                  }
                  catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
                  JOptionPane.showInternalMessageDialog(
                      element.getKnopfleiste(), gewonnen, "Game Over", JOptionPane.PLAIN_MESSAGE, hurraman
                  );

                }
                return;

              case Init:
                errorClip();
                JOptionPane.showInternalMessageDialog(
                    element.getKnopfAufRundenzahl(), spielstart.toString(), benutzerfehler.toString(),
                    JOptionPane.PLAIN_MESSAGE, logo
                );
                return;
              case Running:
                // richtige Runde?
                if (aktuelleSpielrunde == element.getNummer()) {
                  // alles ok
                  try {
                    aktuelleSpielrunde(element, value);
                  }
                  catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
                    log.info(e.getMessage(), e);
                  }
                  return;
                }
                else {
                  if (value == Control.KNOPF_AUF_RUNDENZAHL) {
                    duBistInPanel.setKnopfAufRundenzahl(aktuelleSpielrunde);
                    errorClip();
                    JOptionPane.showInternalMessageDialog(
                        element.getKnopfAufRundenzahl(), duBistInPanel, benutzerfehler.toString(),
                        JOptionPane.PLAIN_MESSAGE, logo
                    );
                  }
                  else if (
                    value == Control.KNOPF_AUF_SUCHMUSTER_LINKS
                        || value == Control.KNOPF_AUF_SUCHMUSTER_RECHTS
                  ) {
                    duBistInPanel.setKnopfAufRundenzahl(aktuelleSpielrunde);
                    errorClip();
                    JOptionPane.showInternalMessageDialog(
                        element.getKnopfleiste(), duBistInPanel, benutzerfehler.toString(),
                        JOptionPane.PLAIN_MESSAGE, logo
                    );
                  }
                }
                return;
            }
          }
        }
      });
      con.gridx = 0;
      con.gridy++;
      con.weightx = 1.0;
      con.weighty = 1.0;
      con.fill = GridBagConstraints.BOTH;
      gridbag.setConstraints(rundeSuchmusterErgebnis[index], con);
      mainpane.add(rundeSuchmusterErgebnis[index]);
    }
    reset();
  }



  @Override
  public void reset() {
    status = Status.Init;
    aktuelleFarbe = DefColors.GRAY;

    aktuelleSpielrunde = 1;
    attemptPattern = String.valueOf(
        new char[] {DefColors.GRAY.charValue(), DefColors.GRAY.charValue(), DefColors.GRAY.charValue(),
            DefColors.GRAY.charValue()}
    );
    solutionPattern = String.valueOf(
        new char[] {DefColors.BLACK.charValue(), DefColors.BLACK.charValue(), DefColors.BLACK.charValue(),
            DefColors.BLACK.charValue()}
    );
    for (RundeSuchmusterErgebnis tmp : rundeSuchmusterErgebnis) {
      for (Knopf knopf : Knopf.values()) {
        tmp.setButtonColor(knopf, DefColors.GRAY);
      }

      tmp.setReport(
          String.valueOf(
              new char[] {DefColors.BLACK.charValue(), DefColors.BLACK.charValue(),
                  DefColors.BLACK.charValue(), DefColors.BLACK.charValue()}
          )
      );

    }

  }



  /**
   * Eine Hilfsmethode, die aus einer for-Schleife aufgerufen wird. Ich habe sie
   * ausgelagert, damit die for-schleife leichter gelesen werden kann.
   * 
   * @param element
   *                aus der for-schleife
   * @param value
   *                event aus der for-schleife
   * @throws LineUnavailableException
   *                                       wird von SoundClip geworfen
   * @throws UnsupportedAudioFileException
   *                                       wird von SoundClip geworfen
   * @throws IOException
   *                                       wird von SoundClip geworfen
   * @throws MastermindException
   */
  private void aktuelleSpielrunde(RundeSuchmusterErgebnis element, Control value)
      throws IOException, UnsupportedAudioFileException, LineUnavailableException {
    if (value == Control.KNOPF_AUF_SUCHMUSTER_RECHTS) {
      element.setButtonColor(element.getKnopfleiste().getKnopf(), DefColors.GRAY);
      alt = null;
    }
    else if (value == Control.KNOPF_AUF_SUCHMUSTER_LINKS) {
      neu = new KnopfUndFarbe();
      neu.setFarbe(aktuelleFarbe);
      neu.setKnopf(element.getKnopfleiste().getKnopf());

      if (neu.equals(alt)) {
        // wenn knopf und rundenfarbe identisch, dann nächste farbe
        // gleiche farben -- durchlaufen

        DefColors neueFarbe = farbkreis.get(alt.getFarbe());
        neu.setFarbe(neueFarbe);
        element.setButtonColor(element.getKnopfleiste().getKnopf(), neueFarbe);
        aktuelleFarbe = neueFarbe;

        changes.firePropertyChange(JSpielfeld12.class.getName(), Control.NULL, Control.COLOR);

      }
      else {
        // hier wir die Rundenfarbe gesetzt
        // farben sind nicht identisch
        // nächste farbe, wenn grau
        if (aktuelleFarbe == DefColors.GRAY) {
          aktuelleFarbe = DefColors.YELLOW;
          element.setButtonColor(element.getKnopfleiste().getKnopf(), aktuelleFarbe);
          neu = new KnopfUndFarbe();
          neu.setFarbe(aktuelleFarbe);
          neu.setKnopf(element.getKnopfleiste().getKnopf());
          changes.firePropertyChange(JSpielfeld12.class.getName(), Control.NULL, Control.COLOR);
        }
        else {
          element.setButtonColor(element.getKnopfleiste().getKnopf(), aktuelleFarbe);
        }
      }
      alt = neu;
    }
    else if (value == Control.KNOPF_AUF_RUNDENZAHL) {
      log.info("Rundenende?");
      attemptPattern = element.getKnopfleiste().getAttemptPattern();
      if (attemptPattern.contains(String.valueOf(DefColors.GRAY.charValue()))) {
        sleepClip();
        JOptionPane.showInternalMessageDialog(
            element.getKnopfAufRundenzahl(), labelRundeZuEnde,
            runde.toString() + " " + String.valueOf(aktuelleSpielrunde), JOptionPane.PLAIN_MESSAGE, sleepman
        );

        return;
      }
      // vergleiche Rundenmuster mit Lösungsmuster
      try {
        String rundenbericht = PatternChecker.roundUpReport(attemptPattern, solutionPattern);
        element.setReport(rundenbericht);
        if (
          rundenbericht.equals(
              String.valueOf(
                  new char[] {DefColors.WHITE.charValue(), DefColors.WHITE.charValue(),
                      DefColors.WHITE.charValue(), DefColors.WHITE.charValue()}
              )
          )
        ) {
          SoundClip.playVictory();
          changes.firePropertyChange(JSpielfeld12.class.getName(), Control.NULL, Control.GEWONNEN);
          return;
          // Spielende
        }
        else if (aktuelleSpielrunde >= DOZE) {
          log.info("verloren");
          changes.firePropertyChange(JSpielfeld12.class.getName(), Control.NULL, Control.VERLOREN);
          return;
        }
        else {
          SoundClip.playCompletionOfALevel();

          aktuelleSpielrunde++;
          attemptPattern = String.valueOf(
              new char[] {DefColors.GRAY.charValue(), DefColors.GRAY.charValue(), DefColors.GRAY.charValue(),
                  DefColors.GRAY.charValue()}
          );

        }
      }
      catch (MastermindException e) {
        // TODO
        log.fatal("Spiel ist kaputt");
        throw new RuntimeException(e);
      }
    }
  }



  @Override
  public int getRounds() {
    return DOZE;
  }



  @Override
  public Status hasStatus() {
    return status;
  }



  @Override
  public DefColors getAktuelleFarbe() {
    return aktuelleFarbe;
  }



  @Override
  public String getDescription() {
    return String.valueOf(DOZE);
  }



  @Override
  public void setAktuelleFarbe(DefColors color) {
    aktuelleFarbe = color;
  }



  @Override
  public void stateChanged(StatusEvent event) {
    status = event.hasStatus();
  }
}
