/*
 * The MIT License
 * Copyright © 2023 Thorsten Schulz <https://javacomm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.javacomm.mastermind.gui;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import javax.swing.JPanel;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.mastermind.tool.DefColors;
import net.javacomm.mastermind.tool.Knopf;
import net.javacomm.window.manager.Control;



/**
 * Eine Spielrunde mit Zählerstand, Suchmuster und Ergebnis. Die übergeordnete
 * Klasse ist JSpielfeld. Ein Spielfeld hat viele RundeSuchmusterErgebnis.
 */
public class RundeSuchmusterErgebnis extends JPanel {

  private final static Logger log = LogManager.getLogger(RundeSuchmusterErgebnis.class);

  private static final long serialVersionUID = -97341866464542053L;
  private JNumberButton numberButton; // Die Zahlen 1,2,3,4 ... Rundenknöpfe
  private SuchmusterPanel suchmusterPanel = new SuchmusterPanel();
  private JRoundReportPanel ergebnis = new JRoundReportPanel();

//  private JReportPanel ergebnis = new JReportPanel();

  private JBlank blank1 = new JBlank(20, 60);
  private JBlank blank2 = new JBlank(30, 60);
  private GridBagLayout gridbag = new GridBagLayout();
  private GridBagConstraints con = new GridBagConstraints();
  private int nummer;

  private PropertyChangeSupport changes = new PropertyChangeSupport(this);

  /**
   * 
   * @param round
   *              die Nummer vom Rundenknopf
   */
  public RundeSuchmusterErgebnis(int round) {
    nummer = round;
    numberButton = new JNumberButton(nummer);
    numberButton.addActionListener(event -> {
      changes.firePropertyChange(
          RundeSuchmusterErgebnis.class.getName(), Control.NULL, Control.KNOPF_AUF_RUNDENZAHL
      );
    });
    suchmusterPanel.addSuchmusterListener(event -> {
      changes.firePropertyChange(RundeSuchmusterErgebnis.class.getName(), Control.NULL, event.getNewValue());
    });

    setBackground(Constants.OLIV);

    con.gridy = 0;
    con.weightx = 1.0;
    con.weighty = 1.0;

    con.gridx = 0;
    gridbag.setConstraints(numberButton, con);
    this.add(numberButton);

    con.gridx = 1;
    gridbag.setConstraints(blank1, con);
    this.add(blank1);

    con.gridx = 2;
    gridbag.setConstraints(suchmusterPanel, con);
    this.add(suchmusterPanel);

    con.gridx = 3;
    gridbag.setConstraints(blank2, con);
    this.add(blank2);

    con.gridx = 4;
    gridbag.setConstraints(ergebnis, con);
    this.add(ergebnis);

    setBorder(null);
  }



  public void addRundenListener(PropertyChangeListener listener) {
    changes.addPropertyChangeListener(listener);
  }



  public void removeRundenListener(PropertyChangeListener listener) {
    changes.removePropertyChangeListener(listener);
  }



  public void removeAllListener() {
    for (PropertyChangeListener pl : changes.getPropertyChangeListeners()) {
      removeRundenListener(pl);
    }
    for (ActionListener al : numberButton.getActionListeners()) {
      numberButton.removeActionListener(al);
    }
    suchmusterPanel.removeAllListener();
  }



  /**
   * Der Rundenbericht wird übergeben und angezeigt. Wenn ein Spiel gelöscht wird,
   * werden über diese Methode alle Rundenberichte zurückgesetzt.
   * 
   * @param report
   *               die Treffer werden angezeigt
   */
  public void setReport(String report) {
    ergebnis.setReport(report);
  }



  /**
   * Der Spieler hat dieses Muster geraten.
   * 
   * @return
   */
  public String getAttemptPattern() {
    return suchmusterPanel.getAttemptPattern();
  }



  /**
   * Jede Reihe hat ihre Nummer von [1..8], [1..12] oder [1..16]. 8, 12 oder 16
   * ist abhängig von der Spielfeldgröße.
   * 
   * @return die geklickte Runde als Zahlenwert
   */
  public int getNummer() {
    return nummer;
  }



  /**
   * Über das Knopfpanel wird die Dialogposition beziehungsweise die Reihe auf dem
   * Bildschirm bestimmt. Eine Anzeigehilfe für den Anwender, damit er weiß, wo
   * der Fehler aufrat.
   * 
   * @return Knopfpanel
   */
  JNumberButton getKnopfAufRundenzahl() {
    return numberButton;
  }



  /**
   * Über das Suchmusterpanel oder die Knopfleiste wird die Dialogposition
   * beziehungsweise die Reihe auf dem Bildschirm bestimmt. Eine Anzeigehilfe für
   * den Anwender, damit er weiß, wo der Fehler aufrat.
   * 
   * @return Suchmusterpanel
   */
  SuchmusterPanel getKnopfleiste() {
    return suchmusterPanel;
  }



  /**
   * Die Knopffarbe für das Ratefeld wird gesetzt. Die Farbe wird nur gesetzt,
   * wenn der Knopf in der Raterunde liegt.
   * 
   * @param color
   *              eine Farbe
   */
  void setButtonColor(Knopf knopf, DefColors color) {
    suchmusterPanel.setButtonColor(knopf, color);
  }

}
