/*
 * The MIT License
 * Copyright © 2023 Thorsten Schulz <https://javacomm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.javacomm.mastermind.gui;

import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import javax.swing.JButton;
import javax.swing.JPanel;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.nexuswob.gui.JMouseButton;
import net.javacomm.mastermind.tool.DefColors;
import net.javacomm.mastermind.tool.Knopf;
import net.javacomm.window.manager.Control;



/**
 * Die Leiste enthält die 4 Farbknöpfe, die der Spieler rät.
 * 
 */
@SuppressWarnings("serial")
public class SuchmusterPanel extends JPanel {

  private final static Logger log = LogManager.getLogger(SuchmusterPanel.class);

  private final static int ANZAHL_KNOEPFE = Knopf.values().length;

  private JMouseButton[] button = new JMouseButton[ANZAHL_KNOEPFE];
  private GridLayout grid = new GridLayout(1, ANZAHL_KNOEPFE, 50, 50);
  private String strInputPattern = "";
  private PropertyChangeSupport changes = new PropertyChangeSupport(this);
  private Knopf knopf;

  /**
   * Nach der Initialisierung sind alle Knöpfe ausgegraut.
   */
  public SuchmusterPanel() {
    setBackground(Constants.OLIV);
    setLayout(grid);

    for (int i = 0; i < ANZAHL_KNOEPFE; i++) {
      button[i] = new JMouseButton();
      add(button[i]);
      button[i].addActionListener(event -> {
        for (int index = 0; index < ANZAHL_KNOEPFE; index++) {
          if (event.getSource().equals(button[index])) {
            knopf = Knopf.toKnopf(index);
            break;
          }
        }
        changes.firePropertyChange(
            SuchmusterPanel.class.getName(), Control.NULL, Control.KNOPF_AUF_SUCHMUSTER_LINKS
        );
      });
      button[i].addRightClickListener(event -> {

        for (int index = 0; index < ANZAHL_KNOEPFE; index++) {
          if (event.getSource().equals(button[index])) {
            knopf = Knopf.toKnopf(index);
            // TODO darf nur ausgegraut werden, wenn er in der richtigen Zeile ist
//            setButtonColor(knopf, DefColors.GRAY);
            break;
          }
        }
        changes.firePropertyChange(
            SuchmusterPanel.class.getName(), Control.NULL, Control.KNOPF_AUF_SUCHMUSTER_RECHTS
        );

      });

    }
    clear();
  }



  /**
   * Alle Knöpfe werden ausgegraut.
   * 
   */
  public void clear() {
    for (JButton tmp : button) {
      tmp.setBackground(DefColors.GRAY.colorValue());
    }
  }



  @Override
  public Dimension getPreferredSize() {
    return new Dimension(400, 60);
  }



  public void addSuchmusterListener(PropertyChangeListener l) {
    changes.addPropertyChangeListener(l);
  }



  public void removeSuchmusterListener(PropertyChangeListener l) {
    changes.removePropertyChangeListener(l);
  }



  void removeAllListener() {
    for (PropertyChangeListener pl : changes.getPropertyChangeListeners()) {
      changes.removePropertyChangeListener(pl);
    }

    for (JMouseButton tmp : button) {
      for (ActionListener al : tmp.getActionListeners()) {
        tmp.removeActionListener(al);
      }

      for (PropertyChangeListener pl : tmp.getPropertyChangeListeners()) {
        tmp.removeRightClickListener(pl);
      }

    }

  }



  /**
   * Das Suchmuster aller vier Knöpfe in der Spielrunde. Das Suchmuster ist ein
   * Akronym aus 4 Buchstaben.
   * 
   * 
   * @return das aktuelle Suchmuster
   */
  public String getAttemptPattern() {
    strInputPattern = "";
    char[] colorChar = new char[ANZAHL_KNOEPFE];
    for (int i = 0; i < ANZAHL_KNOEPFE; i++) {
      colorChar[i] = DefColors.toDefColors(button[i].getBackground()).charValue();
      strInputPattern += String.valueOf(colorChar[i]);
    }
    return strInputPattern;
  }



  /**
   * Welcher der 4 Knöpfe wurde auf der Leiste gedrückt?
   * 
   * @return genau dieser mit einer Knopfnummer
   */
  public Knopf getKnopf() {
    return knopf;
  }



  /**
   * Ein Knopf wird eingefärbt.
   * 
   * @param knopf
   *              einer der 4 Leistenknöpfe
   * @param color
   *              der Knopf erscheint in dieser Farbe
   */
  public void setButtonColor(Knopf knopf, DefColors color) {
    button[knopf.intValue()].setBackground(color.colorValue());
  }

}
