/**
 * Copyright © 2020 Luis Andrés Lange <http://javacomm.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.javacomm.protocol;

import net.javacomm.multilingual.schema.KEY;



/**
 * 
 * Eine Botschaft hat einen HEADER, ein COMMAND und optional einen DATASET.
 * HEADER und COMMAND sind für jede Botschaft zwingend vorgeschrieben. Ein
 * DATASET ist abhängig vom verwendeten COMMAND. COMMAND kann eine
 * Fehlerbeschreibung enthalten.
 * 
 * @author llange
 *
 */
public class MESSAGE {

  public final static String ERRORCODE = "ERRORCODE";
  public final static String ERRORMESSAGE = "ERRORMESSAGE";
  public final static String TEXT = "TEXT";
  public final static String MULTILINGUALKEY = "MULTILINGUALKEY";

  private HEADER header;
  private Command command;
  private String dataset;
  private Integer errorcode;
  private String errormessage;
  private String text;
  private KEY multilingualkey;

  /**
   * Wie lautet der Befehl?
   * 
   * @return das Kommando
   */
  public Command getCommand() {
    return command;
  }



  /**
   * Das Kommando wird beim ENCODING verwendet und allen Attributen eines
   * Kommandos vorangestellt.
   * 
   * @return alle Attribute eines Kommandos
   */
  public String getDataset() {
    return dataset;
  }



  public Integer getErrorCode() {
    return errorcode;
  }



  public String getErrorMessage() {
    return errormessage;
  }



  /**
   * {@link Protocol#REQUEST}, {@link Protocol#RESPONSE}, {@link Protocol#CONFIRM}
   * oder {@link Protocol#ERROR}?
   * 
   * @return der Header
   */
  public HEADER getHeader() {
    return header;
  }



  public KEY getMultilingualkey() {
    return multilingualkey;
  }



  /**
   * Ein beliebiger Kommentar oder eine Beschreibung.
   * 
   * @return text
   */
  public String getText() {
    return text;
  }



  /**
   * 
   * Jedes definierte Kommando aus {@link Protocol}.
   * 
   * @param value
   *              Kommando als String
   */
  public void setCommand(Command value) {
    command = value;
  }



  /**
   * Dataset ist abhängig vom verwendeten Kommando.
   * 
   * @param value
   *              Datensatz als String
   * 
   * @see DIAGNOSTICS
   * @see GETINTERFACES
   * @see REFRESHINTERFACES
   * @see USRLOGIN
   */
  public void setDataset(String value) {
    dataset = value;
  }



  /**
   * Eine Fehlerbeschreibung als Zahlenwert.
   * 
   * @param value
   *              Fehlerbeschreibung
   */
  public void setErrorCode(Integer value) {
    errorcode = value;
  }



  /**
   * Eine genaue Fehlerbescrheibung
   * 
   * @param value
   *              Fehlerbeschreibung
   */
  public void setErrorMessage(String value) {
    errormessage = value;
  }



  public void setHeader(HEADER header) {
    this.header = header;
  }



  public void setMultilingualkey(KEY multilingualkey) {
    this.multilingualkey = multilingualkey;
  }



  /**
   * Freier Text.
   * 
   * @param value
   *              text
   */
  public void setText(String value) {
    text = value;
  }

}
