package net.javacomm.protocol;

import static net.javacomm.protocol.HEADER.CONFIRM;
import static net.javacomm.protocol.HEADER.ERROR;
import static net.javacomm.protocol.HEADER.REQUEST;
import static net.javacomm.protocol.HEADER.RESPONSE;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonSyntaxException;
import jakarta.websocket.DecodeException;
import jakarta.websocket.EncodeException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.regex.Matcher;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import net.javacomm.multilingual.schema.KEY;



public class TestMessageDecoder {

  MessageDecoder messagedecoder;

  @Test
  public void decodeBEAMEROFF_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"BEAMEROFF\","
        + "\"DATASET\":{"
        + "\"USERID\":\"anaconda\""
        + "}}";

    BEAMEROFF beameroff = (BEAMEROFF) messagedecoder.decode(encoded);
    assertEquals(Command.BEAMEROFF, beameroff.getCommand(), "BEAMEROFF");
    assertEquals(REQUEST, beameroff.getHeader(), "REQUEST");
    assertEquals("anaconda", beameroff.getUserid(), "USERID");
  }



  @Test
  public void decodeBEAMEROFF_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"BEAMEROFF\","
        + "\"DATASET\":{"
        + "\"USERID\":\"anaconda\""
        + "}}";

    BEAMEROFF beameroff = (BEAMEROFF) messagedecoder.decode(encoded);
    assertEquals(Command.BEAMEROFF, beameroff.getCommand(), "BEAMEROFF");
    assertEquals(RESPONSE, beameroff.getHeader(), "RESPONSE");
    assertEquals("anaconda", beameroff.getUserid(), "USERID");
  }



  @Test
  public void decodeBREAKROOMS_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"BREAKROOMS\","
        + "\"DATASET\":{"
        + "\"USERID\":\"anaconda\""
        + "}}";

    BREAKROOMS breakrooms = (BREAKROOMS) messagedecoder.decode(encoded);
    assertEquals(Command.BREAKROOMS, breakrooms.getCommand(), "BREAKROOMS");
    assertEquals(REQUEST, breakrooms.getHeader(), "REQUEST");
    assertEquals("anaconda", breakrooms.getUserid(), "USERID");
  }



  @Test
  public void decodeBREAKROOMS_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"BREAKROOMS\","
        + "\"DATASET\":{"
        + "\"USERID\":\"anaconda\","
        + "\"ROOMS\":[\"schach\",\"fussball\"]"
        + "}}";

    BREAKROOMS breakrooms = (BREAKROOMS) messagedecoder.decode(encoded);
    assertEquals(Command.BREAKROOMS, breakrooms.getCommand(), "BREAKROOMS");
    assertEquals(RESPONSE, breakrooms.getHeader(), "RESPONSE");
    assertEquals("anaconda", breakrooms.getUserid(), "USERID");
    assertEquals("schach", breakrooms.getRooms()[0], "schach");
    assertEquals("fussball", breakrooms.getRooms()[1], "fussball");
  }



  @Test
  public void decodeCALLPRIVATECHAT_CONFIRM() throws DecodeException {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"CALLPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"Jester\","
        + "\"REMOTE_SESSIONID\":\"8\"}"
        + "}";

    CALLPRIVATECHAT call = (CALLPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.CALLPRIVATECHAT, call.getCommand(), "command");
    assertEquals(CONFIRM, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("Jester", call.getRemoteNickname(), "remote Nickname");
    assertEquals(null, call.getSenderUID(), "sender uid");
    assertEquals("7", call.getLocalSessionid(), "remote Nickname");
    assertEquals("8", call.getRemoteSessionid(), "remote Nickname");
  }



  @Test
  public void decodeCALLPRIVATECHAT_HEADER() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CALLPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"anaconda\","
        + "\"REMOTE_SESSIONID\":\"7\","
        + "\"ERRORMESSAGE\":\"abc\"}"
        + "}";

    CALLPRIVATECHAT call = (CALLPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.CALLPRIVATECHAT, call.getCommand(), "command");
    assertEquals(ERROR, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("anaconda", call.getRemoteNickname(), "remote Nickname");
    assertEquals(null, call.getSenderUID(), "sender uid");
    assertEquals("7", call.getLocalSessionid(), "remote Nickname");
    assertEquals("7", call.getRemoteSessionid(), "remote Nickname");
    assertEquals("abc", call.getErrorMessage(), "errorMessage");
  }



  @Test
  public void decodeCALLPRIVATECHAT_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CALLPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"SENDER_UID\":\"hdh39678\","
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"REMOTE_NICKNAME\":\"jester\"}}";

    CALLPRIVATECHAT call = (CALLPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.CALLPRIVATECHAT, call.getCommand(), "command");
    assertEquals(REQUEST, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("jester", call.getRemoteNickname(), "remote Nickname");
    assertEquals("hdh39678", call.getSenderUID(), "sender uid");

  }



  @Test
  public void decodeCALLPRIVATECHAT_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CALLPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"Jester\","
        + "\"REMOTE_SESSIONID\":\"8\","
        + "\"TEXT\":\"Der Anwender nimmt keine Anrufe entgegen.\"}"
        + "}";

    CALLPRIVATECHAT call = (CALLPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.CALLPRIVATECHAT, call.getCommand(), "command");
    assertEquals(RESPONSE, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("Jester", call.getRemoteNickname(), "remote Nickname");
    assertEquals(null, call.getSenderUID(), "sender uid");
    assertEquals("7", call.getLocalSessionid(), "remote Nickname");
    assertEquals("8", call.getRemoteSessionid(), "remote Nickname");
    assertEquals("Der Anwender nimmt keine Anrufe entgegen.", call.getText(), "text");
  }



  @Test
  public void decodeCALLPRIVATECHAT_RESPONSE_MULTILINGUALKEY() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CALLPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"Jester\","
        + "\"REMOTE_SESSIONID\":\"8\","
        + "\"TEXT\":\"Der Anwender nimmt keine Anrufe entgegen.\","
        + "\"MULTILINGUALKEY\":\"Server_Lehnt_Chatanfragen_Ab\""
        + "}}";

    CALLPRIVATECHAT call = (CALLPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.CALLPRIVATECHAT, call.getCommand(), "command");
    assertEquals(RESPONSE, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("Jester", call.getRemoteNickname(), "remote Nickname");
    assertEquals(null, call.getSenderUID(), "sender uid");
    assertEquals("7", call.getLocalSessionid(), "remote Nickname");
    assertEquals("8", call.getRemoteSessionid(), "remote Nickname");
    assertEquals("Der Anwender nimmt keine Anrufe entgegen.", call.getText(), "text");
    assertEquals(KEY.SERVER_LEHNT_CHATANFRAGEN_AB, call.getMultilingualkey(), "MULTILINGUALKEY");
  }



  @Test
  public void decodeCALLPRIVATECHAT_ERROR_OHNE_ErrorMessage() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CALLPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"anaconda\","
        + "\"REMOTE_SESSIONID\":\"8\","
        + "\"TEXT\":\"einsame Seelen\","
        + "\"MULTILINGUALKEY\":\"String_Einsame_Seelen\""
        + "}}";

    CALLPRIVATECHAT call = (CALLPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.CALLPRIVATECHAT, call.getCommand(), "command");
    assertEquals(ERROR, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("anaconda", call.getRemoteNickname(), "remote Nickname");
    assertEquals(null, call.getSenderUID(), "sender uid");
    assertEquals("7", call.getLocalSessionid(), "remote Nickname");
    assertEquals("8", call.getRemoteSessionid(), "remote Nickname");
    assertEquals("einsame Seelen", call.getText(), "text");
    assertEquals(KEY.STRING_EINSAME_SEELEN, call.getMultilingualkey(), "MULTILINGUALKEY");
  }



  @Test
  public void decodeCALLPRIVATECHAT_RESPONSE_ohne_TEXT() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CALLPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"Jester\","
        + "\"REMOTE_SESSIONID\":\"8\""
        + "}}";

    CALLPRIVATECHAT call = (CALLPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.CALLPRIVATECHAT, call.getCommand(), "command");
    assertEquals(RESPONSE, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("Jester", call.getRemoteNickname(), "remote Nickname");
    assertEquals(null, call.getSenderUID(), "sender uid");
    assertEquals("7", call.getLocalSessionid(), "remote Nickname");
    assertEquals("8", call.getRemoteSessionid(), "remote Nickname");
    assertNull(call.getText(), "text");
  }



  @Test
  public void decodeCALLREMOTEUSER_CONFIRM() throws DecodeException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"CALLREMOTEUSER\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"Jester\","
        + "\"REMOTE_SESSIONID\":\"8\"}"
        + "}";

    CALLREMOTEUSER call = (CALLREMOTEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.CALLREMOTEUSER, call.getCommand(), "command");
    assertEquals(CONFIRM, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("Jester", call.getRemoteNickname(), "remote Nickname");
    assertEquals("8", call.getRemoteSessionid(), "remote sessionid");
    assertEquals("7", call.getLocalSessionid(), "local sessionid");

  }



  @Test
  public void decodeCALLREMOTEUSER_ERROR() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CALLREMOTEUSER\","
        + "\"DATASET\":{"
        + "\"ERRORMESSAGE\":\"The user is offline\"}"
        + "}";

    CALLREMOTEUSER call = (CALLREMOTEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.CALLREMOTEUSER, call.getCommand(), "command");
    assertEquals(ERROR, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("The user is offline", call.getErrorMessage(), "errorMessage");

  }



  @Test
  public void decodeCALLREMOTEUSER_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CALLREMOTEUSER\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"Jester\","
        + "\"REMOTE_SESSIONID\":\"8\"}"
        + "}";

    CALLREMOTEUSER call = (CALLREMOTEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.CALLREMOTEUSER, call.getCommand(), "command");
    assertEquals(REQUEST, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("Jester", call.getRemoteNickname(), "remote Nickname");
    assertEquals("8", call.getRemoteSessionid(), "remote sessionid");
    assertEquals("7", call.getLocalSessionid(), "local sessionid");

  }



  @Test
  public void decodeCALLREMOTEUSER_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CALLREMOTEUSER\","
        + "\"DATASET\":{"
        + "\"LOCAL_NICKNAME\":\"anaconda\","
        + "\"LOCAL_SESSIONID\":\"7\","
        + "\"REMOTE_NICKNAME\":\"Jester\","
        + "\"REMOTE_SESSIONID\":\"8\"}"
        + "}";

    CALLREMOTEUSER call = (CALLREMOTEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.CALLREMOTEUSER, call.getCommand(), "command");
    assertEquals(RESPONSE, call.getHeader(), "header");
    assertEquals(Protocol.DATASET, call.getDataset(), "dataset");
    assertEquals("anaconda", call.getLocalNickname(), "local Nickname");
    assertEquals("Jester", call.getRemoteNickname(), "remote Nickname");
    assertEquals("8", call.getRemoteSessionid(), "remote sessionid");
    assertEquals("7", call.getLocalSessionid(), "local sessionid");

  }



  @Test
  public void decodeCANDIDATETOPICMEMBER_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CANDIDATETOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\"}"
        + "}";

    CANDIDATETOPICMEMBER result = (CANDIDATETOPICMEMBER) messagedecoder.decode(encoded);
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("fussball", result.getRoom(), "room");

  }



  @Test
  public void decodeCANDIDATETOPICMEMBER_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CANDIDATETOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\","
        + "\"ROOMTYPE\":\"BESPRECHUNGSRAUM\","
        + "\"TOPICMEMBER\":["
        + "{\"USERID\":\"jester\",\"NICKNAME\":\"anaconda\"},"
        + "{\"USERID\":\"luis\",\"NICKNAME\":\"viper\""
        + "}]"
        + "}}";

    CANDIDATETOPICMEMBER result = (CANDIDATETOPICMEMBER) messagedecoder.decode(encoded);
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("fussball", result.getRoom(), "room");
    ChatUser[] chatuser = result.getChatUser();
    assertEquals(2, chatuser.length, "2 Datensätze");
    assertEquals("jester", chatuser[0].getUserid(), "userid0");
    assertEquals("anaconda", chatuser[0].getNickname(), "nickname0");
    assertEquals("luis", chatuser[1].getUserid(), "userid1");
    assertEquals("viper", chatuser[1].getNickname(), "nickname1");
    assertEquals(Roomtype.BESPRECHUNGSRAUM, result.getRoomtype(), "ROOMTYPE");

  }



  @Test
  public void decodeCHANGETOPICMEMBER_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CHANGETOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"Nexus\","
        + "\"ERRORMESSAGE\":\"Room <Nexus> does not exist.\""
        + "}}";

    CHANGETOPICMEMBER result = (CHANGETOPICMEMBER) messagedecoder.decode(encoded);
    assertEquals("12345678", result.getUserid(), "userid");
    assertEquals("Nexus", result.getRoom(), "room");
    assertEquals("Room <Nexus> does not exist.", result.getErrorMessage(), "errorMessage");

  }



  @Test
  public void decodeCHANGETOPICMEMBER_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHANGETOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\","
        + "\"ROOMTYPE\":\"BESPRECHUNGSRAUM\","
        + "\"TOPICMEMBER\":["
        + "{\"USERID\":\"jester\",\"NICKNAME\":\"anaconda\"},"
        + "{\"USERID\":\"luis\",\"NICKNAME\":\"viper\""
        + "}]"

        + "}}";

    CHANGETOPICMEMBER result = (CHANGETOPICMEMBER) messagedecoder.decode(encoded);
    ChatUser[] chatuser = result.getChatUser();
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("fussball", result.getRoom(), "room");
    assertEquals(2, chatuser.length, "chatuser");
    assertEquals("jester", chatuser[0].getUserid(), "userid1");
    assertEquals("anaconda", chatuser[0].getNickname(), "nickname1");
    assertEquals("luis", chatuser[1].getUserid(), "userid2");
    assertEquals("viper", chatuser[1].getNickname(), "nickname2");
    assertEquals(Roomtype.BESPRECHUNGSRAUM, result.getRoomtype(), "");

  }



  @Test
  public void decodeCHANGETOPICMEMBER_REQUEST_DecodeException() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHANGETOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\""
        + "}}";

    CHANGETOPICMEMBER result;
    try {
      result = (CHANGETOPICMEMBER) messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCHANGETOPICMEMBER_REQUEST_LEER() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHANGETOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\","
        + "\"ROOMTYPE\":\"BESPRECHUNGSRAUM\","
        + "\"TOPICMEMBER\":["
        + "]"

        + "}}";

    CHANGETOPICMEMBER result = (CHANGETOPICMEMBER) messagedecoder.decode(encoded);
    ChatUser[] chatuser = result.getChatUser();
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("fussball", result.getRoom(), "room");
    assertEquals(0, chatuser.length, "chatuser");
    assertEquals(Roomtype.BESPRECHUNGSRAUM, result.getRoomtype(), "");

  }



  @Test
  public void decodeCHANGETOPICMEMBER_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CHANGETOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\","
        + "\"ROOMTYPE\":\"BESPRECHUNGSRAUM\","
        + "\"TOPICMEMBER\":["
        + "{\"USERID\":\"jester\",\"NICKNAME\":\"anaconda\"},"
        + "{\"USERID\":\"luis\",\"NICKNAME\":\"viper\""
        + "}]"
        + "}}";

    CHANGETOPICMEMBER result = (CHANGETOPICMEMBER) messagedecoder.decode(encoded);
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("fussball", result.getRoom(), "room");
    ChatUser[] chatuser = result.getChatUser();
    assertEquals(2, chatuser.length, "2 Datensätze");
    assertEquals("jester", chatuser[0].getUserid(), "userid0");
    assertEquals("anaconda", chatuser[0].getNickname(), "nickname0");
    assertEquals("luis", chatuser[1].getUserid(), "userid1");
    assertEquals("viper", chatuser[1].getNickname(), "nickname1");
    assertEquals(Roomtype.BESPRECHUNGSRAUM, result.getRoomtype(), "");

  }



  @Test
  public void decodeCHATMESSAGE_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHATMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"MESSAGE\":\"Hallo meine Freunde\","
        + "\"CHATUSER\":{"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"FOREGROUND_COLOR\":-16777216,"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\"}}}";

    CHATMESSAGE chatmessage = (CHATMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.CHATMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "userid");
    assertEquals("room", chatmessage.getRoom(), "room");
    assertEquals(-16777216, chatuser.getForegroundColor(), "fgColor");
    assertEquals(-1, chatuser.getBackgroundColor(), "bgColor");
    assertEquals("anaconda", chatuser.getNickname(), "nickname");
    assertEquals("Hallo meine Freunde", chatmessage.getMessage(), "message");

  }



  @Test
  public void decodeCHATMESSAGE_REQUEST_2() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CHATMESSAGE\",\"DATASET\":{"
        + "\"ROOM\":\"Tester\","
        + "\"MESSAGE\":\"anhang\","
        + "\"ATTACHMENT\":29098336975847574,"
        + "\"FILENAME\":\"Top 5 Highest Win Rate Openings for White (2025).mp4\","
        + "\"FILESIZE\":67934764,"
        + "\"CHATUSER\":{\"BACKGROUND_COLOR\":-1772550,"
        + "\"FOREGROUND_COLOR\":-14607386,"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\"}}}";

    CHATMESSAGE chatmessage = (CHATMESSAGE) messagedecoder.decode(encoded);
    assertEquals(Command.CHATMESSAGE, chatmessage.getCommand(), "CHATMESSAGE");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "userid");
    assertEquals("Tester", chatmessage.getRoom(), "room");
    assertEquals(-14607386, chatuser.getForegroundColor(), "fgColor");
    assertEquals(-1772550, chatuser.getBackgroundColor(), "bgColor");
    assertEquals("anaconda", chatuser.getNickname(), "nickname");
    assertEquals("anhang", chatmessage.getMessage(), "message");

    assertEquals(29098336975847574l, chatmessage.getAttachment(), "attachment");
    assertEquals("Top 5 Highest Win Rate Openings for White (2025).mp4", chatmessage.getFilename());
    assertEquals(67934764l, chatmessage.getFilesize());

  }



  @Test
  public void decodeCHATMESSAGE_REQUEST_ATTACHMENT() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHATMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"MESSAGE\":\"Hallo meine Freunde\","
        + "\"ATTACHMENT\":33,"
        + "\"FILENAME\":\"name.txt\","
        + "\"FILESIZE\":1024,"
        + "\"CHATUSER\":{"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"FOREGROUND_COLOR\":-16777216,"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\"}}}";

    CHATMESSAGE chatmessage = (CHATMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.CHATMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "userid");
    assertEquals("room", chatmessage.getRoom(), "room");
    assertEquals(-16777216, chatuser.getForegroundColor(), "fgColor");
    assertEquals(-1, chatuser.getBackgroundColor(), "bgColor");
    assertEquals("anaconda", chatuser.getNickname(), "nickname");
    assertEquals("Hallo meine Freunde", chatmessage.getMessage(), "message");
    assertEquals(33, chatmessage.getAttachment(), "ATTACHMENT");
    assertEquals("name.txt", chatmessage.getFilename());
    assertEquals(1024, chatmessage.getFilesize());

  }



  @Test
  public void decodeCHATMESSAGE_REQUEST_UMLAUTE() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHATMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"MESSAGE\":\"äöüßéáúéíóú\","
        + "\"ATTACHMENT\":33,"
        + "\"FILENAME\":\"name.txt\","
        + "\"FILESIZE\":1024,"
        + "\"CHATUSER\":{"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"FOREGROUND_COLOR\":-16777216,"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\"}}}";

    CHATMESSAGE chatmessage = (CHATMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.CHATMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "userid");
    assertEquals("room", chatmessage.getRoom(), "room");
    assertEquals(-16777216, chatuser.getForegroundColor(), "fgColor");
    assertEquals(-1, chatuser.getBackgroundColor(), "bgColor");
    assertEquals("anaconda", chatuser.getNickname(), "nickname");
    assertEquals("äöüßéáúéíóú", chatmessage.getMessage(), "message");
    assertEquals(33, chatmessage.getAttachment(), "ATTACHMENT");
    assertEquals("name.txt", chatmessage.getFilename());
    assertEquals(1024, chatmessage.getFilesize());

  }



  @Test
  public void decodeCHATMESSAGE_REQUEST2() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHATMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"Everyone\","
        + "\"MESSAGE\":\"Hallo meine Freunde\","
        + "\"CHATUSER\":{\"BACKGROUND_COLOR\":-12975360,"
        + "\"FOREGROUND_COLOR\":-65536,"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"python\"}}}";

    CHATMESSAGE chatmessage = (CHATMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.CHATMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "userid");
    assertEquals("Everyone", chatmessage.getRoom(), "room");
    assertEquals(-65536, chatuser.getForegroundColor(), "fgColor");
    assertEquals(-12975360, chatuser.getBackgroundColor(), "bgColor");
    assertEquals("python", chatuser.getNickname(), "nickname");
    assertEquals("Hallo meine Freunde", chatmessage.getMessage(), "message");

  }



  @Test
  public void decodeCHATMESSAGE_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CHATMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"Everyone\","
        + "\"DATETIME\":9223372036854775807,"
        + "\"MESSAGE\":\"Hallo meine Freunde\","
        + "\"CHATUSER\":{\"BACKGROUND_COLOR\":-12975360,"
        + "\"FOREGROUND_COLOR\":-65536,"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"python\"}}}";

    CHATMESSAGE chatmessage = (CHATMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.CHATMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(RESPONSE, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "userid");
    assertEquals("Everyone", chatmessage.getRoom(), "room");
    assertEquals(-65536, chatuser.getForegroundColor(), "fgColor");
    assertEquals(-12975360, chatuser.getBackgroundColor(), "bgColor");
    assertEquals("python", chatuser.getNickname(), "nickname");
    assertEquals("Hallo meine Freunde", chatmessage.getMessage(), "message");
    assertEquals(Long.MAX_VALUE, chatmessage.getDatetime(), "datetime");

  }



  @Test
  public void decodeCHATMESSAGE_RESPONSE_ATTACHMENT() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CHATMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"Everyone\","
        + "\"DATETIME\":666,"
        + "\"MESSAGE\":\"Hallo meine Freunde\","
        + "\"ATTACHMENT\":99,"
        + "\"FILENAME\":\"name.txt\","
        + "\"FILESIZE\":"
        + String.valueOf(Long.MAX_VALUE)
        + ","
        + "\"CHATUSER\":{"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"FOREGROUND_COLOR\":-65536,"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"python\"}}}";

    CHATMESSAGE chatmessage = (CHATMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.CHATMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(RESPONSE, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "userid");
    assertEquals("Everyone", chatmessage.getRoom(), "room");
    assertEquals(-65536, chatuser.getForegroundColor(), "fgColor");
    assertEquals(-1, chatuser.getBackgroundColor(), "bgColor");
    assertEquals("python", chatuser.getNickname(), "nickname");
    assertEquals("Hallo meine Freunde", chatmessage.getMessage(), "message");
    assertEquals(666, chatmessage.getDatetime(), "datetime");
    assertEquals(Long.MAX_VALUE, chatmessage.getFilesize());
    assertEquals("name.txt", chatmessage.getFilename());
    assertEquals(99, chatmessage.getAttachment());

  }



  @Test
  public void decodeCHATMESSAGE_RESPONSE_Ohne_Time() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CHATMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"Everyone\","
        + "\"MESSAGE\":\"Hallo meine Freunde\","
        + "\"CHATUSER\":{\"BACKGROUND_COLOR\":-12975360,"
        + "\"FOREGROUND_COLOR\":-65536,"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"python\"}}}";

    CHATMESSAGE chatmessage = (CHATMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.CHATMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(RESPONSE, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "userid");
    assertEquals("Everyone", chatmessage.getRoom(), "room");
    assertEquals(-65536, chatuser.getForegroundColor(), "fgColor");
    assertEquals(-12975360, chatuser.getBackgroundColor(), "bgColor");
    assertEquals("python", chatuser.getNickname(), "nickname");
    assertEquals("Hallo meine Freunde", chatmessage.getMessage(), "message");
    assertNull(chatmessage.getDatetime(), "datetime");

  }



  @Test
  public void decodeCHATONLINELIST_REQUEST()
      throws DecodeException, JsonSyntaxException, UnsupportedEncodingException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHATONLINELIST\","
        + "\"DATASET\":{\"CHATONLINELIST\":["
        + "{\"ROOM\":\"Nexus\",\"ONLINE\":1},"
        + "{\"ROOM\":\"Everyone\",\"ONLINE\":2},"
        + "{\"ROOM\":\"Developer\",\"ONLINE\":2}]}}";

    CHATONLINELIST chatonline = (CHATONLINELIST) messagedecoder.decodeCHATONLINELIST(encoded);
    assertEquals(Command.CHATONLINELIST, chatonline.getCommand(), "command");
    assertEquals(REQUEST, chatonline.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatonline.getDataset(), "dataset");
    assertEquals(3, chatonline.getChatOnline().length, "3 Felder");
  }



  @Test
  public void decodeCHATONLINELIST_REQUEST_2() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CHATONLINELIST\","
        + "\"DATASET\":{"
        + "\"CHATONLINELIST\":["
        + "{\"ROOM\":\"everyone\",\"ONLINE\":99},"
        + "{\"ROOM\":\"ingress\",\"ONLINE\":7}"
        + "]}}";

    CHATONLINELIST onlinelist = (CHATONLINELIST) messagedecoder.decode(encoded);
    assertEquals(Command.CHATONLINELIST, onlinelist.getCommand(), "command");
    assertEquals(REQUEST, onlinelist.getHeader(), "command");
    assertEquals(Protocol.DATASET, onlinelist.getDataset(), "command");
    ChatOnline[] chatonline = onlinelist.getChatOnline();

    assertEquals("everyone", chatonline[0].getRoom(), "room_0");
    assertEquals(99, chatonline[0].getOnline(), "online_0");
    assertEquals("ingress", chatonline[1].getRoom(), "room_1");
    assertEquals(7, chatonline[1].getOnline(), "online_1");

  }



  @Test
  public void decodeCHATUSERLIST_ERROR() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\",\"COMMAND\":\"CHATUSERLIST\"}";

    CHATUSERLIST chatuseronline = (CHATUSERLIST) messagedecoder.decode(encoded);
    assertEquals(Command.CHATUSERLIST, chatuseronline.getCommand(), "command");
    assertEquals(ERROR, chatuseronline.getHeader(), "header");
  }



  @Test
  public void decodeCHATUSERLIST_ERROR_mitErrorCode() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\",\"COMMAND\":\"CHATUSERLIST\",\"DATASET\":{\"ERRORCODE\":-1,\"ERRORMESSAGE\":\"error\"}}";
    String bla = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CHATUSERLIST\","
        + "\"DATASET\":{\"ERRORCODE\":-1,\"ERRORMESSAGE\":\"error\"}}";

    CHATUSERLIST chatuseronline = (CHATUSERLIST) messagedecoder.decode(encoded);
    assertEquals(Command.CHATUSERLIST, chatuseronline.getCommand(), "command");
    assertEquals(ERROR, chatuseronline.getHeader(), "header");
    assertEquals(-1, chatuseronline.getErrorCode(), "errorcode");
    assertEquals("error", chatuseronline.getErrorMessage(), "errormessage");
  }



  @Test
  public void decodeCHATUSERLIST_REQEUST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CHATUSERLIST\",\"DATASET\":{\"ROOM\":\"Developer\"}}";

    CHATUSERLIST chatuseronline = (CHATUSERLIST) messagedecoder.decode(encoded);
    assertEquals(Command.CHATUSERLIST, chatuseronline.getCommand(), "command");
    assertEquals(REQUEST, chatuseronline.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatuseronline.getDataset(), "dataset");
    assertEquals("Developer", chatuseronline.getRoom(), "room");

  }



  @Test
  public void decodeCHATUSERLIST_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CHATUSERLIST\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"CHATUSERLIST\":["
        + "{\"USERID\":\"12345678\","
        + "\"FOREGROUND_COLOR\":-1,"
        + "\"BACKGROUND_COLOR\":-2,"
        + "\"NICKNAME\":\"louisxiv\"}]}}";

    CHATUSERLIST chatuseronline = (CHATUSERLIST) messagedecoder.decode(encoded);
    assertEquals(Command.CHATUSERLIST, chatuseronline.getCommand(), "command");
    assertEquals(RESPONSE, chatuseronline.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatuseronline.getDataset(), "dataset");
    assertEquals(1, chatuseronline.getChatUser().length, "1 Felder");
    assertEquals("louisxiv", chatuseronline.getChatUser()[0].getNickname(), "Nickname");
    assertEquals("12345678", chatuseronline.getChatUser()[0].getUserid(), "userid");
    assertEquals(-2, chatuseronline.getChatUser()[0].getBackgroundColor(), "bgColor");
    assertEquals(-1, chatuseronline.getChatUser()[0].getForegroundColor(), "fgColor");
    assertEquals("room", chatuseronline.getRoom(), "rootm");
  }



  @Test
  public void decodeCHATUSERLIST_RESPONSE_ohneUSERID() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CHATUSERLIST\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"CHATUSERLIST\":["
        + "{\"FOREGROUND_COLOR\":-1,\"BACKGROUND_COLOR\":-2,\"NICKNAME\":\"louisxiv\"}]}}";

    CHATUSERLIST chatuseronline = (CHATUSERLIST) messagedecoder.decode(encoded);
    assertEquals(Command.CHATUSERLIST, chatuseronline.getCommand(), "command");
    assertEquals(RESPONSE, chatuseronline.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatuseronline.getDataset(), "dataset");
    assertEquals(1, chatuseronline.getChatUser().length, "1 Felder");
    assertEquals("louisxiv", chatuseronline.getChatUser()[0].getNickname(), "Nickname");
    assertEquals(null, chatuseronline.getChatUser()[0].getUserid(), "userid,optional");
    assertEquals(-2, chatuseronline.getChatUser()[0].getBackgroundColor(), "bgColor");
    assertEquals(-1, chatuseronline.getChatUser()[0].getForegroundColor(), "fgColor");
    assertEquals("room", chatuseronline.getRoom(), "room");
  }



  @Test
  public void decodeCONFERENCE_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\","
        + "\"ORGANISATOR_UID\":\"anaconda\","
        + "\"USERID\":\"12345678\","
        + "\"SESSION\":\"3a\","
        + "\"OFFHOOK\":true"
        + "}}";

    CONFERENCE conference = (CONFERENCE) messagedecoder.decode(encoded);
    assertEquals(Command.CONFERENCE, conference.getCommand(), "COMMAND");
    assertEquals(HEADER.REQUEST, conference.getHeader(), "header");
    assertEquals(Protocol.DATASET, conference.getDataset(), "dataset");

    assertEquals("Dienstag Treff", conference.getKonferenzname());
    assertEquals("anaconda", conference.getOrganisatorUid());
    assertEquals("12345678", conference.getUserid());
    assertEquals("3a", conference.getSession());
    assertEquals(true, conference.isOffhook());
  }



  @Test
  public void decodeCONFERENCE_REQUEST_ohne_USERID() {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\","
        + "\"ORGANISATOR_UID\":\"anaconda\","
        + "\"SESSION\":\"3a\","
        + "\"OFFHOOK\":true"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, USERID fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCONFERENCE_REQUEST_ohne_SESSION() {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\","
        + "\"ORGANISATOR_UID\":\"anaconda\","
        + "\"USERID\":\"12345678\","
        + "\"OFFHOOK\":true"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, SESSION fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCONFERENCE_REQUEST_ERROR() {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\""
        + "}}";

    try {
      CONFERENCE conference = (CONFERENCE) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCONFERENCE_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\","
        + "\"ORGANISATOR_UID\":\"anaconda\","
        + "\"USERID\":\"admin678\","
        + "\"OFFHOOK\":true,"
        + "\"KONFERENZRAUMUSER\":[{\"ANWESEND\":true,\"SESSION\":\"ab\",\"MUTE\":true,\"CHATUSER\":{\"USERID\":\"userid1\",\"NICKNAME\":\"admin678\",\"FOREGROUND_COLOR\":-2839,\"BACKGROUND_COLOR\":238}},"
        + "{\"ANWESEND\":true,\"SESSION\":\"cd\",\"MUTE\":true,\"CHATUSER\":{\"USERID\":\"userid2\",\"NICKNAME\":\"freigeist\",\"FOREGROUND_COLOR\":-9283,\"BACKGROUND_COLOR\":-293}},"
        + "{\"ANWESEND\":false,\"MUTE\":false,\"CHATUSER\":{\"USERID\":\"userid3\",\"NICKNAME\":\"Jester\",\"FOREGROUND_COLOR\":-7263,\"BACKGROUND_COLOR\":-9283}},"
        + "{\"ANWESEND\":false,\"MUTE\":false,\"CHATUSER\":{\"USERID\":\"userid4\",\"NICKNAME\":\"picasso\",\"FOREGROUND_COLOR\":-8273,\"BACKGROUND_COLOR\":-23}}]"

        + "}}";

    CONFERENCE conference = (CONFERENCE) messagedecoder.decode(encoded);
    assertEquals(Command.CONFERENCE, conference.getCommand(), "COMMAND");
    assertEquals(HEADER.RESPONSE, conference.getHeader(), "header");
    assertEquals(Protocol.DATASET, conference.getDataset(), "dataset");

    assertEquals("Dienstag Treff", conference.getKonferenzname());
    assertEquals("anaconda", conference.getOrganisatorUid());
    KonferenzraumUser[] konfernzraumuser = conference.getKonferenzraumUser();
    assertEquals(-2839, konfernzraumuser[0].getForegroundColor());
    assertEquals(238, konfernzraumuser[0].getBackgroundColor());
    assertEquals("admin678", konfernzraumuser[0].getNickname());
    assertEquals("userid1", konfernzraumuser[0].getUserid());
    assertTrue(konfernzraumuser[0].isAnwesend());
    assertEquals("ab", konfernzraumuser[0].getSession());
    assertTrue(konfernzraumuser[0].isMute());
    assertEquals("admin678", conference.getUserid());
    assertTrue(conference.isOffhook());

  }



  @Test
  public void decodeCONFERENCE_RESPONSE_ohne_OFFHOOK() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\","
        + "\"ORGANISATOR_UID\":\"anaconda\","
        + "\"USERID\":\"admin678\","
        + "\"KONFERENZRAUMUSER\":[{\"ANWESEND\":true,\"SESSION\":\"ab\",\"CHATUSER\":{\"USERID\":\"userid1\",\"NICKNAME\":\"admin678\",\"FOREGROUND_COLOR\":-2839,\"BACKGROUND_COLOR\":238}},"
        + "{\"ANWESEND\":true,\"SESSION\":\"cd\",\"CHATUSER\":{\"USERID\":\"userid2\",\"NICKNAME\":\"freigeist\",\"FOREGROUND_COLOR\":-9283,\"BACKGROUND_COLOR\":-293}},"
        + "{\"ANWESEND\":false,\"CHATUSER\":{\"USERID\":\"userid3\",\"NICKNAME\":\"Jester\",\"FOREGROUND_COLOR\":-7263,\"BACKGROUND_COLOR\":-9283}},"
        + "{\"ANWESEND\":false,\"CHATUSER\":{\"USERID\":\"userid4\",\"NICKNAME\":\"picasso\",\"FOREGROUND_COLOR\":-8273,\"BACKGROUND_COLOR\":-23}}]"

        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, OFFHOOK fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeCONFERENCE_RESPONSE_JERSEY() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\","
        + "\"ORGANISATOR_UID\":\"anaconda\","
        + "\"KONFERENZRAUMUSER\":[{\"ANWESEND\":true,\"SESSION\":\"2b\",\"CHATUSER\":{\"USERID\":\"userid1\",\"NICKNAME\":\"admin678\",\"FOREGROUND_COLOR\":-2839,\"BACKGROUND_COLOR\":238}},"
        + "{\"ANWESEND\":true,\"SESSION\":\"1a\",\"CHATUSER\":{\"USERID\":\"userid2\",\"NICKNAME\":\"freigeist\",\"FOREGROUND_COLOR\":-9283,\"BACKGROUND_COLOR\":-293}},"
        + "{\"ANWESEND\":false,\"CHATUSER\":{\"USERID\":\"userid3\",\"NICKNAME\":\"Jester\",\"FOREGROUND_COLOR\":-7263,\"BACKGROUND_COLOR\":-9283}},"
        + "{\"ANWESEND\":false,\"CHATUSER\":{\"USERID\":\"userid4\",\"NICKNAME\":\"picasso\",\"FOREGROUND_COLOR\":-8273,\"BACKGROUND_COLOR\":-23}}]"

        + "}}";

    GsonBuilder builder = new GsonBuilder();
    builder.registerTypeAdapter(CONFERENCE.class, new ConferenceDeserializer());
    Gson gson = builder.create();
    CONFERENCE conference = gson.fromJson(encoded, CONFERENCE.class);

    assertEquals(Command.CONFERENCE, conference.getCommand(), "COMMAND");
    assertEquals(HEADER.RESPONSE, conference.getHeader(), "header");
    assertEquals(Protocol.DATASET, conference.getDataset(), "dataset");

    assertEquals("Dienstag Treff", conference.getKonferenzname());
    assertEquals("anaconda", conference.getOrganisatorUid());
    KonferenzraumUser[] konfernzraumuser = conference.getKonferenzraumUser();
    assertEquals(-2839, konfernzraumuser[0].getForegroundColor());
    assertEquals(238, konfernzraumuser[0].getBackgroundColor());
    assertEquals("admin678", konfernzraumuser[0].getNickname());
    assertEquals("userid1", konfernzraumuser[0].getUserid());
    assertEquals("2b", konfernzraumuser[0].getSession());
    assertTrue(konfernzraumuser[0].isAnwesend());

  }



  @Test
  public void decodeCONFERENCE_RESPONSE_OHNE_KONFERENZRAUMUSER() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\","
        + "\"ORGANISATOR_UID\":\"anaconda\","
        + "\"USERID\":\"admin678\","
        + "\"OFFHOOK\":true,"
        + "\"KONFERENZRAUMUSER\":[]"
        + "}}";

    CONFERENCE conference = (CONFERENCE) messagedecoder.decode(encoded);
    assertEquals(Command.CONFERENCE, conference.getCommand(), "COMMAND");
    assertEquals(HEADER.RESPONSE, conference.getHeader(), "header");
    assertEquals(Protocol.DATASET, conference.getDataset(), "dataset");

    assertEquals("Dienstag Treff", conference.getKonferenzname());
    assertEquals("anaconda", conference.getOrganisatorUid());
    KonferenzraumUser[] konfernzraumuser = conference.getKonferenzraumUser();
    assertNotNull(konfernzraumuser);
    assertEquals(0, konfernzraumuser.length);
    assertTrue(conference.isOffhook());
  }



  @Test
  public void decodeCONFERENCE_RESPONSE_OHNE_KONFERENZNAME() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"ORGANISATOR_UID\":\"anaconda\","
        + "\"KONFERENZRAUMUSER\":[]"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, KONFERENZNAME fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCONFERENCE_RESPONSE_OHNE_ORGANISATOR() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CONFERENCE\","
        + "\"DATASET\":{"
        + "\"KONFERENZNAME\":\"Dienstag Treff\","
        + "\"KONFERENZRAUMUSER\":[]"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, ORGANISATOR fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCONFERENCEAUDIO() throws DecodeException {
//    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CONFERENCEAUDIO\","
//        + "\"DATASET\":{"
//        + "\"TARGET\":[\"ab\",\"a1\"],"
//        + "\"CONTENT\":\"ich bin ein base64 textstring\""
//        + "}}";

    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEAUDIO\"," + "\"DATASET\":{"
        + "\"TARGET\":[\"ab\",\"a1\"],"
        + "\"CONTENT\":[105,99,104,32,98,105,110,32,101,105,110,32,98,97,115,101,54,52,32,116,101,120,116,115,116,114,105,110,103]}}";

    CONFERENCEAUDIO conferenceaudio = (CONFERENCEAUDIO) messagedecoder.decode(encoded);
    assertEquals(Command.CONFERENCEAUDIO, conferenceaudio.getCommand(), "COMMAND");
    assertEquals(HEADER.REQUEST, conferenceaudio.getHeader(), "header");
    assertEquals(Protocol.DATASET, conferenceaudio.getDataset(), "dataset");

//    Arrays.equals(null, null)

    assertTrue(
        Arrays.equals(
            new byte[] {105, 99, 104, 32, 98, 105, 110, 32, 101, 105, 110, 32, 98, 97, 115, 101, 54, 52, 32,
                116, 101, 120, 116, 115, 116, 114, 105, 110, 103},
            conferenceaudio.getContent()
        ), "ich bin ein base64 textstring"
    );

    String[] target = conferenceaudio.getReceiverSessions();
    assertEquals("ab", target[0]);
    assertEquals("a1", target[1]);

  }



  @Test
  public void decodeCONFERENCEAUDIO_MIT_USERID() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CONFERENCEAUDIO\","
        + "\"DATASET\":{"
        + "\"USERID\":\"admin678\","
        + "\"TARGET\":[\"ab\",\"a1\"],"
        + "\"CONTENT\":[105,99,104,32,98,105,110,32,101,105,110,32,98,97,115,101,54,52,32,116,101,120,116,115,116,114,105,110,103]}}";

    CONFERENCEAUDIO conferenceaudio = (CONFERENCEAUDIO) messagedecoder.decode(encoded);
    assertEquals(Command.CONFERENCEAUDIO, conferenceaudio.getCommand(), "COMMAND");
    assertEquals(HEADER.REQUEST, conferenceaudio.getHeader(), "header");
    assertEquals(Protocol.DATASET, conferenceaudio.getDataset(), "dataset");

//    assertEquals("ich bin ein base64 textstring".getBytes(), conferenceaudio.getContent());
    assertTrue(
        Arrays.equals(
            new byte[] {105, 99, 104, 32, 98, 105, 110, 32, 101, 105, 110, 32, 98, 97, 115, 101, 54, 52, 32,
                116, 101, 120, 116, 115, 116, 114, 105, 110, 103},
            conferenceaudio.getContent()
        )
    );
    String[] target = conferenceaudio.getReceiverSessions();
    assertEquals("ab", target[0]);
    assertEquals("a1", target[1]);
    assertEquals("admin678", conferenceaudio.getUserid());

  }



  @Test
  public void decodeCONFERENCEAUDIO_MIT_ORIGIN() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CONFERENCEAUDIO\","
        + "\"DATASET\":{"
        + "\"TARGET\":[\"ab\",\"a1\"],"
        + "\"CONTENT\":[105,99,104,32,98,105,110,32,101,105,110,32,98,97,115,101,54,52,32,116,101,120,116,115,116,114,105,110,103]}}";

    CONFERENCEAUDIO conferenceaudio = (CONFERENCEAUDIO) messagedecoder.decode(encoded);
    assertEquals(Command.CONFERENCEAUDIO, conferenceaudio.getCommand(), "COMMAND");
    assertEquals(HEADER.REQUEST, conferenceaudio.getHeader(), "header");
    assertEquals(Protocol.DATASET, conferenceaudio.getDataset(), "dataset");

//    assertEquals("ich bin ein base64 textstring".getBytes(), conferenceaudio.getContent());
    assertTrue(
        Arrays.equals(
            new byte[] {105, 99, 104, 32, 98, 105, 110, 32, 101, 105, 110, 32, 98, 97, 115, 101, 54, 52, 32,
                116, 101, 120, 116, 115, 116, 114, 105, 110, 103},
            conferenceaudio.getContent()
        )
    );
    String[] target = conferenceaudio.getReceiverSessions();
    assertEquals("ab", target[0]);
    assertEquals("a1", target[1]);

  }



  @Test
  public void decodeCONFERENCEAUDIO_TARGET_NULL() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CONFERENCEAUDIO\","
        + "\"DATASET\":{"
        + "\"CONTENT\":[105,99,104,32,98,105,110,32,101,105,110,32,98,97,115,101,54,52,32,116,101,120,116,115,116,114,105,110,103]"
        + "}}";

    try {
      CONFERENCEAUDIO conferenceaudio = (CONFERENCEAUDIO) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeCONFERENCEMUTE_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEMUTE\"," + "\"DATASET\":{"
        + "\"KONFERENZRAUM\":\"Dienstag Treff\","
        + "\"ORGANISATOR\":\"admin678\","
        + "\"USERID\":\"max45678\","
        + "\"RECEIVER_SESSIONS\":[\"ab\",\"cd\"],"
        + "\"MUTE\":true"
        + "}}";

    CONFERENCEMUTE mute = (CONFERENCEMUTE) messagedecoder.decode(encoded);
    assertEquals(REQUEST, mute.getHeader(), "REQUEST");
    assertEquals(Command.CONFERENCEMUTE, mute.getCommand(), "COMMAND");
    assertEquals(Protocol.DATASET, mute.getDataset(), "DATASET");
    assertEquals("Dienstag Treff", mute.getKonferenzraum(), "KONFERENZRAUM");
    assertEquals("admin678", mute.getOrganisator(), "ORGANISATOR");
    assertEquals("max45678", mute.getUserid(), "USERID");
    assertEquals("ab", mute.getReceiverSessions()[0], "RECEIVER_SESSIONS");
    assertEquals("cd", mute.getReceiverSessions()[1], "RECEIVER_SESSIONS");
    assertEquals(Boolean.TRUE, mute.isMute(), "MUTE");

  }



  @Test
  public void decodeCONFERENCEMUTE_REQUEST_ohne_USERID() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEMUTE\"," + "\"DATASET\":{"
        + "\"KONFERENZRAUM\":\"Dienstag Treff\","
        + "\"ORGANISATOR\":\"admin678\","
        + "\"RECEIVER_SESSIONS\":[\"ab\",\"cd\"],"
        + "\"MUTE\":true"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, USERID fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeCONFERENCEMUTE_REQUEST_ohne_ORGANISATOR() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEMUTE\"," + "\"DATASET\":{"
        + "\"KONFERENZRAUM\":\"Dienstag Treff\","
        + "\"USERID\":\"max45678\","
        + "\"RECEIVER_SESSIONS\":[\"ab\",\"cd\"],"
        + "\"MUTE\":true"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, ORGANISATOR fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeCONFERENCEMUTE_REQUEST_ohne_KONFERENZRAUM() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEMUTE\"," + "\"DATASET\":{"
        + "\"ORGANISATOR\":\"admin678\","
        + "\"USERID\":\"max45678\","
        + "\"RECEIVER_SESSIONS\":[\"ab\",\"cd\"],"
        + "\"MUTE\":true"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, KONFERENZRAUM fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeCONFERENCEMUTE_REQUEST_ohne_MUTE() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEMUTE\"," + "\"DATASET\":{"
        + "\"KONFERENZRAUM\":\"Dienstag Treff\","
        + "\"ORGANISATOR\":\"admin678\","
        + "\"USERID\":\"max45678\","
        + "\"RECEIVER_SESSIONS\":[\"ab\",\"cd\"]"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, MUTE fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCONFERENCEMUTE_REQUEST_ohne_RECEIVER_SESSIONS() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEMUTE\"," + "\"DATASET\":{"
        + "\"KONFERENZRAUM\":\"Dienstag Treff\","
        + "\"ORGANISATOR\":\"admin678\","
        + "\"USERID\":\"max45678\","
        + "\"MUTE\":true"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, RECEIVER_SESSIONS() fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCONFERENCEVIDEO() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEVIDEO\"," + "\"DATASET\":{"
        + "\"KONFERENZRAUM\":\"Dienstag Treff\","
        + "\"ORGANISATOR\":\"max45678\","
        + "\"RECEIVER_SESSIONS\":[\"ab\",\"a1\"],"
        + "\"SHARING\":true"
        + "}}";

    CONFERENCEVIDEO conferenceVideo = (CONFERENCEVIDEO) messagedecoder.decode(encoded);
    assertEquals(Command.CONFERENCEVIDEO, conferenceVideo.getCommand(), "COMMAND");
    assertEquals(HEADER.REQUEST, conferenceVideo.getHeader(), "header");
    assertEquals(Protocol.DATASET, conferenceVideo.getDataset(), "dataset");

    String[] receiver = conferenceVideo.getReceiverSessions();
    assertEquals("ab", receiver[0]);
    assertEquals("a1", receiver[1]);

    assertTrue(conferenceVideo.isShared(), "SHARING");
    assertEquals("max45678", conferenceVideo.getOrganisator());
    assertEquals("Dienstag Treff", conferenceVideo.getKonferenzraum());

  }



  @Test
  @DisplayName(value = "CONFERENCEVIDEO ohne ORGANISATOR")
  public void decodeCONFERENCEVIDEO_ohne_ORGANISATOR() {

    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEVIDEO\"," + "\"DATASET\":{"
        + "\"RECEIVER_SESSIONS\":[\"ab\",\"a1\"],"
        + "\"CONNECT\":\"true\""
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, ORGANISATOR fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  @DisplayName(value = "CONFERENCEVIDEO ohne RECEIVER_SESSIONS")
  public void decodeCONFERENCEVIDEO_ohne_Receiver() {

    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEVIDEO\"," + "\"DATASET\":{"
        + "\"ORGANISATOR\":\"max45678\","
        + "\"CONNECT\":\"true\""
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, RECEIVER_SESSIONS fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  @DisplayName(value = "CONFERENCEVIDEO ohne CONNECT")
  public void decodeCONFERENCEVIDEO_ohne_CONNCET() {

    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"CONFERENCEVIDEO\"," + "\"DATASET\":{"
        + "\"ORGANISATOR\":\"max45678\","
        + "\"RECEIVER_SESSIONS\":[\"ab\",\"a1\"]"
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, CONNECT fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeCREATETEMPROOM_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ROOMTYPE\":\"BESPRECHUNGSRAUM\","
        + "\"ERRORCODE\":1062,"
        + "\"ERRORMESSAGE\":\"duplicate key entry\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(ERROR, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals(1062, temproom.getErrorCode(), "errorcode");
    assertEquals("duplicate key entry", temproom.getErrorMessage(), "user");
  }



  @Test
  public void decodeCREATETEMPROOM_ERROR_mit_ROOMTYPE() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ROOMTYPE\":\"PAUSENRAUM\","
        + "\"ERRORCODE\":1062,"
        + "\"ERRORMESSAGE\":\"duplicate key entry\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(ERROR, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals("PAUSENRAUM", temproom.getRoomtype().name(), "ROOMTYPE");
    assertEquals(1062, temproom.getErrorCode(), "errorcode");
    assertEquals("duplicate key entry", temproom.getErrorMessage(), "user");
  }



  @Test
  public void decodeCREATETEMPROOM_ERROR_MULTILINGUALKEY() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ROOMTYPE\":\"BESPRECHUNGSRAUM\","
        + "\"ERRORCODE\":1062,"
        + "\"ERRORMESSAGE\":\"duplicate key entry\","
        + "\"MULTILINGUALKEY\":\"Server_Zu_kurzer_Raumname\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(ERROR, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals(1062, temproom.getErrorCode(), "errorcode");
    assertEquals("duplicate key entry", temproom.getErrorMessage(), "user");
    assertEquals(KEY.SERVER_ZU_KURZER_RAUMNAME, temproom.getMultilingualkey(), "SERVER_KURZER_RAUMNAME");
  }



  @Test
  public void decodeCREATETEMPROOM_ERROR_PROTECTED() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"louisxiv\","
        + "\"ROOM\":\"anaconda\","
        + "\"ENTRY\":\"PROTECTED\","
        + "\"LIFETIME\":\"TEMPORARY\","
        + "\"ROOMTYPE\":\"GRUPPENRAUM\","
        + "\"ERRORMESSAGE\":\"The password is missing.\"}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(ERROR, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("louisxiv", temproom.getUser(), "user");
    assertEquals("The password is missing.", temproom.getErrorMessage(), "errorMessage");
    assertEquals(Entry.PROTECTED, temproom.getEntry(), "entry");
    assertEquals(Lifetime.TEMPORARY, temproom.getLifetime(), "lifetime");
    assertNull(temproom.getErrorCode(), "errorcode");
  }



  @Test
  public void decodeCREATETEMPROOM_ERROR_TEXT() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ROOMTYPE\":\"FORUM\","
        + "\"ERRORCODE\":1062,"
        + "\"ERRORMESSAGE\":\"duplicate key entry\","
        + "\"TEXT\":\"fehler\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(ERROR, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals(1062, temproom.getErrorCode(), "errorcode");
    assertEquals("duplicate key entry", temproom.getErrorMessage(), "user");
    assertEquals("fehler", temproom.getText(), "text");
  }



  @Test
  public void decodeCREATETEMPROOM_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"NICKNAME\":\"viper\","
        + "\"ROOM\":\"anaconda\","
        + "\"ENTRY\":\"PUBLIC\","
        + "\"LIFETIME\":\"TEMPORARY\","
        + "\"ROOMTYPE\":\"FORUM\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(REQUEST, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals(Entry.PUBLIC, temproom.getEntry(), "PUBLIC");
    assertEquals(Lifetime.TEMPORARY, temproom.getLifetime(), "TEMPORARY");
    assertEquals("viper", temproom.getNickname(), "nickname");
    assertEquals("FORUM", temproom.getRoomtype().name(), "ROOMTYPE");
  }



  @Test
  public void decodeCREATETEMPROOM_REQUEST_DecodeException() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ENTRY\":\"PUBLIC\","
        + "\"LIFETIME\":\"TEMPORARY\""
        + "}}";

    CREATETEMPROOM temproom;
    try {
      temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, NICKNAME");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeCREATETEMPROOM_REQUEST_PROTECTED_temporary() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"NICKNAME\":\"viper\","
        + "\"ROOM\":\"anaconda\","
        + "\"ENTRY\":\"PROTECTED\","
        + "\"LIFETIME\":\"TEMPORARY\","
        + "\"PASSWORD\":\"geheim\","
        + "\"ROOMTYPE\":\"PAUSENRAUM\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(REQUEST, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals("PROTECTED", temproom.getEntry().name());
    assertEquals("TEMPORARY", temproom.getLifetime().name(), "TEMPORARY");
    assertEquals("geheim", temproom.getPassword(), "PASSWORD");
    assertEquals("viper", temproom.getNickname(), "NICKNAME");
    assertEquals("PAUSENRAUM", temproom.getRoomtype().name(), "PAUSENRAUM");
  }



  @Test
  public void decodeCREATETEMPROOM_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ENTRY\":\"PUBLIC\","
        + "\"LIFETIME\":\"TEMPORARY\","
        + "\"ROOMTYPE\":\"PAUSENRAUM\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(RESPONSE, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals("PUBLIC", temproom.getEntry().name(), "entry");
    assertEquals("TEMPORARY", temproom.getLifetime().name(), "lifetime");
    assertEquals("PAUSENRAUM", temproom.getRoomtype().name(), "ROOMTYPE");
  }



  @Test
  public void decodeCREATETEMPROOM_RESPONSE_MULTILINGUALKEY() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ENTRY\":\"PUBLIC\","
        + "\"LIFETIME\":\"TEMPORARY\","
        + "\"ROOMTYPE\":\"PAUSENRAUM\","
        + "\"MULTILINGUALKEY\":\"Server_Der_Chatraum_wurde_angelegt\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(RESPONSE, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals("PUBLIC", temproom.getEntry().name(), "entry");
    assertEquals("TEMPORARY", temproom.getLifetime().name(), "lifetime");
    assertEquals("PAUSENRAUM", temproom.getRoomtype().name(), "ROOMTYPE");
    assertEquals(KEY.SERVER_DER_CHATRAUM_WURDE_ANGELEGT, temproom.getMultilingualkey(), "MULTILINGUALKEY");
  }



  @Test
  public void decodeCREATETEMPROOM_RESPONSE_PROTECTED_temporary() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ENTRY\":\"PROTECTED\","
        + "\"LIFETIME\":\"TEMPORARY\","
        + "\"PASSWORD\":\"geheim\","
        + "\"ROOMTYPE\":\"PAUSENRAUM\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(RESPONSE, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals(Entry.PROTECTED, temproom.getEntry(), "PROTECTED");
    assertEquals("TEMPORARY", temproom.getLifetime().name(), "TEMPORARY");
    assertEquals("geheim", temproom.getPassword(), "PASSWORD");
    assertEquals("PAUSENRAUM", temproom.getRoomtype().name(), "PAUSENRAUM");
  }



  @Test
  public void decodeCREATETEMPROOM_RESPONSE_TEXT() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"CREATETEMPROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"llange\","
        + "\"ROOM\":\"anaconda\","
        + "\"ENTRY\":\"PUBLIC\","
        + "\"LIFETIME\":\"TEMPORARY\","
        + "\"ROOMTYPE\":\"PAUSENRAUM\","
        + "\"TEXT\":\"Hallo\""
        + "}}";

    CREATETEMPROOM temproom = (CREATETEMPROOM) messagedecoder.decode(encoded);
    assertEquals(Command.CREATETEMPROOM, temproom.getCommand(), "command");
    assertEquals(RESPONSE, temproom.getHeader(), "header");
    assertEquals(Protocol.DATASET, temproom.getDataset(), "dataset");
    assertEquals("anaconda", temproom.getRoom(), "room");
    assertEquals("llange", temproom.getUser(), "user");
    assertEquals("PUBLIC", temproom.getEntry().name(), "entry");
    assertEquals("TEMPORARY", temproom.getLifetime().name(), "lifetime");
    assertEquals("Hallo", temproom.getText(), "text");
    assertEquals("PAUSENRAUM", temproom.getRoomtype().name(), "ROOMTYPE");
  }



  @Test
  public void decodeDELETEROOM_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"DELETEROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"louisxiv\","
        + "\"ROOM\":\"nexus\","
        + "\"TEXT\":\"the room does not exist.\""
        + "}}";

    DELETEROOM delete = (DELETEROOM) messagedecoder.decode(encoded);
    assertEquals(Command.DELETEROOM, delete.getCommand(), "command");
    assertEquals(ERROR, delete.getHeader(), "header");
    assertEquals("louisxiv", delete.getUserid(), "user");
    assertEquals("nexus", delete.getRoom(), "room");
    assertEquals("the room does not exist.", delete.getText(), "text");

  }



  @Test
  public void decodeDELETEROOM_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"DELETEROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"louisxiv\","
        + "\"ROOM\":\"nexus\""
        + "}}";

    DELETEROOM delete = (DELETEROOM) messagedecoder.decode(encoded);
    assertEquals(Command.DELETEROOM, delete.getCommand(), "command");
    assertEquals(REQUEST, delete.getHeader(), "header");
    assertEquals("louisxiv", delete.getUserid(), "user");
    assertEquals("nexus", delete.getRoom(), "room");

  }



  @Test
  public void decodeDELETEROOM_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DELETEROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"louisxiv\","
        + "\"ROOM\":\"nexus\""
        + "}}";

    DELETEROOM delete = (DELETEROOM) messagedecoder.decode(encoded);
    assertEquals(Command.DELETEROOM, delete.getCommand(), "command");
    assertEquals(RESPONSE, delete.getHeader(), "header");
    assertEquals("louisxiv", delete.getUserid(), "user");
    assertEquals("nexus", delete.getRoom(), "room");

  }



  @Test
  public void decodeDELETEROOM_RESPONSE_TEXT() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DELETEROOM\","
        + "\"DATASET\":{"
        + "\"USER\":\"louisxiv\","
        + "\"ROOM\":\"nexus\","
        + "\"TEXT\":\"bla\""
        + "}}";

    DELETEROOM delete = (DELETEROOM) messagedecoder.decode(encoded);
    assertEquals(Command.DELETEROOM, delete.getCommand(), "command");
    assertEquals(RESPONSE, delete.getHeader(), "header");
    assertEquals("louisxiv", delete.getUserid(), "user");
    assertEquals("nexus", delete.getRoom(), "room");
    assertEquals("bla", delete.getText(), "text");

  }



  @Test
  public void decodeDELETEUPLOADFILES_CONFIRM() throws DecodeException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"DELETEUPLOADFILES\"" + "}";

    DELETEUPLOADFILES deletefiles = (DELETEUPLOADFILES) messagedecoder.decode(encoded);

    assertEquals(Command.DELETEUPLOADFILES, deletefiles.getCommand(), "command");
    assertEquals(CONFIRM, deletefiles.getHeader(), "header");
  }



  @Test
  public void decodeDELETEUPLOADFILES_ERROR() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"DELETEUPLOADFILES\","
        + "\"DATASET\":{"
        + "\"TEXT\":\"A database error has occured\","
        + "\"ERRORCODE\":34,"
        + "\"ERRORMESSAGE\":\"fehler\""
        + "}}";

    DELETEUPLOADFILES deletefiles = (DELETEUPLOADFILES) messagedecoder.decode(encoded);

    assertEquals(Command.DELETEUPLOADFILES, deletefiles.getCommand(), "command");
    assertEquals(ERROR, deletefiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, deletefiles.getDataset(), "dataset");
    assertEquals(34, deletefiles.getErrorCode(), "errorcode");
    assertEquals("fehler", deletefiles.getErrorMessage(), "message");
    assertEquals("A database error has occured", deletefiles.getText(), "text");
  }



  @Test
  public void decodeDELETEUPLOADFILES_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"DELETEUPLOADFILES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"192.168.1.41\""
        + "}}";

    DELETEUPLOADFILES deletefiles = (DELETEUPLOADFILES) messagedecoder.decode(encoded);

    assertEquals(Command.DELETEUPLOADFILES, deletefiles.getCommand(), "command");
    assertEquals(REQUEST, deletefiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, deletefiles.getDataset(), "dataset");
    assertEquals("louisxiv", deletefiles.getUserid(), "userid");
    assertEquals("192.168.1.41", deletefiles.getIp(), "ip");
  }



  @Test
  public void decodeDELETEUSER_CONFIRM() throws DecodeException {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"DELETE_USER\"" + "}";

    DELETEUSER deleteuser = (DELETEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.DELETE_USER, deleteuser.getCommand(), "command");
    assertEquals(CONFIRM, deleteuser.getHeader(), "header");
  }



  @Test
  public void decodeDELETEUSER_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"DELETE_USER\"" + "}";

    DELETEUSER deleteuser = (DELETEUSER) messagedecoder.decode(encoded);
    assertEquals(Command.DELETE_USER, deleteuser.getCommand(), "command");
    assertEquals(ERROR, deleteuser.getHeader(), "header");

  }



  @Test
  public void decodeDELETEUSER_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"DELETE_USER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\""
        + "}}";

    DELETEUSER deleteuser = (DELETEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.DELETE_USER, deleteuser.getCommand(), "command");
    assertEquals(REQUEST, deleteuser.getHeader(), "header");
    assertEquals(Protocol.DATASET, deleteuser.getDataset(), "dataset");
    assertEquals("louisxiv", deleteuser.getUserid(), "userid");
  }



  @Test
  public void decodeDELETEUSER_REQUEST_DecodeException() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"DELETE_USER\","
        + "\"DATASE\":{"
        + "\"USERID\":\"louisxiv\""
        + "}}";

    DELETEUSER deleteuser;
    try {
      deleteuser = (DELETEUSER) messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      // .error(e.getMessage(), e.getCause());
    }

    /* assertEquals("command", Command.DELETE_USER, deleteuser.getCommand());
     * assertEquals("header", REQUEST, deleteuser.getHeader());
     * assertEquals("dataset", Protocol.DATASET, deleteuser.getDataset());
     * assertEquals("userid", "louisxiv", deleteuser.getUserid()); */
  }



  @Test
  public void decodeDIAL_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_PORT\":29477,"
        + "\"RECEIVER_IP\":\"localhost\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\","
        + "\"ERRORMESSAGE\":\"Der Anwender hat in der Zwischenzeit seinen Nickname geändert.\""
        + "}}";

    DIAL dial = (DIAL) messagedecoder.decode(encoded);
    assertEquals(Command.DIAL, dial.getCommand(), "COMMAND");
    assertEquals(
        "Der Anwender hat in der Zwischenzeit seinen Nickname geändert.", dial.getErrorMessage(),
        "ErrorMessage"
    );
    assertEquals(ERROR, dial.getHeader(), "HEADER");
    assertEquals("localhost", dial.getReceiverIp(), "RECEIVER_IP");
    assertEquals("anaconda", dial.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals(29477, dial.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("12345678", dial.getCallerUserid(), "CALLER_USERID");
    assertEquals("niedrig", dial.getCallerVoice(), "CALLER_VOICE");
    assertEquals("javaanaconda", dial.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertNull(dial.getReceiverUserid(), "RECEIVER_USERID");
  }



  @Test
  public void decodeDIAL_ERROR_database_1() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_NICKNAME\":\"jester\","
        + "\"ERRORMESSAGE\":\"Der Anwender hat in der Zwischenzeit seinen Nickname geändert.\""
        + "}}";

    DIAL dial = (DIAL) messagedecoder.decode(encoded);
    assertEquals(Command.DIAL, dial.getCommand(), "COMMAND");
    assertEquals(
        "Der Anwender hat in der Zwischenzeit seinen Nickname geändert.", dial.getErrorMessage(),
        "ErrorMessage"
    );
    assertEquals(ERROR, dial.getHeader(), "HEADER");
    assertNull(dial.getReceiverIp(), "RECEIVER_IP");
    assertEquals("anaconda", dial.getCallerNickname(), "CALLER_NICKNAME");
    assertNull(dial.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("12345678", dial.getCallerUserid(), "CALLER_USERID");
    assertEquals("niedrig", dial.getCallerVoice(), "CALLER_VOICE");
    assertEquals("jester", dial.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertNull(dial.getReceiverUserid(), "RECEIVER_USERID");
  }



  @Test
  public void decodeDIAL_ERROR_MULTILINGUALKEY() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_PORT\":29477,"
        + "\"RECEIVER_IP\":\"localhost\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\","
        + "\"ERRORMESSAGE\":\"Der Anwender hat in der Zwischenzeit seinen Nickname geändert.\","
        + "\"MULTILINGUALKEY\":\"String_nimmt_keine_Anrufe_entgegen\""
        + "}}";

    DIAL dial = (DIAL) messagedecoder.decode(encoded);
    assertEquals(Command.DIAL, dial.getCommand(), "COMMAND");
    assertEquals(
        "Der Anwender hat in der Zwischenzeit seinen Nickname geändert.", dial.getErrorMessage(),
        "ErrorMessage"
    );
    assertEquals(ERROR, dial.getHeader(), "HEADER");
    assertEquals("localhost", dial.getReceiverIp(), "RECEIVER_IP");
    assertEquals("anaconda", dial.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals(29477, dial.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("12345678", dial.getCallerUserid(), "CALLER_USERID");
    assertEquals("niedrig", dial.getCallerVoice(), "CALLER_VOICE");
    assertEquals("javaanaconda", dial.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals(KEY.STRING_NIMMT_KEINE_ANRUFE_ENTGEGEN, dial.getMultilingualkey(), "MULTILINGUALKEY");
    assertNull(dial.getReceiverUserid(), "RECEIVER_USERID");
  }



  @Test
  public void decodeDIAL_ERROR_MULTILINGUALKEY_ohne_ERRORMESSAGE() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_PORT\":29477,"
        + "\"RECEIVER_IP\":\"localhost\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\","
        + "\"MULTILINGUALKEY\":\"String_nimmt_keine_Anrufe_entgegen\""
        + "}}";

    DIAL dial = (DIAL) messagedecoder.decode(encoded);
    assertEquals(Command.DIAL, dial.getCommand(), "COMMAND");
    assertNull(dial.getErrorMessage(), "ERRORMESSAGE");
    assertEquals(ERROR, dial.getHeader(), "HEADER");
    assertEquals("localhost", dial.getReceiverIp(), "RECEIVER_IP");
    assertEquals("anaconda", dial.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals(29477, dial.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("12345678", dial.getCallerUserid(), "CALLER_USERID");
    assertEquals("niedrig", dial.getCallerVoice(), "CALLER_VOICE");
    assertEquals("javaanaconda", dial.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals(KEY.STRING_NIMMT_KEINE_ANRUFE_ENTGEGEN, dial.getMultilingualkey(), "MULTILINGUALKEY");
    assertNull(dial.getReceiverUserid(), "RECEIVER_USERID");
  }



  @Test
  public void decodeDIAL_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\""
        + "}}";

    DIAL dial = (DIAL) messagedecoder.decode(encoded);
    assertEquals(Command.DIAL, dial.getCommand(), "COMMAND");
    assertEquals(REQUEST, dial.getHeader(), "HEADER");
    assertNull(dial.getReceiverIp(), "RECEIVER_IP");
    assertEquals("anaconda", dial.getCallerNickname(), "CALLER_NICKNAME");
    assertNull(dial.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("12345678", dial.getCallerUserid(), "CALLER_USERID");
    assertEquals("niedrig", dial.getCallerVoice(), "CALLER_VOICE");
    assertEquals("javaanaconda", dial.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertNull(dial.getReceiverUserid(), "RECEIVER_USERID");
  }



  @Test
  public void decodeDIAL_REQUEST_ohne_CALLERVOICE() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\""
        + "}}";

    DIAL dial;
    try {
      dial = (DIAL) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen");
    }
    catch (DecodeException e) {
      assert true;
    }
  }



  @Test
  public void decodeDIAL_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_PORT\":29477,"
        + "\"RECEIVER_IP\":\"localhost\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\","
        + "\"RECEIVER_USERID\":\"java123\","
        + "\"CALLER_SESSION\":\"4d\","
        + "\"RECEIVER_SESSION\":\"4e\""
        + "}}";

    DIAL dial = (DIAL) messagedecoder.decode(encoded);
    assertEquals(Command.DIAL, dial.getCommand(), "COMMAND");
    assertEquals(RESPONSE, dial.getHeader(), "HEADER");
    assertEquals("localhost", dial.getReceiverIp(), "RECEIVER_IP");
    assertEquals("anaconda", dial.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals(29477, dial.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("12345678", dial.getCallerUserid(), "CALLER_USERID");
    assertEquals("niedrig", dial.getCallerVoice(), "CALLER_VOICE");
    assertEquals("javaanaconda", dial.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("java123", dial.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("4d", dial.getCallerSession(), "CALLER_SESSION");
    assertEquals("4e", dial.getReceiverSession(), "RECEIVER_SESSION");
  }



  @Test
  public void decodeDIAL_RESPONSE_mit_SESSION() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"CALLER_SESSION\":\"8b\","
        + "\"RECEIVER_IP\":\"localhost\","
        + "\"RECEIVER_PORT\":29477,"
        + "\"RECEIVER_IP\":\"localhost\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\","
        + "\"RECEIVER_USERID\":\"java123\","
        + "\"RECEIVER_SESSION\":\"7c\""
        + "}}";

    DIAL dial = (DIAL) messagedecoder.decode(encoded);
    assertEquals(Command.DIAL, dial.getCommand(), "COMMAND");
    assertEquals(RESPONSE, dial.getHeader(), "HEADER");
    assertEquals("localhost", dial.getReceiverIp(), "RECEIVER_IP");
    assertEquals("anaconda", dial.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals(29477, dial.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("12345678", dial.getCallerUserid(), "CALLER_USERID");
    assertEquals("niedrig", dial.getCallerVoice(), "CALLER_VOICE");
    assertEquals("javaanaconda", dial.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("java123", dial.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("7c", dial.getReceiverSession(), "SESSION");
    assertEquals("8b", dial.getCallerSession(), "CALLER_SESSION");
  }



  @Test
  public void decodeDIAL_RESPONSE_ohne_CALLERVOICE() {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"RECEIVER_PORT\":29477,"
        + "\"RECEIVER_IP\":\"localhost\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\","
        + "\"RECEIVER_USERID\":\"java123\""
        + "}}";

    DIAL dial;
    try {
      dial = (DIAL) messagedecoder.decode(encoded);
      fail("Excrption nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeDIAL_RESPONSE_OHNE_IP_UND_OHNE_PORT() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"12345678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_NICKNAME\":\"javaanaconda\","
        + "\"RECEIVER_USERID\":\"java123\","
        + "\"CALLER_SESSION\":\"3c\","
        + "\"RECEIVER_SESSION\":\"3d\""
        + "}}";

    DIAL dial = (DIAL) messagedecoder.decode(encoded);
    assertEquals(Command.DIAL, dial.getCommand(), "COMMAND");
    assertEquals(RESPONSE, dial.getHeader(), "HEADER");
    assertNull(dial.getReceiverIp(), "RECEIVER_IP");
    assertEquals("anaconda", dial.getCallerNickname(), "CALLER_NICKNAME");
    assertNull(dial.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("12345678", dial.getCallerUserid(), "CALLER_USERID");
    assertEquals("niedrig", dial.getCallerVoice(), "CALLER_VOICE");
    assertEquals("javaanaconda", dial.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("java123", dial.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("3c", dial.getCallerSession(), "CALLER_SESSION");
    assertEquals("3d", dial.getReceiverSession(), "RECEIVER_SESSION");
  }



  @Test
  public void decodeDIAL_RESPONSE_ohne_PORT() {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DIAL\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"NICKNAME\":\"anaconda\","
        + "\"IP\":\"0.0.0.0\","
        + "\"VOICE\":\"niedrig\""
        + "}}";

    try {
      DIAL dial = (DIAL) messagedecoder.decode(encoded);
      fail("Exception für fehlenden PORT wurde nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeDOWNLOAD_ERROR() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"DOWNLOAD\","
        + "\"DATASET\":{"
        + "\"USERID\":\"python\","
        + "\"SLOT\":1,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"TEXT\":\"The file\","
        + "\"INDEX\":2873}"
        + "}";

    DOWNLOAD download = (DOWNLOAD) messagedecoder.decode(encoded);

    assertEquals(Command.DOWNLOAD, download.getCommand(), "command");
    assertEquals(ERROR, download.getHeader(), "header");
    assertEquals(Protocol.DATASET, download.getDataset(), "dataset");
    assertEquals("python", download.getUserid(), "python");
    assertEquals(1, download.getSlot(), "slot");
    assertEquals("hallo.mp3", download.getFilename(), "filename");
    assertEquals(2873, download.getIndex(), "index");
    assertEquals("The file", download.getText(), "text");
  }



  @Test
  public void decodeDOWNLOAD_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"DOWNLOAD\","
        + "\"DATASET\":{"
        + "\"USERID\":\"python\","
        + "\"SLOT\":1,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"INDEX\":2873}"
        + "}";

    DOWNLOAD download = (DOWNLOAD) messagedecoder.decode(encoded);

    assertEquals(Command.DOWNLOAD, download.getCommand(), "command");
    assertEquals(REQUEST, download.getHeader(), "header");
    assertEquals(Protocol.DATASET, download.getDataset(), "dataset");
    assertEquals("python", download.getUserid(), "python");
    assertEquals(1, download.getSlot(), "slot");
    assertEquals("hallo.mp3", download.getFilename(), "filename");
    assertEquals(2873, download.getIndex(), "index");
  }



  @Test
  public void decodeDOWNLOAD_RESPONSE() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DOWNLOAD\","
        + "\"DATASET\":{"
        + "\"USERID\":\"python\","
        + "\"SLOT\":1,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"IP\":\"127.0.0.1\","
        + "\"PATHFILE\":\"c:/hallo.mp3\","
        + "\"INDEX\":2873}"
        + "}";

    DOWNLOAD download;
    try {
      download = (DOWNLOAD) messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {}

  }



  @Test
  public void decodeDOWNLOAD_RESPONSE_PORT() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"DOWNLOAD\","
        + "\"DATASET\":{"
        + "\"USERID\":\"python\","
        + "\"SLOT\":1,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"IP\":\"127.0.0.1\","
        + "\"PATHFILE\":\"c:/hallo.mp3\","
        + "\"INDEX\":2873,"
        + "\"PORT\":8443"
        + "}"
        + "}";

    DOWNLOAD download = (DOWNLOAD) messagedecoder.decode(encoded);

    assertEquals(Command.DOWNLOAD, download.getCommand(), "command");
    assertEquals(RESPONSE, download.getHeader(), "header");
    assertEquals(Protocol.DATASET, download.getDataset(), "dataset");
    assertEquals("python", download.getUserid(), "python");
    assertEquals(1, download.getSlot(), "slot");
    assertEquals("hallo.mp3", download.getFilename(), "filename");
    assertEquals(2873, download.getIndex(), "index");
    assertEquals("c:/hallo.mp3", download.getPathfile(), "pathfile");
    assertEquals("127.0.0.1", download.getIp(), "index");
    assertEquals(8443, download.getPort(), "port");
  }



  @Test
  public void decodeENTERROOM_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"ENTERROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"anaconda\""
        + "}}";

    ENTERROOM enterroom = (ENTERROOM) messagedecoder.decode(encoded);
    assertEquals(Command.ENTERROOM, enterroom.getCommand(), "command");
    assertEquals(ERROR, enterroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, enterroom.getDataset(), "dataset");
    assertEquals("12345678", enterroom.getUserid(), "userid");
    assertEquals("anaconda", enterroom.getRoom(), "room");
  }



  @Test
  public void decodeENTERROOM_ERROR_Duplicate_Key() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"ENTERROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"anaconda\","
        + "\"ERRORCODE\":1047"
        + "}}";

    ENTERROOM enterroom = (ENTERROOM) messagedecoder.decode(encoded);
    assertEquals(Command.ENTERROOM, enterroom.getCommand(), "command");
    assertEquals(ERROR, enterroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, enterroom.getDataset(), "dataset");
    assertEquals("12345678", enterroom.getUserid(), "userid");
    assertEquals("anaconda", enterroom.getRoom(), "room");
    assertEquals(1047, enterroom.getErrorCode(), "errorCode");
  }



  @Test
  public void decodeENTERROOM_ERROR_Invalid_User() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"ENTERROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"anaconda\","
        + "\"ERRORMESSAGE\":\"You are not in a valid group to join the room.\""
        + "}}";

    ENTERROOM enterroom = (ENTERROOM) messagedecoder.decode(encoded);
    assertEquals(Command.ENTERROOM, enterroom.getCommand(), "command");
    assertEquals(ERROR, enterroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, enterroom.getDataset(), "dataset");
    assertEquals("12345678", enterroom.getUserid(), "userid");
    assertEquals("anaconda", enterroom.getRoom(), "room");
    assertEquals(
        "You are not in a valid group to join the room.", enterroom.getErrorMessage(), "errorMessage"
    );
  }



  @Test
  public void decodeENTERROOM_REQUEST() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"ENTERROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"anaconda\""
        + "}}";

    try {
      ENTERROOM enterroom = (ENTERROOM) messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen, NICKNAME fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeENTERROOM_REQUEST_Nickname() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"ENTERROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"NICKNAME\":\"viper\","
        + "\"ROOM\":\"anaconda\""
        + "}}";

    ENTERROOM enterroom = (ENTERROOM) messagedecoder.decode(encoded);
    assertEquals(Command.ENTERROOM, enterroom.getCommand(), "command");
    assertEquals(REQUEST, enterroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, enterroom.getDataset(), "dataset");
    assertEquals("12345678", enterroom.getUserid(), "userid");
    assertEquals("viper", enterroom.getNickname(), "nickname");
    assertEquals("anaconda", enterroom.getRoom(), "room");
  }



  @Test
  public void decodeENTERROOM_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"ENTERROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"NICKNAME\":\"viper\","
        + "\"ROOM\":\"anaconda\","
        + "\"ROOMTYPE\":\"PAUSENRAUM\""
        + "}}";

    ENTERROOM enterroom = (ENTERROOM) messagedecoder.decode(encoded);
    assertEquals(Command.ENTERROOM, enterroom.getCommand(), "command");
    assertEquals(RESPONSE, enterroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, enterroom.getDataset(), "dataset");
    assertEquals("12345678", enterroom.getUserid(), "userid");
    assertEquals("viper", enterroom.getNickname(), "nickname");
    assertEquals("anaconda", enterroom.getRoom(), "room");
  }



  @Test
  public void decodeENTERROOM_RESPONSE_DecodeException() {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"ENTERROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"anaconda\""
        + "}}";

    ENTERROOM enterroom;
    try {
      enterroom = (ENTERROOM) messagedecoder.decode(encoded);
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeENTERROOM_RESPONSE_ROOMTYPE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"ENTERROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"NICKNAME\":\"viper\","
        + "\"ROOM\":\"anaconda\","
        + "\"ROOMTYPE\":\"BESPRECHUNGSRAUM\""
        + "}}";

    ENTERROOM enterroom = (ENTERROOM) messagedecoder.decode(encoded);
    assertEquals(Command.ENTERROOM, enterroom.getCommand(), "command");
    assertEquals(RESPONSE, enterroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, enterroom.getDataset(), "dataset");
    assertEquals("12345678", enterroom.getUserid(), "userid");
    assertEquals("viper", enterroom.getNickname(), "nickname");
    assertEquals("anaconda", enterroom.getRoom(), "room");
    assertEquals(Roomtype.BESPRECHUNGSRAUM, enterroom.getRoomtype(), "roomtype");
  }



  @Test
  public void decodeFILETYPES_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"FILETYPES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"userid123\""
        + "}"
        + "}";

    FILETYPES filetypes = (FILETYPES) messagedecoder.decode(encoded);
    assertEquals(Command.FILETYPES, filetypes.getCommand(), "Command");
    assertEquals(REQUEST, filetypes.getHeader(), "REQUEST");
    assertEquals("userid123", filetypes.getUserid(), "USERID");
  }



  @Test
  public void decodeFILETYPES_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"FILETYPES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"userid123\","
        + "\"FILETYPES\":["
        + "{\"SUFFIX\":\"pdf\",\"DESCRIPTION\":\"Adobe\",CHECKED:true},"
        + "{\"SUFFIX\":\"txt\",\"DESCRIPTION\":\"Simple Text Format\",CHECKED:false}"
        + "]}"
        + "}";

    FILETYPES filetypes = (FILETYPES) messagedecoder.decode(encoded);
    assertEquals(Command.FILETYPES, filetypes.getCommand(), "Command");
    assertEquals(RESPONSE, filetypes.getHeader(), "RESPONSE");
    assertEquals("userid123", filetypes.getUserid(), "USERID");

    BlacklistTypes pdf = filetypes.getFiletypes()[0];
    BlacklistTypes txt = filetypes.getFiletypes()[1];

    assertEquals("pdf", pdf.getSuffix(), "pdfSuffix");
    assertEquals("Adobe", pdf.getDescription(), "Adobe");
    assertEquals(true, pdf.getChecked(), "pdfChecked");

    assertEquals("txt", txt.getSuffix(), "txtSuffix");
    assertEquals("Simple Text Format", txt.getDescription(), "Simple Text Format");
    assertEquals(false, txt.getChecked(), "txtChecked");

  }



  @Test
  public void decodeFTSDOWNLOAD_ERROR() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"FTSDOWNLOAD\","
        + "\"DATASET\":{"
        + "\"SLOT\":1,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"PATH\":\"c:/\","
        + "\"ERRORMESSAGE\":\"errormeaage\","
        + "\"TEXT\":\"ich bin ein text\""
        + "}"
        + "}";

    FTSDOWNLOAD fts = (FTSDOWNLOAD) messagedecoder.decode(encoded);

    assertEquals(Command.FTSDOWNLOAD, fts.getCommand(), "command");
    assertEquals(ERROR, fts.getHeader(), "header");
    assertEquals(Protocol.DATASET, fts.getDataset(), "dataset");
    assertEquals(1, fts.getSlot(), "slot");
    assertEquals("hallo.mp3", fts.getFilename(), "filename");
    assertEquals("c:/", fts.getPath(), "pathfile");
    assertEquals("errormeaage", fts.getErrorMessage(), "errorMessage");
    assertEquals("ich bin ein text", fts.getText(), "freier text");
  }



  @Test
  public void decodeFTSDOWNLOAD_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"FTSDOWNLOAD\","
        + "\"DATASET\":{"
        + "\"SLOT\":1,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"PATH\":\"c:/\"}"
        + "}";

    FTSDOWNLOAD fts = (FTSDOWNLOAD) messagedecoder.decode(encoded);

    assertEquals(Command.FTSDOWNLOAD, fts.getCommand(), "command");
    assertEquals(REQUEST, fts.getHeader(), "header");
    assertEquals(Protocol.DATASET, fts.getDataset(), "dataset");
    assertEquals(1, fts.getSlot(), "slot");
    assertEquals("hallo.mp3", fts.getFilename(), "filename");
    assertEquals("c:/", fts.getPath(), "pathfile");
  }



  @Test
  public void decodeFTSDOWNLOAD_REQUEST_ERROR() {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"FTSDOWNLOAD\","
        + "\"DATASET\":{\"SLOT\":0,"
        + "\"FILENAME\":\"wir_sind_helden2.mp4\"}}";

    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeFTSDOWNLOAD_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"FTSDOWNLOAD\","
        + "\"DATASET\":{"
        + "\"SLOT\":1,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"PATH\":\"c:/\","
        + "\"MAXDATABLOCKS\":848"
        + "}"
        + "}";

    FTSDOWNLOAD fts = (FTSDOWNLOAD) messagedecoder.decode(encoded);

    assertEquals(Command.FTSDOWNLOAD, fts.getCommand(), "command");
    assertEquals(RESPONSE, fts.getHeader(), "header");
    assertEquals(Protocol.DATASET, fts.getDataset(), "dataset");
    assertEquals(1, fts.getSlot(), "slot");
    assertEquals("hallo.mp3", fts.getFilename(), "filename");
    assertEquals("c:/", fts.getPath(), "pathfile");
    assertEquals(848, fts.getMaxDatablocks(), "maxdatablocks");
  }



  @Test
  public void decodeHISTORYMESSAGE_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"HISTORYMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\"}}";

    HISTORYMESSAGE history = (HISTORYMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.HISTORYMESSAGE, history.getCommand(), "command");
    assertEquals(REQUEST, history.getHeader(), "header");
    assertEquals(Protocol.DATASET, history.getDataset(), "dataset");
    assertEquals("room", history.getRoom(), "room");

  }



  @Test
  public void decodeHISTORYMESSAGE_RESPONSE_1_Nachricht() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"HISTORYMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"RECORD\":["
        + "{\"ROOM\":\"chat\",\"DATETIME\":666,\"MESSAGE\":\"Hallo\",\"CHATUSER\":{\"BACKGROUND_COLOR\":-1,\"FOREGROUND_COLOR\":-2,\"NICKNAME\":\"python\",\"USERID\":\"louisxiv\"}}"
        + "]"
        + "}}";

    HISTORYMESSAGE history = (HISTORYMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.HISTORYMESSAGE, history.getCommand(), "command");
    assertEquals(RESPONSE, history.getHeader(), "header");
    assertEquals(Protocol.DATASET, history.getDataset(), "dataset");
    assertEquals("room", history.getRoom(), "room");
    RecordInterface[] record = history.getAttachment();
    assertEquals(1, record.length, "1 Datensatz");
    assertEquals(666, record[0].getDatetime(), "datetime");
    assertEquals("Hallo", record[0].getMessage(), "message");
    assertEquals("chat", record[0].getRoom(), "room");
    ChatUser chatuser = record[0].getChatUser();
    assertEquals(-1, chatuser.getBackgroundColor(), "bgColor");
    assertEquals(-2, chatuser.getForegroundColor(), "fgColor");
    assertEquals("python", chatuser.getNickname(), "fgColor");
    assertEquals("louisxiv", chatuser.getUserid(), "fgColor");

  }



  @Test
  public void decodeHISTORYMESSAGE_RESPONSE_1_Nachricht_JERSEY() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"HISTORYMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"RECORD\":["
        + "{\"ROOM\":\"chat\",\"DATETIME\":666,\"MESSAGE\":\"Hallo\",\"CHATUSER\":{\"BACKGROUND_COLOR\":-1,\"FOREGROUND_COLOR\":-2,\"NICKNAME\":\"python\",\"USERID\":\"louisxiv\"}}"
        + "]"
        + "}}";

    GsonBuilder builder = new GsonBuilder();
    builder.registerTypeAdapter(HISTORYMESSAGE.class, new HistoryMessageDeserializer());
    Gson gson = builder.create();
    HISTORYMESSAGE history = gson.fromJson(encoded, HISTORYMESSAGE.class);

    assertEquals(Command.HISTORYMESSAGE, history.getCommand(), "command");
    assertEquals(RESPONSE, history.getHeader(), "header");
    assertEquals(Protocol.DATASET, history.getDataset(), "dataset");
    assertEquals("room", history.getRoom(), "room");
    RecordInterface[] record = history.getAttachment();
    assertEquals(1, record.length, "1 Datensatz");
    assertEquals(666, record[0].getDatetime(), "datetime");
    assertEquals("Hallo", record[0].getMessage(), "message");
    assertEquals("chat", record[0].getRoom(), "room");
    ChatUser chatuser = record[0].getChatUser();
    assertEquals(-1, chatuser.getBackgroundColor(), "bgColor");
    assertEquals(-2, chatuser.getForegroundColor(), "fgColor");
    assertEquals("python", chatuser.getNickname(), "fgColor");
    assertEquals("louisxiv", chatuser.getUserid(), "fgColor");

  }



  @Test
  public void decodeHISTORYMESSAGE_RESPONSE_ATTACHEMENT() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"HISTORYMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"RECORD\":["
        + "{\"ROOM\":\"chat\","
        + "\"DATETIME\":1458579290065,"
        + "\"MESSAGE\":\"Hallo\","
        + "\"ATTACHMENT\":99,"
        + "\"FILENAME\":\"name.txt\","
        + "\"FILESIZE\":"
        + String.valueOf(Long.MAX_VALUE)
        + ","
        + "\"CHATUSER\":{"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"FOREGROUND_COLOR\":-2,"
        + "\"NICKNAME\":\"python\","
        + "\"USERID\":\"louisxiv\""
        + "}}"
        + "]"
        + "}}";

    HISTORYMESSAGE history = (HISTORYMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.HISTORYMESSAGE, history.getCommand(), "command");
    assertEquals(RESPONSE, history.getHeader(), "header");
    assertEquals(Protocol.DATASET, history.getDataset(), "dataset");
    assertEquals("room", history.getRoom(), "room");
    Attachment[] record = history.getAttachment();
    assertEquals(1, record.length, "1 Datensatz");
    assertEquals(1458579290065L, record[0].getDatetime(), "datetime");
    assertEquals("Hallo", record[0].getMessage(), "message");
    assertEquals("chat", record[0].getRoom(), "room");
    ChatUser chatuser = record[0].getChatUser();
    assertEquals(-1, chatuser.getBackgroundColor(), "bgColor");
    assertEquals(-2, chatuser.getForegroundColor(), "fgColor");
    assertEquals("python", chatuser.getNickname(), "fgColor");
    assertEquals("louisxiv", chatuser.getUserid(), "fgColor");

    assertEquals("name.txt", record[0].getFilename());
    assertEquals(Long.MAX_VALUE, record[0].getFilesize());
    assertEquals(99, record[0].getAttachment());

  }



  @Test
  public void decodeHISTORYMESSAGE_RESPONSE_ATTACHEMENT_JERSEY() {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"HISTORYMESSAGE\","
        + "\"DATASET\":{"
        + "\"ROOM\":\"room\","
        + "\"RECORD\":["
        + "{\"ROOM\":\"chat\","
        + "\"DATETIME\":1458579290065,"
        + "\"MESSAGE\":\"Hallo\","
        + "\"ATTACHMENT\":99,"
        + "\"FILENAME\":\"name.txt\","
        + "\"FILESIZE\":"
        + String.valueOf(Long.MAX_VALUE)
        + ","
        + "\"CHATUSER\":{"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"FOREGROUND_COLOR\":-2,"
        + "\"NICKNAME\":\"python\","
        + "\"USERID\":\"louisxiv\""
        + "}}"
        + "]"
        + "}}";

    GsonBuilder builder = new GsonBuilder();
    builder.registerTypeAdapter(HISTORYMESSAGE.class, new HistoryMessageDeserializer());
    Gson gson = builder.create();
    HISTORYMESSAGE history = gson.fromJson(encoded, HISTORYMESSAGE.class);

    assertNotNull(history);
    assertEquals(Command.HISTORYMESSAGE, history.getCommand(), "command");
    assertEquals(RESPONSE, history.getHeader(), "header");
    assertEquals(Protocol.DATASET, history.getDataset(), "dataset");
    assertEquals("room", history.getRoom(), "room");
    Attachment[] record = history.getAttachment();
    assertEquals(1, record.length, "1 Datensatz");
    assertEquals(1458579290065L, record[0].getDatetime(), "datetime");
    assertEquals("Hallo", record[0].getMessage(), "message");
    assertEquals("chat", record[0].getRoom(), "room");
    ChatUser chatuser = record[0].getChatUser();
    assertEquals(-1, chatuser.getBackgroundColor(), "bgColor");
    assertEquals(-2, chatuser.getForegroundColor(), "fgColor");
    assertEquals("python", chatuser.getNickname(), "fgColor");
    assertEquals("louisxiv", chatuser.getUserid(), "fgColor");

    assertEquals("name.txt", record[0].getFilename(), "FILENAME");
    assertEquals(Long.MAX_VALUE, record[0].getFilesize(), "FILESIZE");
    assertEquals(99, record[0].getAttachment(), "ATTACHMENT");

  }



  @Test
  public void decodeIMAGE_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"IMAGE\","
        + "\"DATASET\":{"
        + "\"SENDERUID\":\"louisxiv\","
        + "\"SENDERNICKNAME\":\"anaconda\","
        + "\"RECEIVERNICKNAME\":\"freigeißt\","
        + "\"MULTILINGUALKEY\":\"Server_Der_Empfaenger_ist_verschwunden\""
        + "}}";

    IMAGE image = (IMAGE) messagedecoder.decode(encoded);

    assertEquals(Command.IMAGE, image.getCommand(), "command");
    assertEquals(ERROR, image.getHeader(), "header");
    assertEquals(Protocol.DATASET, image.getDataset(), "dataset");
    assertEquals("anaconda", image.getSenderNickname());
    assertEquals("louisxiv", image.getSenderUid());
    assertEquals("freigeißt", image.getReceiverNickname());
    assertEquals(KEY.SERVER_DER_EMPFAENGER_IST_VERSCHWUNDEN, image.getMultilingualkey());
  }



  @Test
  public void decodeIMAGE_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"IMAGE\","
        + "\"DATASET\":{"
        + "\"SENDERUID\":\"louisxiv\","
        + "\"SENDERNICKNAME\":\"anaconda\","
        + "\"RECEIVERNICKNAME\":\"freigeißt\""
        + "}}";

    IMAGE image = (IMAGE) messagedecoder.decode(encoded);

    assertEquals(Command.IMAGE, image.getCommand(), "command");
    assertEquals(REQUEST, image.getHeader(), "header");
    assertEquals(Protocol.DATASET, image.getDataset(), "dataset");
    assertEquals("anaconda", image.getSenderNickname());
    assertEquals("louisxiv", image.getSenderUid());
    assertEquals("freigeißt", image.getReceiverNickname());
  }



  @Test
  public void decodeIMAGE_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"IMAGE\","
        + "\"DATASET\":{"
        + "\"SENDERUID\":\"louisxiv\","
        + "\"SENDERNICKNAME\":\"anaconda\","
        + "\"RECEIVERNICKNAME\":\"freigeißt\","
        + "\"RECEIVERUID\":\"user5678\""
        + "}}";

    IMAGE image = (IMAGE) messagedecoder.decode(encoded);

    assertEquals(Command.IMAGE, image.getCommand(), "command");
    assertEquals(RESPONSE, image.getHeader(), "header");
    assertEquals(Protocol.DATASET, image.getDataset(), "dataset");
    assertEquals("anaconda", image.getSenderNickname());
    assertEquals("louisxiv", image.getSenderUid());
    assertEquals("freigeißt", image.getReceiverNickname());
    assertEquals("user5678", image.getReceiverUid());

  }



  @Test
  public void decodeINCOMINGCALL_CONFIRM() throws DecodeException {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\","
        + "\"RECEIVER_IP\":\"0.0.0.0\","
        + "\"RECEIVER_PORT\":29470"
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(CONFIRM, incoming.getHeader(), "CONFIRM");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertEquals(29477, incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("0.0.0.0", incoming.getReceiverIp(), "RECEIVER_IP");
    assertEquals(29470, incoming.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("192.168.1.67", incoming.getCallerIp(), "CALLER_IP");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
  }



  @Test
  public void decodeINCOMINGCALL_CONFIRM_ohne_CALLERVOICE() {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\","
        + "\"RECEIVER_IP\":\"0.0.0.0\","
        + "\"RECEIVER_PORT\":29470"
        + "}}";

    INCOMINGCALL incoming;
    try {
      incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
      fail("Exception wurde nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeINCOMINGCALL_CONFIRM_OHNE_IP_UND_PORT() throws DecodeException {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(CONFIRM, incoming.getHeader(), "CONFIRM");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertNull(incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertNull(incoming.getReceiverIp(), "RECEIVER_IP");
    assertNull(incoming.getReceiverPort(), "RECEIVER_PORT");
    assertNull(incoming.getCallerIp(), "CALLER_IP");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
  }



  @Test
  public void decodeINCOMINGCALL_CONFIRM_SESSION() throws DecodeException {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"CALLER_SESSION\":\"8b\","
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\","
        + "\"RECEIVER_IP\":\"0.0.0.0\","
        + "\"RECEIVER_PORT\":29470,"
        + "\"RECEIVER_SESSION\":\"7c\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(CONFIRM, incoming.getHeader(), "CONFIRM");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertEquals(29477, incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("0.0.0.0", incoming.getReceiverIp(), "RECEIVER_IP");
    assertEquals(29470, incoming.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("192.168.1.67", incoming.getCallerIp(), "CALLER_IP");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
    assertEquals("7c", incoming.getReceiverSession(), "RECEIVER_SESSION");
    assertEquals("8b", incoming.getCallerSession(), "CALLER_SESSION");
  }



  @Test
  public void decodeINCOMINGCALL_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"RECEIVER_NICKNAME\":\"llange\","
        + "\"ERRORMESSAGE\":\"'llange' ist nicht mehr online.\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(ERROR, incoming.getHeader(), "HEADER");
    assertEquals("llange", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("'llange' ist nicht mehr online.", incoming.getErrorMessage(), "ERRORMESSAGE");
  }



  @Test
  public void decodeINCOMINGCALL_ERROR_hat_aufgelegt() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\",\"COMMAND\":\"INCOMINGCALL\",\"DATASET\":{\"RECEIVER_NICKNAME\":\"anaconda\",\"MULTILINGUALKEY\":\"String_hat_aufgelegt\"}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(ERROR, incoming.getHeader(), "HEADER");
    assertEquals("anaconda", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals(KEY.STRING_HAT_AUFGELEGT, incoming.getMultilingualkey(), "MULTILINGUALKEY");
  }



  @Test
  public void decodeINCOMINGCALL_ERROR_MULTILINGUALKEY() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"RECEIVER_NICKNAME\":\"llange\","
        + "\"ERRORMESSAGE\":\"'llange' ist nicht mehr online.\","
        + "\"MULTILINGUALKEY\":\"String_ist_nicht_mehr_online\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(ERROR, incoming.getHeader(), "HEADER");
    assertEquals("llange", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("'llange' ist nicht mehr online.", incoming.getErrorMessage(), "ERRORMESSAGE");
    assertEquals(KEY.STRING_IST_NICHT_MEHR_ONLINE, incoming.getMultilingualkey(), "MULTILINGUALKEY");
  }



  @Test
  public void decodeINCOMINGCALL_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"CALLER_SESSION\":\"7b\","
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(REQUEST, incoming.getHeader(), "HEADER");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertEquals("192.168.1.67", incoming.getCallerIp(), "CALLER_IP");
    assertEquals(29477, incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("7b", incoming.getCallerSession(), "CALLER_SESSION");
  }



  @Test
  public void decodeINCOMINGCALL_REQUEST_mitSESSION() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"CALLER_SESSION\":\"7b\","
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\","
        + "\"RECEIVER_SESSION\":\"7c\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(REQUEST, incoming.getHeader(), "HEADER");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertEquals("192.168.1.67", incoming.getCallerIp(), "CALLER_IP");
    assertEquals(29477, incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("7c", incoming.getReceiverSession(), "RECEIVER_SESSION");
    assertEquals("7b", incoming.getCallerSession(), "CALLER_SESSION");
  }



  @Test
  public void decodeINCOMINGCALL_REQUEST_ohne_CALLERVOICE() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\""
        + "}}";

    INCOMINGCALL incoming;
    try {
      incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeINCOMINGCALL_REQUEST_OHNE_IP_UND_PORT() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_VOICE\":\"niedrig\","
        + "\"CALLER_SESSION\":\"7b\","
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(REQUEST, incoming.getHeader(), "HEADER");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertNull(incoming.getCallerIp(), "CALLER_IP");
    assertNull(incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("7b", incoming.getCallerSession(), "CALLER_SESSION");
  }



  @Test
  public void decodeINCOMINGCALL_REQUEST_ohne_PORT() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_NICKNAME\":\"hallo123\""
        + "}}";

    INCOMINGCALL incoming;
    try {
      incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen, weil der Port fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeINCOMINGCALL_REQUEST_ohne_VOICE() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"CALLER_NICKNAME\":\"hallo123\""
        + "}}";

    INCOMINGCALL incoming;
    try {
      incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen, weil VOICE fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeINCOMINGCALL_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"CALLER_VOICE\":niedrig,"
        + "\"CALLER_SESSION\":\"7b\","
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\","
        + "\"RECEIVER_IP\":\"0.0.0.0\","
        + "\"RECEIVER_PORT\":29470,"
        + "\"RECEIVER_VOICE\":\"hoch\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(RESPONSE, incoming.getHeader(), "HEADER");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertEquals("192.168.1.67", incoming.getCallerIp(), "CALLER_IP");
    assertEquals(29477, incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
    assertEquals("7b", incoming.getCallerSession(), "CALLER_SESSION");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("0.0.0.0", incoming.getReceiverIp(), "RECEIVER_IP");
    assertEquals(29470, incoming.getReceiverPort(), "RECEIVER_PORT");
  }



  @Test
  public void decodeINCOMINGCALL_RESPONSE_ohne_CALLERVOICE() {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\","
        + "\"RECEIVER_IP\":\"0.0.0.0\","
        + "\"RECEIVER_PORT\":29470"
        + "}}";

    INCOMINGCALL incoming;
    try {
      incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
      fail("Exception wurde nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeINCOMINGCALL_RESPONSE_OHNE_IP_UND_PORT() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_VOICE\":niedrig,"
        + "\"CALLER_SESSION\":\"7b\","
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\","
        + "\"RECEIVER_VOICE\":\"hoch\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(RESPONSE, incoming.getHeader(), "HEADER");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertNull(incoming.getCallerIp(), "CALLER_IP");
    assertNull(incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
    assertEquals("7b", incoming.getCallerSession(), "CALLER_SESSION");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertNull(incoming.getReceiverIp(), "RECEIVER_IP");
    assertNull(incoming.getReceiverPort(), "RECEIVER_PORT");
  }



  @Test
  public void decodeINCOMINGCALL_RESPONSE_RECEIVER_SESSION() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"INCOMINGCALL\","
        + "\"DATASET\":{"
        + "\"CALLER_NICKNAME\":\"hallo123\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"CALLER_IP\":\"192.168.1.67\","
        + "\"CALLER_PORT\":29477,"
        + "\"CALLER_VOICE\":niedrig,"
        + "\"CALLER_SESSION\":8b,"
        + "\"RECEIVER_USERID\":\"uid123\","
        + "\"RECEIVER_NICKNAME\":\"java\","
        + "\"RECEIVER_IP\":\"0.0.0.0\","
        + "\"RECEIVER_PORT\":29470,"
        + "\"RECEIVER_SESSION\":\"7c\","
        + "\"RECEIVER_VOICE\":\"hoch\""
        + "}}";

    INCOMINGCALL incoming = (INCOMINGCALL) messagedecoder.decode(encoded);
    assertEquals(Command.INCOMINGCALL, incoming.getCommand(), "COMMAND");
    assertEquals(RESPONSE, incoming.getHeader(), "HEADER");
    assertEquals("anaconda", incoming.getCallerUserid(), "CALLER_USERID");
    assertEquals("192.168.1.67", incoming.getCallerIp(), "CALLER_IP");
    assertEquals(29477, incoming.getCallerPort(), "CALLER_PORT");
    assertEquals("niedrig", incoming.getCallerVoice(), "CALLER_VOICE");
    assertEquals("hallo123", incoming.getCallerNickname(), "CALLER_NICKNAME");
    assertEquals("uid123", incoming.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("java", incoming.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("0.0.0.0", incoming.getReceiverIp(), "RECEIVER_IP");
    assertEquals(29470, incoming.getReceiverPort(), "RECEIVER_PORT");
    assertEquals("7c", incoming.getReceiverSession(), "RECEIVER_SESSION");
    assertEquals("8b", incoming.getCallerSession(), "CALLER_SESSION");
    assertEquals("hoch", incoming.getReceiverVoice(), "RECEIVER_VOICE");
  }



  @Test
  public void decodeKEEPALIVE_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"KEEPALIVE\"}";

    KEEPALIVE keepalive = (KEEPALIVE) messagedecoder.decode(encoded);
    assertEquals(Command.KEEPALIVE, keepalive.getCommand(), "command");
    assertEquals(REQUEST, keepalive.getHeader(), "header");
    assertEquals(null, keepalive.getDataset(), "dataset");
  }



  @Test
  public void decodeKEEPALIVE_REQUEST_JERSEY() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"KEEPALIVE\"}";

    Gson gson = new GsonBuilder().registerTypeAdapter(KEEPALIVE.class, new KeepaliveDeserializer()).create();
    KEEPALIVE keepalive = gson.fromJson(encoded, KEEPALIVE.class);
    assertEquals(Command.KEEPALIVE, keepalive.getCommand(), "command");
    assertEquals(REQUEST, keepalive.getHeader(), "header");
    assertNull(keepalive.getDataset(), "DATASET vorhanden");
  }



  @Test
  public void decodeKEEPALIVE_REQUEST_ohne_Attribut_SESSIONID() {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"KEEPALIVE\"," + "\"DATASET\":{" + "}}";

    try {
      KEEPALIVE keepalive = (KEEPALIVE) messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeKEEPALIVE_SESSION_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"KEEPALIVE\","
        + "\"DATASET\":{"
        + "\"SESSIONID\":\"e\""
        + "}}";

    KEEPALIVE keepalive = (KEEPALIVE) messagedecoder.decode(encoded);
    assertEquals(Command.KEEPALIVE, keepalive.getCommand(), "command");
    assertEquals(REQUEST, keepalive.getHeader(), "header");
    assertEquals("e", keepalive.getSessionId(), "header");
  }



  @Test
  public void decodeKEEPALIVE_SESSION_REQUEST_JERSEY() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"KEEPALIVE\","
        + "\"DATASET\":{"
        + "\"SESSIONID\":\"e\""
        + "}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(KEEPALIVE.class, new KeepaliveDeserializer()).create();
    KEEPALIVE keepalive = gson.fromJson(encoded, KEEPALIVE.class);
    assertEquals(Command.KEEPALIVE, keepalive.getCommand());
    assertEquals(REQUEST, keepalive.getHeader());
    assertEquals("e", keepalive.getSessionId(), "SESSIONID fehlt");
  }



  @Test
  public void decodeLEAVEPRIVATECHAT_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"LEAVEPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"GONE_SESSIONID\":\"5\"}}";

    LEAVEPRIVATECHAT chat = (LEAVEPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.LEAVEPRIVATECHAT, chat.getCommand(), "command");
    assertEquals(REQUEST, chat.getHeader(), "header");
    assertEquals(Protocol.DATASET, chat.getDataset(), "dataset");
    assertEquals("5", chat.getGoneSessionid(), "errorMessage");

  }



  @Test
  public void decodeLEAVEPRIVATECHAT_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"LEAVEPRIVATECHAT\","
        + "\"DATASET\":{"
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"GONE_SESSIONID\":\"5\"}}";

    LEAVEPRIVATECHAT chat = (LEAVEPRIVATECHAT) messagedecoder.decode(encoded);

    assertEquals(Command.LEAVEPRIVATECHAT, chat.getCommand(), "command");
    assertEquals(RESPONSE, chat.getHeader(), "header");
    assertEquals(Protocol.DATASET, chat.getDataset(), "dataset");
    assertEquals("5", chat.getGoneSessionid(), "gone");
    assertEquals("6", chat.getRemotesession(), "remote");

  }



  @Test
  public void decodeLEAVEROOM_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"LEAVEROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"anaconda\""
        + "}}";

    LEAVEROOM leaveroom = (LEAVEROOM) messagedecoder.decode(encoded);
    assertEquals(Command.LEAVEROOM, leaveroom.getCommand(), "command");
    assertEquals(ERROR, leaveroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, leaveroom.getDataset(), "dataset");
    assertEquals("12345678", leaveroom.getUserid(), "userid");
    assertEquals("anaconda", leaveroom.getRoom(), "room");
  }



  @Test
  public void decodeLEAVEROOM_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"LEAVEROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"anaconda\""
        + "}}";

    LEAVEROOM leaveroom = (LEAVEROOM) messagedecoder.decode(encoded);
    assertEquals(Command.LEAVEROOM, leaveroom.getCommand(), "command");
    assertEquals(REQUEST, leaveroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, leaveroom.getDataset(), "dataset");
    assertEquals("12345678", leaveroom.getUserid(), "userid");
    assertEquals("anaconda", leaveroom.getRoom(), "room");
  }



  @Test
  public void decodeLEAVEROOM_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"LEAVEROOM\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ROOM\":\"anaconda\""
        + "}}";

    LEAVEROOM leaveroom = (LEAVEROOM) messagedecoder.decode(encoded);
    assertEquals(Command.LEAVEROOM, leaveroom.getCommand(), "command");
    assertEquals(RESPONSE, leaveroom.getHeader(), "header");
    assertEquals(Protocol.DATASET, leaveroom.getDataset(), "dataset");
    assertEquals("12345678", leaveroom.getUserid(), "userid");
    assertEquals("anaconda", leaveroom.getRoom(), "room");
  }



  @Test
  public void decodeMICROERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"MICROERROR\","
        + "\"DATASET\":{"
        + "\"FROM_NICKNAME\":\"anaconda\","
        + "\"FROM_USERID\":\"jester\","
        + "\"TEXT\":\"Die Gegenseite ist nicht mehr online\","
        + "\"TO_USERID\":\"userid123\""
        + "}}";

    MICROERROR microerror = (MICROERROR) messagedecoder.decode(encoded);
    assertEquals(Command.MICROERROR, microerror.getCommand(), "MICROERROR");
    assertEquals(REQUEST, microerror.getHeader(), "REQUEST");
    assertEquals("anaconda", microerror.getFromNickname(), "FROM_NICKNAME");
    assertEquals("jester", microerror.getFromUserid(), "FROM_USERID");
    assertEquals("Die Gegenseite ist nicht mehr online", microerror.getText(), "TEXT");
    assertEquals("userid123", microerror.getToUserid(), "TO_USERID");

  }



  @Test
  public void decodeONCALL_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"ONCALL\","
        + "\"DATASET\":{"
        + "\"USERID\":\"anaconda\","
        + "\"ONCALL\":true"
        + "}}";

    ONCALL oncall = (ONCALL) messagedecoder.decode(encoded);
    assertEquals(Command.ONCALL, oncall.getCommand(), "COMMAND");
    assertEquals(REQUEST, oncall.getHeader(), "HEADER");
    assertEquals("anaconda", oncall.getUserid(), "userid");
    assertEquals(true, oncall.isOnCall(), "onCall");
  }



  @Test
  public void decodeONCALL_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"ONCALL\","
        + "\"DATASET\":{"
        + "\"USERID\":\"anaconda\","
        + "\"ONCALL\":false"
        + "}}";

    ONCALL oncall = (ONCALL) messagedecoder.decode(encoded);
    assertEquals(Command.ONCALL, oncall.getCommand(), "COMMAND");
    assertEquals(RESPONSE, oncall.getHeader(), "HEADER");
    assertEquals("anaconda", oncall.getUserid(), "userid");
    assertEquals(false, oncall.isOnCall(), "onCall");
  }



  @Test
  public void decodeONHOOK_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"ONHOOK\","
        + "\"DATASET\":{"
        + "\"RECEIVER_USERID\":\"jester\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"TO\":true,"
        + "\"SIGNAL\":true"
        + "}}";
    ONHOOK onhook = (ONHOOK) messagedecoder.decode(encoded);
    assertEquals(Command.ONHOOK, onhook.getCommand(), "ONHOOK");
    assertEquals(REQUEST, onhook.getHeader(), "REQUEST");
    assertEquals("jester", onhook.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("anaconda", onhook.getCallerUserid(), "CALLER_USERID");
    assertEquals(true, onhook.isTo(), "TO");
  }



  @Test
  public void decodeONHOOK_REQUEST_ohne_TO() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"ONHOOK\","
        + "\"DATASET\":{"
        + "\"RECEIVER_USERID\":\"jester\","
        + "\"CALLER_USERID\":\"anaconda\""
        + "}}";
    try {
      ONHOOK onhook = (ONHOOK) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen, TO");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeONHOOK_REQUEST_SIGNAL() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"ONHOOK\","
        + "\"DATASET\":{"
        + "\"RECEIVER_USERID\":\"jester\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"TO\":true,"
        + "\"SIGNAL\":true"
        + "}}";
    ONHOOK onhook = (ONHOOK) messagedecoder.decode(encoded);
    assertEquals(Command.ONHOOK, onhook.getCommand(), "ONHOOK");
    assertEquals(REQUEST, onhook.getHeader(), "REQUEST");
    assertEquals("jester", onhook.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("anaconda", onhook.getCallerUserid(), "CALLER_USERID");
    assertEquals(true, onhook.isTo(), "TO");
    assertEquals(true, onhook.isSignal(), "SIGNAL");
  }



  @Test
  public void decodeONHOOK_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"ONHOOK\","
        + "\"DATASET\":{"
        + "\"RECEIVER_USERID\":\"jester\","
        + "\"CALLER_USERID\":\"anaconda\","
        + "\"TO\":true"
        + "}}";
    ONHOOK onhook = (ONHOOK) messagedecoder.decode(encoded);
    assertEquals(Command.ONHOOK, onhook.getCommand(), "ONHOOK");
    assertEquals(RESPONSE, onhook.getHeader(), "RESPONSE");
    assertEquals("jester", onhook.getReceiverUserid(), "RECEIVER_USERID");
    assertEquals("anaconda", onhook.getCallerUserid(), "CALLER_USERID");
    assertEquals(true, onhook.isTo(), "TO");
  }



  @Test
  public void decodePRIVATECHATFILE_ERROR() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"PRIVATECHATFILE\"," + "\"DATASET\":{}}";

    PRIVATECHATFILE chatfile = (PRIVATECHATFILE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATECHATFILE, chatfile.getCommand(), "command");
    assertEquals(ERROR, chatfile.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatfile.getDataset(), "dataset");
  }



  @Test
  public void decodePRIVATECHATFILE_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATECHATFILE\","
        + "\"DATASET\":{"
        + "\"USERID\":\"anaconda\","
        + "\"FILENAME\":\"dateiname\","
        + "\"REMOTE_SESSION_ID\":\"6\","
        + "\"REMOTE_NICKNAME\":\"anaconda\","
        + "\"ABSOLUTE_FILENAME\":\"/absoluter/dateiname\","
        + "\"FILESIZE\":0,"
        + "\"MESSAGE\":\"\"}}";

    PRIVATECHATFILE chatfile = (PRIVATECHATFILE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATECHATFILE, chatfile.getCommand(), "command");
    assertEquals(REQUEST, chatfile.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatfile.getDataset(), "dataset");
    assertEquals("anaconda", chatfile.getUserid());
    assertEquals("dateiname", chatfile.getFilename());
    assertEquals("6", chatfile.getRemoteSessionId());
    assertEquals("/absoluter/dateiname", chatfile.getAbsoluteFilename());
    assertEquals(0, chatfile.getFilesize());
    assertEquals("", chatfile.getMessage());
    assertEquals("anaconda", chatfile.getRemoteNickname());

  }



  @Test
  public void decodePRIVATECHATFILE_REQUEST_OHNE_MESSAGE() {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATECHATFILE\","
        + "\"DATASET\":{"
        + "\"NUMBER\":33,"
        + "\"USERID\":\"anaconda\","
        + "\"FILENAME\":\"dateiname\","
        + "\"REMOTE_SESSION_ID\":\"6\","
        + "\"ABSOLUTE_FILENAME\":\"/absoluter/dateiname\","
        + "\"FILESIZE\":0"
        + "}}";

    try {
      PRIVATECHATFILE chatfile = (PRIVATECHATFILE) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodePRIVATECHATFILE_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\",\"COMMAND\":\"PRIVATECHATFILE\",\"DATASET\":"
        + "{\"NUMBER\":31,"
        + "\"REMOTE_NICKNAME\":\"nickname\","
        + "\"RECEIVER_USERID\":\"louisxiv\","
        + "\"USERID\":\"4nsb1j9r\","
        + "\"FILENAME\":\"98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4\","
        + "\"ABSOLUTE_FILENAME\":\"C:\\\\Users\\\\llange\\\\Videos\\\\98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4\","
        + "\"FILESIZE\":11319456,"
        + "\"MESSAGE\":\"123\","
        + "\"REMOTE_SESSION_ID\":\"6a\""
        + "}}";

    PRIVATECHATFILE chatfile = (PRIVATECHATFILE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATECHATFILE, chatfile.getCommand(), "command");
    assertEquals(RESPONSE, chatfile.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatfile.getDataset(), "dataset");
    assertEquals("4nsb1j9r", chatfile.getUserid());
    assertEquals("98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4", chatfile.getFilename());
    assertEquals(
        "C:\\Users\\llange\\Videos\\98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4",
        chatfile.getAbsoluteFilename()
    );
    assertEquals(11319456, chatfile.getFilesize());
    assertEquals("123", chatfile.getMessage());
    assertEquals("6a", chatfile.getRemoteSessionId());
    assertEquals(31, chatfile.getNumber());
    assertEquals("nickname", chatfile.getRemoteNickname());

  }



  @Test
  public void decodePRIVATECHATFILE_RESPONSE_ohne_MESSAGE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\",\"COMMAND\":\"PRIVATECHATFILE\",\"DATASET\":{"
        + "\"NUMBER\":77,"
        + "\"REMOTE_NICKNAME\":\"bla\","
        + "\"RECEIVER_USERID\":\"louisxiv\","
        + "\"USERID\":\"4nsb1j9r\","
        + "\"FILENAME\":\"98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4\","
        + "\"ABSOLUTE_FILENAME\":\"C:\\\\Users\\\\llange\\\\Videos\\\\98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4\","
        + "\"FILESIZE\":11319456,"
        + "\"REMOTE_SESSION_ID\":\"6a\""
        + "}}";

    PRIVATECHATFILE chatfile = (PRIVATECHATFILE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATECHATFILE, chatfile.getCommand(), "command");
    assertEquals(RESPONSE, chatfile.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatfile.getDataset(), "dataset");
    assertEquals("4nsb1j9r", chatfile.getUserid());
    assertEquals("98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4", chatfile.getFilename());
    assertEquals(
        "C:\\Users\\llange\\Videos\\98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4",
        chatfile.getAbsoluteFilename()
    );
    assertEquals(11319456, chatfile.getFilesize());
    assertEquals("6a", chatfile.getRemoteSessionId());
    assertEquals(null, chatfile.getMessage());
    assertEquals("bla", chatfile.getRemoteNickname());

  }



  @Test
  public void decodePRIVATECHATFILE_RESPONSE_OHNE_REMOTE_SESSION_ID() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\",\"COMMAND\":\"PRIVATECHATFILE\",\"DATASET\":"
        + "{\"NUMBER\":32,"
        + "\"REMOTE_NICKNAME\":\"\","
        + "\"RECEIVER_USERID\":\"louisxiv\","
        + "\"USERID\":\"4nsb1j9r\","
        + "\"FILENAME\":\"98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4\","
        + "\"ABSOLUTE_FILENAME\":\"C:\\\\Users\\\\llange\\\\Videos\\\\98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4\","
        + "\"FILESIZE\":11319456,"
        + "\"MESSAGE\":\"123\"}}";

    PRIVATECHATFILE chatfile = (PRIVATECHATFILE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATECHATFILE, chatfile.getCommand(), "command");
    assertEquals(RESPONSE, chatfile.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatfile.getDataset(), "dataset");
    assertEquals("4nsb1j9r", chatfile.getUserid());
    assertEquals("98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4", chatfile.getFilename());
    assertEquals(
        "C:\\Users\\llange\\Videos\\98   Tampa Curhat Beat _ Karl Wine   Baila Remix DVJ CraSh_.mp4",
        chatfile.getAbsoluteFilename()
    );
    assertEquals(11319456, chatfile.getFilesize());
    assertEquals("123", chatfile.getMessage());
    assertEquals("", chatfile.getRemoteNickname());

  }



  @Test
  public void decodePRIVATEMESSAGE_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"LOCAL_SESSIONID\":\"5\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"DATETIME\":123,"
        + "\"MESSAGE\":\"Hallo\"}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("5", chatmessage.getLocalSessionid(), "local");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertEquals(123, chatmessage.getDatetime(), "datetime");
    assertNull(chatmessage.getAttachment());
    assertNull(chatmessage.getFilename());
    assertNull(chatmessage.getFilesize());
  }



  @Test
  public void decodePRIVATEMESSAGE_REQUEST_2() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"SENDER_UID\":\"python\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"DATETIME\":123,"
        + "\"MESSAGE\":\"Hallo\"}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("python", chatmessage.getSenderUID(), "local");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertEquals(123, chatmessage.getDatetime(), "datetime");

  }



  @Test
  public void decodePRIVATEMESSAGE_REQUEST_3() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"SENDER_UID\":\"python\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"DATETIME\":123,"
        + "\"MESSAGE\":\"Hallo\"}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("python", chatmessage.getSenderUID(), "local");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertEquals(123, chatmessage.getDatetime(), "datetime");

  }



  @Test
  public void decodePRIVATEMESSAGE_REQUEST_4() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"SENDER_UID\":\"louisxiv\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"MESSAGE\":\"Hallo\"}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("louisxiv", chatmessage.getSenderUID(), "senderUID");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertNull(chatmessage.getDatetime(), "datetime");

  }



  @Test
  public void decodePRIVATEMESSAGE_REQUEST_5() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"SENDER_UID\":\"louisxiv\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"MESSAGE\":\"Hallo\","
        + "\"CHATUSER\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\","
        + "\"FOREGROUND_COLOR\":-2,"
        + "\"BACKGROUND_COLOR\":-3"

        + "}}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("louisxiv", chatmessage.getSenderUID(), "senderUID");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertNull(chatmessage.getDatetime(), "datetime");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "uid");
    assertEquals("anaconda", chatuser.getNickname(), "nickname");
    assertEquals(-2, chatuser.getForegroundColor(), "forground");
    assertEquals(-3, chatuser.getBackgroundColor(), "background");
  }



  @Test
  public void decodePRIVATEMESSAGE_REQUEST_ATTACHMENT() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"LOCAL_SESSIONID\":\"5\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"DATETIME\":123,"
        + "\"MESSAGE\":\"Hallo\","
        + "\"ATTACHMENT\":33,"
        + "\"FILENAME\":\"einTest.mp3\","
        + "\"FILESIZE\":8192"
        + "}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(REQUEST, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("5", chatmessage.getLocalSessionid(), "local");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertEquals(123, chatmessage.getDatetime(), "datetime");
    assertEquals(33, chatmessage.getAttachment(), "NUMBER");
    assertEquals("einTest.mp3", chatmessage.getFilename(), "FILENAME");
    assertEquals(8192, chatmessage.getFilesize(), "FILESIZE");

  }



  @Test
  public void decodePRIVATEMESSAGE_REQUEST_OHNE_REMOTESESSIONID() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"LOCAL_SESSIONID\":\"5\","
        + "\"DATETIME\":123,"
        + "\"MESSAGE\":\"Hallo\","
        + "\"NUMBER\":33,"
        + "\"FILENAME\":\"einTest.mp3\","
        + "\"FILESIZE\":8192"
        + "}}";

    try {
      PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen, obwohl REMOTE_SESSIONID fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodePRIVATEMESSAGE_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"LOCAL_SESSIONID\":\"5\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"DATETIME\":123,"
        + "\"MESSAGE\":\"Hallo\""
        + "}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(RESPONSE, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("5", chatmessage.getLocalSessionid(), "local");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertEquals(123, chatmessage.getDatetime(), "datetime");
    assertNull(chatmessage.getAttachment());
    assertNull(chatmessage.getFilename());
    assertNull(chatmessage.getFilesize());

  }



  @Test
  public void decodePRIVATEMESSAGE_RESPONSE_2() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"SENDER_UID\":\"anaconda\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"DATETIME\":123,"
        + "\"MESSAGE\":\"Hallo\"}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(RESPONSE, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("anaconda", chatmessage.getSenderUID(), "local");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertEquals(123, chatmessage.getDatetime(), "datetime");

  }



  @Test
  public void decodePRIVATEMESSAGE_RESPONSE_5() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"SENDER_UID\":\"louisxiv\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"MESSAGE\":\"Hallo\","
        + "\"CHATUSER\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\","
        + "\"FOREGROUND_COLOR\":-2,"
        + "\"BACKGROUND_COLOR\":-3"

        + "}}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(RESPONSE, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("louisxiv", chatmessage.getSenderUID(), "senderUID");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertNull(chatmessage.getDatetime(), "datetime");
    ChatUser chatuser = chatmessage.getChatUser();
    assertEquals("louisxiv", chatuser.getUserid(), "uid");
    assertEquals("anaconda", chatuser.getNickname(), "nickname");
    assertEquals(-2, chatuser.getForegroundColor(), "forground");
    assertEquals(-3, chatuser.getBackgroundColor(), "background");
  }



  @Test
  public void decodePRIVATEMESSAGE_RESPONSE_ATTACHMENT() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"PRIVATEMESSAGE\","
        + "\"DATASET\":{"
        + "\"LOCAL_SESSIONID\":\"5\","
        + "\"REMOTE_SESSIONID\":\"6\","
        + "\"DATETIME\":123,"
        + "\"MESSAGE\":\"Hallo\","
        + "\"NUMBER\":33,"
        + "\"FILENAME\":\"einTest.mp3\","
        + "\"FILESIZE\":8192"
        + "}}";

    PRIVATEMESSAGE chatmessage = (PRIVATEMESSAGE) messagedecoder.decode(encoded);

    assertEquals(Command.PRIVATEMESSAGE, chatmessage.getCommand(), "command");
    assertEquals(RESPONSE, chatmessage.getHeader(), "header");
    assertEquals(Protocol.DATASET, chatmessage.getDataset(), "dataset");
    assertEquals("5", chatmessage.getLocalSessionid(), "local");
    assertEquals("6", chatmessage.getRemoteSessionid(), "remote");
    assertEquals("Hallo", chatmessage.getMessage(), "message");
    assertEquals(123, chatmessage.getDatetime(), "datetime");

  }



  @Test
  public void decodePROYECTORCLOSING_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"PROYECTORCLOSING\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\""
        + "}}";

    PROYECTORCLOSING proyector = (PROYECTORCLOSING) messagedecoder.decode(encoded);
    assertEquals(Command.PROYECTORCLOSING, proyector.getCommand(), "PROYECTORCLOSING");
    assertEquals(REQUEST, proyector.getHeader(), "REQUEST");
    assertEquals(Protocol.DATASET, proyector.getDataset(), "DATASET");
    assertEquals("louisxiv", proyector.getUserid(), "userid");
  }



  @Test
  public void decodePROYECTORCLOSING_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"PROYECTORCLOSING\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\""
        + "}}";

    PROYECTORCLOSING proyector = (PROYECTORCLOSING) messagedecoder.decode(encoded);
    assertEquals(Command.PROYECTORCLOSING, proyector.getCommand(), "PROYECTORCLOSING");
    assertEquals(RESPONSE, proyector.getHeader(), "RESPONSE");
    assertEquals(Protocol.DATASET, proyector.getDataset(), "DATASET");
    assertEquals("louisxiv", proyector.getUserid(), "userid");
  }



  @Test
  public void decodeREADROOMS_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"READROOMS\","
        + "\"DATASET\":{"
        + "\"USERID\":\"anaconda\","
        + "\"ROOMTYPE\":4"
        + "}}";

    READROOMS readrooms = (READROOMS) messagedecoder.decode(encoded);
    assertEquals(Command.READROOMS, readrooms.getCommand(), "READROOMS");
    assertEquals(REQUEST, readrooms.getHeader(), "REQUEST");
    assertEquals("anaconda", readrooms.getUserid(), "USERID");
    assertEquals(Roomtype.GRUPPENRAUM, readrooms.getRoomtype(), "ROOMTYPE");
  }



  @Test
  public void decodeREADROOMS_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"READROOMS\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOMTYPE\":4,"
        + "\"ROOMS\":[\"schach\",\"radsport\"]"
        + "}}";

    READROOMS readrooms = (READROOMS) messagedecoder.decode(encoded);
    assertEquals(Command.READROOMS, readrooms.getCommand(), "READROOMS");
    assertEquals(RESPONSE, readrooms.getHeader(), "RESPONSE");
    assertEquals("louisxiv", readrooms.getUserid(), "USERID");
    assertEquals(Roomtype.GRUPPENRAUM, readrooms.getRoomtype(), "ROOMTYPE");
    assertEquals(2, readrooms.getRooms().length, "2 Räume");
  }



  @Test
  public void decodeREADROOMS_RESPONSE_EMPTY() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"READROOMS\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOMTYPE\":4,"
        + "\"ROOMS\":[]"
        + "}}";

    READROOMS readrooms = (READROOMS) messagedecoder.decode(encoded);
    assertEquals(Command.READROOMS, readrooms.getCommand(), "READROOMS");
    assertEquals(RESPONSE, readrooms.getHeader(), "RESPONSE");
    assertEquals("louisxiv", readrooms.getUserid(), "USERID");
    assertEquals(Roomtype.GRUPPENRAUM, readrooms.getRoomtype(), "ROOMTYPE");
    assertEquals(0, readrooms.getRooms().length, "0 Räume");
  }



  @Test
  public void decodeREADTOPICROOMOWNER_REQUEST_OWNER() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"READTOPICROOMOWNER\","
        + "\"DATASET\":"
        + "{\"OWNER\":\"viper\","
        + "\"USERID\":\"12345678\""
        + "}}";

    READTOPICROOMOWNER roomlist = (READTOPICROOMOWNER) messagedecoder.decode(encoded);
    assertEquals(Command.READTOPICROOMOWNER, roomlist.getCommand(), "command");
    assertEquals(REQUEST, roomlist.getHeader(), "header");
    assertEquals("viper", roomlist.getOwner(), "owner");
    assertEquals("12345678", roomlist.getUserid(), "usweid");
  }



  @Test
  public void decodeREADTOPICROOMOWNER_RESPONSE_LIST() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"READTOPICROOMOWNER\","
        + "\"DATASET\":"
        + "{\"OWNER\":\"viper\","
        + "\"USERID\":\"12345678\","
        + "\"ROOMLIST\":"
        + "["
        + "{\"ROOM\":\"JFZI\",\"ENTRY\":\"PROTECTED\",\"LIFETIME\":\"PERMANENT\",\"OWNER\":\"viper\"}"
        + "]"
        + "}}";

    READTOPICROOMOWNER roomlist = (READTOPICROOMOWNER) messagedecoder.decode(encoded);
    assertEquals(Command.READTOPICROOMOWNER, roomlist.getCommand(), "command");
    assertEquals(RESPONSE, roomlist.getHeader(), "header");
    assertEquals("viper", roomlist.getOwner(), "owner");
    assertEquals("12345678", roomlist.getUserid(), "userid");
    Room[] rooms = roomlist.getRoom();
    assertEquals(1, rooms.length, "roomlist 0 size");
    assertEquals("JFZI", rooms[0].getChatid(), "chatid");
    assertEquals("PROTECTED", rooms[0].getEntry().name(), "ENTRY");
    assertEquals(Lifetime.PERMANENT, rooms[0].getLifetime(), "LIFETIME");
    assertEquals("viper", rooms[0].getOwner(), "OWNER");
  }



  @Test
  public void decodeREADTOPICROOMOWNER_RESPONSE_OWNER() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"READTOPICROOMOWNER\","
        + "\"DATASET\":"
        + "{\"OWNER\":\"viper\","
        + "\"USERID\":\"12345678\","
        + "\"ROOMLIST\":[]"
        + "}}";

    READTOPICROOMOWNER roomlist = (READTOPICROOMOWNER) messagedecoder.decode(encoded);
    assertEquals(Command.READTOPICROOMOWNER, roomlist.getCommand(), "command");
    assertEquals(RESPONSE, roomlist.getHeader(), "header");
    assertEquals("viper", roomlist.getOwner(), "owner");
    assertEquals("12345678", roomlist.getUserid(), "usweid");
    assertNotNull(roomlist.getRoom(), "roomlist not null");
    assertEquals(0, roomlist.getRoom().length, "roomlist 0 size");
  }



  @Test
  public void decodeROOMLIST_REQUEST()
      throws DecodeException, JsonSyntaxException, UnsupportedEncodingException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"ROOMLIST\"" + "}";

    ROOMLIST roomlist = (ROOMLIST) messagedecoder.decodeROOMLIST(encoded);
    assertEquals(Command.ROOMLIST, roomlist.getCommand(), "command");
    assertEquals(REQUEST, roomlist.getHeader(), "header");

  }



  @Test
  public void decodeROOMLIST_RESPONSE()
      throws DecodeException, JsonSyntaxException, UnsupportedEncodingException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"ROOMLIST\","
        + "\"DATASET\":{\"ROOMLIST\":["
        + "{\"ENTRY\":\"PUBLIC\",\"LIFETIME\":\"PERMANENT\",\"ROOM\":\"Everyone\",\"ROOMTYPE\":\"BESPRECHUNGSRAUM\"},"
        + "{\"ENTRY\":\"PROTECTED\",\"LIFETIME\":\"TEMPORARY\",\"ROOM\":\"Developer\",\"ROOMTYPE\":\"PAUSENRAUM\"}"
        + "]}}";

    ROOMLIST roomlist = (ROOMLIST) messagedecoder.decodeROOMLIST(encoded);
    assertEquals(Command.ROOMLIST, roomlist.getCommand(), "command");
    assertEquals(RESPONSE, roomlist.getHeader(), "header");
    assertEquals(Protocol.DATASET, roomlist.getDataset(), "dataset");
    assertEquals(2, roomlist.getRoom().length, "2 Felder");
    assertEquals(Lifetime.TEMPORARY, roomlist.getRoom()[1].getLifetime(), "TEMPORARY");
    assertEquals(Lifetime.PERMANENT, roomlist.getRoom()[0].getLifetime(), "PERMANENT");

    assertEquals(Roomtype.BESPRECHUNGSRAUM, roomlist.getRoom()[0].getRoomtype(), "");
    assertEquals(Roomtype.PAUSENRAUM, roomlist.getRoom()[1].getRoomtype(), "PAUSENRAUM");

  }



  @Test
  public void decodeSCREENCAST_ERROR() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SCREENCAST\","
        + "\"DATASET\":"
        + "{\"RECEIVER_NICKNAME\":\"anaconda\","
        + "\"ERRORMESSAGE\":\"freigeist ist nicht mehr online.\""
        + "}}";

    SCREENCAST screencast = (SCREENCAST) messagedecoder.decode(encoded);

    assertEquals("freigeist ist nicht mehr online.", screencast.getErrorMessage(), "errormessage");
    assertEquals("anaconda", screencast.getReceiverNickname(), "RECEIVER_NICKNAME");

  }



  @Test
  public void decodeSCREENCAST_ERROR_ohne_RECEIVER_NICKNAME() {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SCREENCAST\","
        + "\"DATASET\":{"
        + "\"RECEIVER_USERID\":\"admin678\","
        + "\"ERRORMESSAGE\":\"freigeist ist nicht mehr online.\""
        + "}}";

    SCREENCAST screencast;
    try {
      screencast = (SCREENCAST) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeSCREENCAST_ERROR_RECEIVERUSERID() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SCREENCAST\","
        + "\"DATASET\":"
        + "{\"RECEIVER_NICKNAME\":\"anaconda\","
        + "\"RECEIVER_USERID\":\"admin678\","
        + "\"ERRORMESSAGE\":\"freigeist ist nicht mehr online.\""
        + "}}";

    SCREENCAST screencast = (SCREENCAST) messagedecoder.decode(encoded);

    assertEquals("freigeist ist nicht mehr online.", screencast.getErrorMessage(), "errormessage");
    assertEquals("anaconda", screencast.getReceiverNickname(), "RECEIVER_NICKNAME");
    assertEquals("admin678", screencast.getReceiverUserid(), "RECEIVER_USERID");

  }



  @Test
  public void decodeSCREENCAST_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"SCREENCAST\","
        + "\"DATASET\":"
        + "{\"CALLER_USERID\":\"admin678\","
        + "\"CALLER_NICKNAME\":\"anaconda\","
        + "\"RECEIVER_NICKNAME\":\"freigeist\","
        + "\"CALLER_IP\":\"javacomm.net\","
        + "\"CALLER_PORT\":29478"
        + "}}";

    SCREENCAST screencast = (SCREENCAST) messagedecoder.decode(encoded);

    assertEquals("anaconda", screencast.getCallerNickname(), "nickname");
    assertEquals("admin678", screencast.getCallerUserid(), "userid");
    assertEquals("freigeist", screencast.getReceiverNickname(), "RECEIVER_NICKNAME");

  }



  @Test
  public void decodeSCREENCAST_REQUEST_DecoderException() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"SCREENCAST\"," + "\"DATASET\":{}}";

    try {
      SCREENCAST screencast = (SCREENCAST) messagedecoder.decode(encoded);
      fail("DecodeException wurde nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeSCREENCAST_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SCREENCAST\","
        + "\"DATASET\":"
        + "{\"CALLER_USERID\":\"caller78\","
        + "\"RECEIVER_USERID\":\"admin678\","
        + "\"RECEIVER_NICKNAME\":\"anaconda\""
        + "}}";

    SCREENCAST screencast = (SCREENCAST) messagedecoder.decode(encoded);

    assertEquals("caller78", screencast.getCallerUserid(), "caller userid");
    assertEquals("anaconda", screencast.getReceiverNickname(), "nickname");
    assertEquals("admin678", screencast.getReceiverUserid(), "userid");

  }



  @Test
  public void decodeSEARCHFILES_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"SEARCHFILES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"SEARCHFILES\":["
        + "{\"FILE\":\"abc.123\"},"
        + "{\"FILE\":\"hallo.456\"}]}}";

    SEARCHFILES searchfiles = (SEARCHFILES) messagedecoder.decode(encoded);
    String[] files = searchfiles.getPatterns();

    assertEquals(Command.SEARCHFILES, searchfiles.getCommand(), "command");
    assertEquals(REQUEST, searchfiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, searchfiles.getDataset(), "dataset");
    assertEquals("louisxiv", searchfiles.getUserid(), "userid");
    assertEquals("abc.123", files[0], "files1");
    assertEquals("hallo.456", files[1], "files1");
  }



  @Test
  public void decodeSEARCHFILES_REQUEST_2() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"SEARCHFILES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"SEARCHFILES\":["
        + "]}}";

    SEARCHFILES searchfiles = (SEARCHFILES) messagedecoder.decode(encoded);
    String[] files = searchfiles.getPatterns();

    assertEquals(Command.SEARCHFILES, searchfiles.getCommand(), "command");
    assertEquals(REQUEST, searchfiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, searchfiles.getDataset(), "dataset");
    assertEquals("louisxiv", searchfiles.getUserid(), "userid");
    assertEquals(0, files.length, "0");
  }



  @Test
  public void decodeSEARCHFILES_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SEARCHFILES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"START\":true,"
        + "\"SEARCHFILE\":["
        + "{\"INDEX\":1,"
        + "\"FILENAME\":\"abc.123\","
        + "\"FILESIZE\":123},"
        + "{\"INDEX\":2,"
        + "\"FILENAME\":\"hallo.456\","
        + "\"FILESIZE\":456}"
        + "]}}";

    SEARCHFILES searchfiles = (SEARCHFILES) messagedecoder.decode(encoded);
    Searchfile[] files = searchfiles.getSearchfiles();

    assertEquals(Command.SEARCHFILES, searchfiles.getCommand(), "command");
    assertEquals(RESPONSE, searchfiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, searchfiles.getDataset(), "dataset");
    assertEquals("louisxiv", searchfiles.getUserid(), "userid");
    assertEquals(true, searchfiles.hasStarted(), "start");

    assertEquals("abc.123", files[0].getFilename(), "files1");
    assertEquals(1, files[0].getIndex(), "index1");
    assertEquals("hallo.456", files[1].getFilename(), "files2");
    assertEquals(2, files[1].getIndex(), "index2");
  }



  @Test
  public void decodeSEARCHFILES_RESPONSE_JERSEY() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SEARCHFILES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"START\":true,"
        + "\"SEARCHFILE\":["
        + "{\"INDEX\":1,"
        + "\"FILENAME\":\"abc.123\","
        + "\"FILESIZE\":123},"
        + "{\"INDEX\":2,"
        + "\"FILENAME\":\"hallo.456\","
        + "\"FILESIZE\":456}"
        + "]}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SEARCHFILES.class, new SearchFilesDeserializer())
        .create();
    SEARCHFILES searchfiles = gson.fromJson(encoded, SEARCHFILES.class);

    Searchfile[] files = searchfiles.getSearchfiles();

    assertEquals(Command.SEARCHFILES, searchfiles.getCommand(), "COMMAND");
    assertEquals(RESPONSE, searchfiles.getHeader(), "HEADER");
    assertEquals(Protocol.DATASET, searchfiles.getDataset(), "DATASET");
    assertEquals("louisxiv", searchfiles.getUserid(), "USERID");
    assertEquals(true, searchfiles.hasStarted(), "START");

    assertEquals("abc.123", files[0].getFilename(), "FILENAME 1");
    assertEquals(1, files[0].getIndex(), "INDEX 1");
    assertEquals(123, files[0].getFilesize(), "FILESIZE 1");

    assertEquals("hallo.456", files[1].getFilename(), "FILENAME 2");
    assertEquals(2, files[1].getIndex(), "INDEX 2");
    assertEquals(456, files[1].getFilesize(), "FILESIZE 2");
  }



  @Test
  public void decodeSEARCHFILES_RESPONSE_ohne_START() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SEARCHFILES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"SEARCHFILE\":["
        + "{\"INDEX\":1,"
        + "\"FILENAME\":\"abc.123\","
        + "\"FILESIZE\":123},"
        + "{\"INDEX\":2,"
        + "\"FILENAME\":\"hallo.456\","
        + "\"FILESIZE\":890}"
        + "]}}";

    SEARCHFILES searchfiles = (SEARCHFILES) messagedecoder.decode(encoded);
    Searchfile[] files = searchfiles.getSearchfiles();

    assertEquals(Command.SEARCHFILES, searchfiles.getCommand(), "command");
    assertEquals(RESPONSE, searchfiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, searchfiles.getDataset(), "dataset");
    assertEquals("louisxiv", searchfiles.getUserid(), "userid");
    assertEquals(false, searchfiles.hasStarted(), "start");

    assertEquals("abc.123", files[0].getFilename(), "files1");
    assertEquals(1, files[0].getIndex(), "index1");
    assertEquals(123, files[0].getFilesize(), "filesize1");

    assertEquals("hallo.456", files[1].getFilename(), "files2");
    assertEquals(2, files[1].getIndex(), "index2");
    assertEquals(890, files[1].getFilesize(), "filesize2");
  }



  @Test
  public void decodeSEARCHFILES_RESPONSE_ohne_START_JERSEY() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SEARCHFILES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"SEARCHFILE\":["
        + "{\"INDEX\":1,"
        + "\"FILENAME\":\"abc.123\","
        + "\"FILESIZE\":123},"
        + "{\"INDEX\":2,"
        + "\"FILENAME\":\"hallo.456\","
        + "\"FILESIZE\":890}"
        + "]}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SEARCHFILES.class, new SearchFilesDeserializer())
        .create();

    SEARCHFILES searchfiles = gson.fromJson(encoded, SEARCHFILES.class);
    Searchfile[] files = searchfiles.getSearchfiles();

    assertEquals(Command.SEARCHFILES, searchfiles.getCommand(), "command");
    assertEquals(RESPONSE, searchfiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, searchfiles.getDataset(), "dataset");
    assertEquals("louisxiv", searchfiles.getUserid(), "userid");
    assertEquals(true, searchfiles.hasStarted(), "start");

    assertEquals("abc.123", files[0].getFilename(), "files1");
    assertEquals(1, files[0].getIndex(), "index1");
    assertEquals(123, files[0].getFilesize(), "filesize1");

    assertEquals("hallo.456", files[1].getFilename(), "files2");
    assertEquals(2, files[1].getIndex(), "index2");
    assertEquals(890, files[1].getFilesize(), "filesize2");
  }



  @Test
  public void decodeSEARCHFILES_RESPONSE_OHNE_USERID_JERSEY() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SEARCHFILES\","
        + "\"DATASET\":{"
        + "\"START\":true,"
        + "\"SEARCHFILE\":["
        + "{\"INDEX\":1,"
        + "\"FILENAME\":\"abc.123\","
        + "\"FILESIZE\":123},"
        + "{\"INDEX\":2,"
        + "\"FILENAME\":\"hallo.456\","
        + "\"FILESIZE\":456}"
        + "]}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SEARCHFILES.class, new SearchFilesDeserializer())
        .create();

    SEARCHFILES searchfiles = gson.fromJson(encoded, SEARCHFILES.class);

    Searchfile[] files = searchfiles.getSearchfiles();

    assertEquals(Command.SEARCHFILES, searchfiles.getCommand(), "command");
    assertEquals(RESPONSE, searchfiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, searchfiles.getDataset(), "dataset");
    assertNull(searchfiles.getUserid(), "USERID");
    assertEquals(true, searchfiles.hasStarted(), "start");

    assertEquals("abc.123", files[0].getFilename(), "files1");
    assertEquals(1, files[0].getIndex(), "index1");
    assertEquals("hallo.456", files[1].getFilename(), "files2");
    assertEquals(2, files[1].getIndex(), "index2");
  }



  @Test
  public void decodeSEARCHFILES_RESPONSE2_JERSEY() {

    String encoded = "{\"HEADER\":\"RESPONSE\",\"COMMAND\":\"SEARCHFILES\",\"DATASET\":{\"START\":true,\"SEARCHFILE\":[{\"INDEX\":0,\"FILENAME\":\"1.mp4\",\"FILESIZE\":666},{\"INDEX\":1,\"FILENAME\":\"2.mp4\",\"FILESIZE\":666},{\"INDEX\":2,\"FILENAME\":\"3.mp4\",\"FILESIZE\":666}]}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SEARCHFILES.class, new SearchFilesDeserializer())
        .create();

    SEARCHFILES result = gson.fromJson(encoded, SEARCHFILES.class);

    Searchfile[] files = result.getSearchfiles();

    assertEquals(3, result.getSearchfiles().length, "3 Dateien");

  }



  @Test
  public void decodeSIGNIN_CONFIRM() throws DecodeException, EncodeException, IOException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99"
        + "}}";

    SIGNIN result = (SIGNIN) messagedecoder.decode(encoded);
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertEquals("password", result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "ip");
    assertTrue(true);
  }



  @Test
  public void decodeSIGNIN_CONFIRM_JERSEY() {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99"
        + "}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertEquals("password", result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "background");
    assertTrue(true);
  }



  @Test
  @DisplayName("SIGIN CONFIRM MULTILINGUALKEY")
  public void decodeSIGNIN_CONFIRM_MULTILINGUALKEY() throws DecodeException, EncodeException, IOException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"MULTILINGUALKEY\":\"Server_confirm_Passwort\""
        + "}}";

    SIGNIN result = (SIGNIN) messagedecoder.decode(encoded);
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertEquals("password", result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "ip");
    assertEquals(KEY.SERVER_CONFIRM_PASSWORT, result.getMultilingualkey(), "MULTILINGUALKEY");
    assertTrue(true);
  }



  @Test
  @DisplayName("SIGIN CONFIRM MULTILINGUALKEY JERSEY")
  public void decodeSIGNIN_CONFIRM_MULTILINGUALKEY_JERSEY() {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"MULTILINGUALKEY\":\"Server_confirm_Passwort\""
        + "}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(CONFIRM, result.getHeader(), "HEADER");
    assertEquals(Command.SIGNIN, result.getCommand(), "COMMAND");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertEquals("password", result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "USERID");
    assertEquals("nickname", result.getNickname(), "NICKNAME");
    assertEquals("localhost", result.getIp(), "IP");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "BACKGROUND_COLOR");
    assertEquals(KEY.SERVER_CONFIRM_PASSWORT, result.getMultilingualkey(), "MULTILINGUALKEY");
    assertTrue(true);
  }



  @Test
  public void decodeSIGNIN_CONFIRM_ohne_IP() throws JsonSyntaxException, UnsupportedEncodingException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"llange@javacomm.net\","
        + "\"USERID\":\"louisxiv\","
        + "\"PASSWORD\":\"PASSWORD\","
        + "\"FOREGROUND_COLOR\":-16777216,"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"NICKNAME\":\"anaconda\"}}";

    SIGNIN result;
    try {
      result = (SIGNIN) messagedecoder.decodeSIGNIN(encoded);
      fail("DecodeException wurde nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeSIGNIN_CONFIRM_TEXT() throws DecodeException, EncodeException, IOException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"TEXT\":\"ich bin ein Test\""
        + "}}";

    SIGNIN result = (SIGNIN) messagedecoder.decode(encoded);
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertEquals("password", result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "ip");
    assertEquals("ich bin ein Test", result.getText(), "TEXT");
    assertTrue(true);
  }



  @Test
  public void decodeSIGNIN_CONFIRM_TEXT_JERSEY() {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"TEXT\":\"ich bin ein Test\""
        + "}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertEquals("password", result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "ip");
    assertEquals("ich bin ein Test", result.getText(), "TEXT");
    assertTrue(true);
  }



  @Test
  public void decodeSIGNIN_ERROR() throws EncodeException, JsonSyntaxException, UnsupportedEncodingException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":{"
        + "\"EMAIL\":\"llange@javacomm.net\","
        + "\"ERRORMESSAGE\":\"123\","
        + "\"TEXT\":\"User could not be added.\""
        + "}}";

    SIGNIN result;
    try {
      result = (SIGNIN) messagedecoder.decodeSIGNIN(encoded);
      fail("EncodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeSIGNIN_ERROR_ohne_Text() throws DecodeException, EncodeException, IOException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":{"
        + "\"EMAIL\":\"llange@javacomm.net\","
        + "\"ERRORCODE\":33,"
        + "\"ERRORMESSAGE\":\"123\""
        + "}}";

    SIGNIN result = (SIGNIN) messagedecoder.decodeSIGNIN(encoded);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("llange@javacomm.net", result.getEmail(), "email");
    assertEquals(33, result.getErrorCode(), "errorcode");
    assertEquals("123", result.getErrorMessage(), "123");

  }



  @Test
  public void decodeSIGNIN_ERROR_ohne_Text_JERSEY() {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":{"
        + "\"EMAIL\":\"llange@javacomm.net\","
        + "\"ERRORCODE\":33,"
        + "\"ERRORMESSAGE\":\"123\""
        + "}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("llange@javacomm.net", result.getEmail(), "email");
    assertEquals(33, result.getErrorCode(), "ERRORCODE");
    assertEquals("123", result.getErrorMessage(), "ERRORMESSAGE");

  }



  @Test
  public void decodeSIGNIN_ERROR_URL() throws DecodeException, EncodeException, IOException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":{\"EMAIL\":\"abcd\","
        + "\"ERRORCODE\":0,"
        + "\"ERRORMESSAGE\":\"Die EMail-Adresse ist ungültig.\","
        + "\"TEXT\":\"Deine Email-Adresse ist ung%C3%BCltig.\"}}";

    SIGNIN result = (SIGNIN) messagedecoder.decodeSIGNIN(encoded);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("abcd", result.getEmail(), "email");
    assertEquals(0, result.getErrorCode(), "errorcode");
    assertEquals(
        "Die EMail-Adresse ist ungültig.", result.getErrorMessage(), "Die EMail-Adresse ist ungültig."
    );
    assertEquals("Deine Email-Adresse ist ung%C3%BCltig.", result.getText(), "text");

  }



  @Test
  public void decodeSIGNIN_ERROR_URL_JERSEY() {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":{\"EMAIL\":\"abcd\","
        + "\"ERRORCODE\":0,"
        + "\"ERRORMESSAGE\":\"Die EMail-Adresse ist ungültig.\","
        + "\"TEXT\":\"Deine Email-Adresse ist ung%C3%BCltig.\"}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("abcd", result.getEmail(), "email");
    assertEquals(0, result.getErrorCode(), "errorcode");
    assertEquals(
        "Die EMail-Adresse ist ungültig.", result.getErrorMessage(), "Die EMail-Adresse ist ungültig."
    );
    assertEquals("Deine Email-Adresse ist ung%C3%BCltig.", result.getText(), "text");

  }



  @Test
  public void decodeSIGNIN_ERROR2() throws DecodeException, EncodeException, IOException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"llange@javacomm.net\","
        + "\"ERRORCODE\":33,"
        + "\"ERRORMESSAGE\":\"123\","
        + "\"TEXT\":\"User could not be added.\"}}";

    SIGNIN result = (SIGNIN) messagedecoder.decodeSIGNIN(encoded);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("llange@javacomm.net", result.getEmail(), "email");
    assertEquals(33, result.getErrorCode(), "errorcode");
    assertEquals("123", result.getErrorMessage(), "errormessage");
    assertEquals("User could not be added.", result.getText(), "text");

  }



  @Test
  public void decodeSIGNIN_ERROR2_JERSEY() {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"llange@javacomm.net\","
        + "\"ERRORCODE\":33,"
        + "\"ERRORMESSAGE\":\"123\","
        + "\"TEXT\":\"User could not be added.\"}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("llange@javacomm.net", result.getEmail(), "email");
    assertEquals(33, result.getErrorCode(), "errorcode");
    assertEquals("123", result.getErrorMessage(), "errormessage");
    assertEquals("User could not be added.", result.getText(), "text");

  }



  @Test
  public void decodeSIGNIN_ERROR3()
      throws DecodeException, EncodeException, JsonSyntaxException, UnsupportedEncodingException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"\","
        + "\"ERRORCODE\":1062,"
        + "\"ERRORMESSAGE\":\"Duplicate entry '' for key 'MAIL'\","
        + "\"TEXT\":\"User could not be added.\"}}";

    SIGNIN result = (SIGNIN) messagedecoder.decodeSIGNIN(encoded);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("", result.getEmail(), "email");
    assertEquals(1062, result.getErrorCode(), "errorcode");
    assertEquals("Duplicate entry '' for key 'MAIL'", result.getErrorMessage(), "errormessage");
    assertEquals("User could not be added.", result.getText(), "text");

  }



  @Test
  public void decodeSIGNIN_ERROR3_JERSEY() {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"\","
        + "\"ERRORCODE\":1062,"
        + "\"ERRORMESSAGE\":\"Duplicate entry '' for key 'MAIL'\","
        + "\"TEXT\":\"User could not be added.\"}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("", result.getEmail(), "email");
    assertEquals(1062, result.getErrorCode(), "errorcode");
    assertEquals("Duplicate entry '' for key 'MAIL'", result.getErrorMessage(), "errormessage");
    assertEquals("User could not be added.", result.getText(), "text");

  }



  @Test
  @DisplayName("die Email Adresse ist ungültig")
  public void decodeSIGNIN_MULTILINGUALKEY()
      throws DecodeException, EncodeException, JsonSyntaxException, UnsupportedEncodingException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"\","
        + "\"ERRORCODE\":1062,"
        + "\"ERRORMESSAGE\":\"Duplicate entry '' for key 'MAIL'\","
        + "\"TEXT\":\"User could not be added.\","
        + "\"MULTILINGUALKEY\":\"Server_Email_invalid\""
        + "}}";

    SIGNIN result = (SIGNIN) messagedecoder.decodeSIGNIN(encoded);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("", result.getEmail(), "email");
    assertEquals(1062, result.getErrorCode(), "errorcode");
    assertEquals("Duplicate entry '' for key 'MAIL'", result.getErrorMessage(), "errormessage");
    assertEquals("User could not be added.", result.getText(), "text");
    assertEquals(KEY.SERVER_EMAIL_INVALID, result.getMultilingualkey(), "MULTILINGUALKEY");
  }



  @Test
  @DisplayName("die Email Adresse ist ungültig Jersey")
  public void decodeSIGNIN_MULTILINGUALKEY_JERSEY() {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"\","
        + "\"ERRORCODE\":1062,"
        + "\"ERRORMESSAGE\":\"Duplicate entry '' for key 'MAIL'\","
        + "\"TEXT\":\"User could not be added.\","
        + "\"MULTILINGUALKEY\":\"Server_Email_invalid\""
        + "}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals(ERROR, result.getHeader(), "header");
    assertEquals("", result.getEmail(), "email");
    assertEquals(1062, result.getErrorCode(), "errorcode");
    assertEquals("Duplicate entry '' for key 'MAIL'", result.getErrorMessage(), "errormessage");
    assertEquals("User could not be added.", result.getText(), "text");
    assertEquals(KEY.SERVER_EMAIL_INVALID, result.getMultilingualkey(), "MULTILINGUALKEY");
  }



  @Test
  public void decodeSIGNIN_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":{"
        + "\"EMAIL\":\"llange@javacomm.net\","
        + "\"IP\":\"localhost\""
        + "}}";

    SIGNIN signin = (SIGNIN) messagedecoder.decode(encoded);
    assertEquals(Command.SIGNIN, signin.getCommand(), "command");
    assertEquals(REQUEST, signin.getHeader(), "header");
    assertEquals(Protocol.DATASET, signin.getDataset(), "dataset");
    assertEquals("llange@javacomm.net", signin.getEmail(), "email");
    assertEquals("localhost", signin.getIp(), "room");
  }



  @Test
  public void decodeSIGNIN_REQUEST_ohne_IP() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":{"
        + "\"EMAIL\":\"llange@javacomm.net\""
        + "}}";

    SIGNIN signin = (SIGNIN) messagedecoder.decode(encoded);
    assertEquals(Command.SIGNIN, signin.getCommand(), "command");
    assertEquals(REQUEST, signin.getHeader(), "header");
    assertEquals(Protocol.DATASET, signin.getDataset(), "dataset");
    assertEquals("llange@javacomm.net", signin.getEmail(), "email");
    assertNull(signin.getIp(), "IP");

  }



  @Test
  public void decodeSIGNIN_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"NEUES_PASSWORD\":\"pass1234\","
        + "\"CONFIRMATION_KEY\":\"key12345\","
        + "\"MULTILINGUALKEY\":\"Server_neues_Passwort\""
        + "}}";

    SIGNIN result = (SIGNIN) messagedecoder.decode(encoded);
    assertEquals(RESPONSE, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertEquals("password", result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "ip");
    assertEquals("pass1234", result.getNeuesPassword(), "neuesPassword");
    assertEquals("key12345", result.getConfirmationKey(), "confirmationKey");
    assertTrue(true);
  }



  @Test
  public void decodeSIGNIN_RESPONSE_JERSEY() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"NEUES_PASSWORD\":\"pass1234\","
        + "\"CONFIRMATION_KEY\":\"key12345\","
        + "\"MULTILINGUALKEY\":\"Server_neues_Passwort\""
        + "}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(RESPONSE, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertEquals("password", result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "ip");
    assertEquals("pass1234", result.getNeuesPassword(), "neuesPassword");
    assertEquals("key12345", result.getConfirmationKey(), "confirmationKey");
    assertTrue(true);
  }



  @Test
  public void decodeSIGNIN_RESPONSE_ohne_MULTILINGUALKEY() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"PASSWORD\":\"password\","
        + "\"USERID\":\"userid\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"NEUES_PASSWORD\":\"pass1234\","
        + "\"CONFIRMATION_KEY\":\"key12345\""
        + "}}";

    try {
      messagedecoder.decode(encoded);
      fail("Exception wurde nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeSIGNIN_RESPONSE_ohne_PASSWORD_ohne_IP()
      throws DecodeException, EncodeException, IOException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"USERID\":\"userid\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"NEUES_PASSWORD\":\"pass1234\","
        + "\"CONFIRMATION_KEY\":\"key12345\","
        + "\"MULTILINGUALKEY\":\"Server_neues_Passwort\""
        + "}}";

    SIGNIN result = (SIGNIN) messagedecoder.decode(encoded);
    assertEquals(RESPONSE, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertNull(result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertNull(result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "ip");
    assertEquals("pass1234", result.getNeuesPassword(), "neuesPassword");
    assertEquals("key12345", result.getConfirmationKey(), "confirmationKey");
    assertTrue(true);
  }



  @Test
  public void decodeSIGNIN_RESPONSE_ohne_PASSWORD_ohne_IP_JERSEY() {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"SIGNIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"team@javacomm.net\","
        + "\"NICKNAME\":\"nickname\","
        + "\"USERID\":\"userid\","
        + "\"FOREGROUND_COLOR\":33,"
        + "\"BACKGROUND_COLOR\":99,"
        + "\"NEUES_PASSWORD\":\"pass1234\","
        + "\"CONFIRMATION_KEY\":\"key12345\","
        + "\"MULTILINGUALKEY\":\"Server_neues_Passwort\""
        + "}}";

    Gson gson = new GsonBuilder().registerTypeAdapter(SIGNIN.class, new SigninDeserializer()).create();
    SIGNIN result = gson.fromJson(encoded, SIGNIN.class);
    assertEquals(RESPONSE, result.getHeader(), "header");
    assertEquals(Command.SIGNIN, result.getCommand(), "command");
    assertEquals("team@javacomm.net", result.getEmail(), "eMail");
    assertNull(result.getPassword(), "password");
    assertEquals("userid", result.getUserid(), "userid");
    assertEquals("nickname", result.getNickname(), "nickname");
    assertNull(result.getIp(), "ip");
    assertEquals(33, result.getForegroundColor(), "foreground");
    assertEquals(99, result.getBackgroundColor(), "ip");
    assertEquals("pass1234", result.getNeuesPassword(), "neuesPassword");
    assertEquals("key12345", result.getConfirmationKey(), "confirmationKey");
    assertTrue(true);
  }



  @Test
  public void decodeSIGNIN_Schlechtfall_COMMAND_USRLOGIN() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"USRLOGIN\",\"DATASET\":{\"EMAIL\":\"team@javacomm.net\"}}";
    try {
      MESSAGE result = messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeSIGNIN_Schlechtfall_kein_DATASET() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"SIGNIN\"}";
    try {
      MESSAGE result = messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeSIGNIN_Schlechtfall_keine_EMAIL() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"SIGNIN\",\"DATASET\":{}}";
    try {
      MESSAGE result = messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeSIGNIN_Schlechtfall_nur_String() {
    String encoded = "";
    try {
      messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeTCPAUDIO() throws DecodeException, UnsupportedEncodingException {

    final String polyfon = "Polyfon zwitschernd aßen Mäxchens Vögel Rüben, Joghurt und Quark";
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"TCPAUDIO\","
        + "\"DATASET\":{"
        + "\"TARGET\":\"8c\","
        + "\"ORIGIN\":\"7b\","
        + "\"CONTENT\":[80,111,108,121,102,111,110,32,122,119,105,116,115,99,104,101,114,110,100,32,97,-61,-97,101,110,32,77,-61,-92,120,99,104,101,110,115,32,86,-61,-74,103,101,108,32,82,-61,-68,98,101,110,44,32,74,111,103,104,117,114,116,32,117,110,100,32,81,117,97,114,107]"
        + "}}";

    TCPAUDIO result = (TCPAUDIO) messagedecoder.decodeTCPAUDIO(encoded);
    assertEquals(REQUEST, result.getHeader());
    assertEquals(Command.TCPAUDIO, result.getCommand());
    assertEquals(Protocol.DATASET, result.getDataset(), "dataset");
    assertEquals("8c", result.getTarget(), "TARGET");
    assertEquals("7b", result.getOrigin(), "ORIGIN");
    assertEquals(polyfon, new String(result.getContent()));
  }



  @Test
  public void decodeTOPICMEMBER_Exception() {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"TOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\","
        + "\"TOPICMEMBER\":["
        + "{\"USERID\":\"jester\",\"NICKNAME\":\"anaconda\"},"
        + "{\"USERID\":\"luis\",\"NICKNAME\":\"viper\""
        + "}]"
        + "}}";

    try {
      TOPICMEMBER result = (TOPICMEMBER) messagedecoder.decode(encoded);
      fail("DecodeException wurde nicht geworfen, ROOMTYPE fehlt");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeTOPICMEMBER_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"TOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\"}"
        + "}";

    TOPICMEMBER result = (TOPICMEMBER) messagedecoder.decode(encoded);
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("fussball", result.getRoom(), "room");

  }



  @Test
  public void decodeTOPICMEMBER_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"TOPICMEMBER\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"ROOM\":\"fussball\","
        + "\"ROOMTYPE\":\"FORUM\","
        + "\"TOPICMEMBER\":["
        + "{\"USERID\":\"jester\",\"NICKNAME\":\"anaconda\"},"
        + "{\"USERID\":\"luis\",\"NICKNAME\":\"viper\""
        + "}]"
        + "}}";

    TOPICMEMBER result = (TOPICMEMBER) messagedecoder.decode(encoded);
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("fussball", result.getRoom(), "room");
    ChatUser[] chatuser = result.getChatUser();
    assertEquals(2, chatuser.length, "2 Datensätze");
    assertEquals("jester", chatuser[0].getUserid(), "userid0");
    assertEquals("anaconda", chatuser[0].getNickname(), "nickname0");
    assertEquals("luis", chatuser[1].getUserid(), "userid1");
    assertEquals("viper", chatuser[1].getNickname(), "nickname1");
    assertEquals("FORUM", result.getRoomtype().toString(), "FORUM");

  }



  @Test
  public void decodeTRANSFER_ERROR() throws DecodeException {

    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"TRANSFER\","
        + "\"DATASET\":{\"SLOT\":1,"
        + "\"FILENAME\":\"hallo.txt\","
        + "\"PATHFILE\":\"/root/path/to/file\","
        + "\"ERRORMESSAGE\":\"Die Datei ist ungültig\"}}";

    TRANSFER transfer = (TRANSFER) messagedecoder.decode(encoded);

    assertEquals(Command.TRANSFER, transfer.getCommand(), "command");
    assertEquals(ERROR, transfer.getHeader(), "header");
    assertEquals(Protocol.DATASET, transfer.getDataset(), "dataset");
    assertEquals(1, transfer.getSlot(), "slot");
    assertEquals("hallo.txt", transfer.getFilename(), "filename");
    assertEquals("/root/path/to/file", transfer.getPathfile(), "pathfile");
    assertEquals("Die Datei ist ungültig", transfer.getErrorMessage(), "errorMessage");

  }



  @Test
  public void decodeTRANSFER_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"TRANSFER\","
        + "\"DATASET\":{"
        + "\"SLOT\":1,"
        + "\"BLOCKINDEX\":33,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"PATHFILE\":\"c:/\","
        + "\"MAXDATABLOCKS\":848"
        + "}"
        + "}";

    TRANSFER transfer = (TRANSFER) messagedecoder.decode(encoded);

    assertEquals(Command.TRANSFER, transfer.getCommand(), "command");
    assertEquals(REQUEST, transfer.getHeader(), "header");
    assertEquals(Protocol.DATASET, transfer.getDataset(), "dataset");
    assertEquals(1, transfer.getSlot(), "slot");
    assertEquals("hallo.mp3", transfer.getFilename(), "filename");
    assertEquals("c:/", transfer.getPathfile(), "pathfile");

    assertEquals(33, transfer.getBlockindex(), "blockindex");
    assertEquals(848, transfer.getMaxDatablocks(), "maxdatablcoks");

  }



  @Test
  public void decodeTRANSFER_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"TRANSFER\","
        + "\"DATASET\":{"
        + "\"SLOT\":1,"
        + "\"BLOCKINDEX\":33,"
        + "\"FILENAME\":\"hallo.mp3\","
        + "\"PATHFILE\":\"c:/\","
        + "\"ENDFLAG\":true,"
        + "\"PAYLOAD\":\"273237\","
        + "\"MAXDATABLOCKS\":848"
        + "}"
        + "}";

    TRANSFER transfer = (TRANSFER) messagedecoder.decode(encoded);

    assertEquals(Command.TRANSFER, transfer.getCommand(), "command");
    assertEquals(RESPONSE, transfer.getHeader(), "header");
    assertEquals(Protocol.DATASET, transfer.getDataset(), "dataset");
    assertEquals(1, transfer.getSlot(), "slot");
    assertEquals("hallo.mp3", transfer.getFilename(), "filename");
    assertEquals("c:/", transfer.getPathfile(), "pathfile");

    assertEquals(33, transfer.getBlockindex(), "blockindex");
    assertEquals(true, transfer.getEndflag(), "endflag");
    assertEquals("273237", transfer.getPayload(), "pathfile");
    assertEquals(848, transfer.getMaxDatablocks(), "maxdatablocks");

  }



  @Test
  public void decodeUPDATEFILETYPES_CONFIRM() throws DecodeException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"UPDATEFILETYPES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"userid123\""
        + "}"
        + "}";

    UPDATEFILETYPES updatetypes = (UPDATEFILETYPES) messagedecoder.decode(encoded);
    assertEquals(Command.UPDATEFILETYPES, updatetypes.getCommand(), "Command");
    assertEquals(CONFIRM, updatetypes.getHeader(), "CONFIRM");
    assertEquals("userid123", updatetypes.getUserid(), "USERID");
  }



  @Test
  public void decodeUPDATEFILETYPES_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPDATEFILETYPES\","
        + "\"DATASET\":{"
        + "\"USERID\":\"userid123\","
        + "\"FILETYPES\":["
        + "{\"SUFFIX\":\"pdf\",\"DESCRIPTION\":\"Adobe\",CHECKED:true},"
        + "{\"SUFFIX\":\"txt\",\"DESCRIPTION\":\"Simple Text Format\",CHECKED:false}"
        + "]}"
        + "}";

    UPDATEFILETYPES filetypes = (UPDATEFILETYPES) messagedecoder.decode(encoded);
    assertEquals(Command.UPDATEFILETYPES, filetypes.getCommand(), "Command");
    assertEquals(REQUEST, filetypes.getHeader(), "REQUEST");
    assertEquals("userid123", filetypes.getUserid(), "USERID");

    BlacklistTypes pdf = filetypes.getFiletypes()[0];
    BlacklistTypes txt = filetypes.getFiletypes()[1];

    assertEquals("pdf", pdf.getSuffix(), "pdfSuffix");
    assertEquals("Adobe", pdf.getDescription(), "Adobe");
    assertEquals(true, pdf.getChecked(), "pdfChecked");

    assertEquals("txt", txt.getSuffix(), "txtSuffix");
    assertEquals("Simple Text Format", txt.getDescription(), "Simple Text Format");
    assertEquals(false, txt.getChecked(), "txtChecked");
  }



  @Test
  public void decodeUPDATEPHONE_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"UPDATEPHONE\"" + "}";

    UPDATEPHONE phone = (UPDATEPHONE) messagedecoder.decode(encoded);
    assertEquals(Command.UPDATEPHONE, phone.getCommand(), "COMMAND");
    assertEquals(ERROR, phone.getHeader(), "HEADER");

  }



  @Test
  public void decodeUPDATEPHONE_REQUEST() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPDATEPHONE\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"IP\":\"127.0.0.1\","
        + "\"PORT\":29477,"
        + "\"ONCALL\":false,"
        + "\"ONBUSY\":false"
        + "}}";

    UPDATEPHONE phone = (UPDATEPHONE) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATEPHONE, phone.getCommand(), "COMMAND");
    assertEquals(REQUEST, phone.getHeader(), "HEADER");
    assertEquals("12345678", phone.getUserid(), "userid");
    assertEquals("127.0.0.1", phone.getIp(), "ip");
    assertEquals(29477, phone.getPort(), "port");
    assertEquals(false, phone.isOnBusy(), "onBusy");
  }



  @Test
  public void decodeUPDATEPHONE_REQUEST_OHNE_PORT() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPDATEPHONE\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ONCALL\":false,"
        + "\"ONBUSY\":false"
        + "}}";

    UPDATEPHONE phone = (UPDATEPHONE) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATEPHONE, phone.getCommand(), "COMMAND");
    assertEquals(REQUEST, phone.getHeader(), "HEADER");
    assertEquals("12345678", phone.getUserid(), "userid");
    assertEquals(false, phone.isOnBusy(), "onBusy");
  }



  @Test
  public void decodeUPDATEPHONE_REQUEST_ONCALL() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPDATEPHONE\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"IP\":\"127.0.0.1\","
        + "\"PORT\":29477,"
        + "\"ONCALL\":true,"
        + "\"ONBUSY\":true"
        + "}}";

    UPDATEPHONE phone = (UPDATEPHONE) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATEPHONE, phone.getCommand(), "COMMAND");
    assertEquals(REQUEST, phone.getHeader(), "HEADER");
    assertEquals("12345678", phone.getUserid(), "userid");
    assertEquals("127.0.0.1", phone.getIp(), "ip");
    assertEquals(29477, phone.getPort(), "port");
    assertEquals(true, phone.isOnBusy(), "onBusy");
  }



  @Test
  public void decodeUPDATEPHONE_REQUEST_Schlechtfall() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPDATEPHONE\","
        + "\"DATASET\":{"
        + "\"IP\":\"127.0.0.1\","
        + "\"PORT\":29477"
        + "}}";

    try {
      UPDATEPHONE phone = (UPDATEPHONE) messagedecoder.decode(encoded);
      fail("Exception nicht geworfen");
    }
    catch (DecodeException e) {
      assertEquals("Das Attribut USERID wurde nicht definiert.", e.getMessage());
    }

  }



  @Test
  public void decodeUPDATEPHONE_RESPONSE() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"UPDATEPHONE\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"IP\":\"93.0.0.1\","
        + "\"PORT\":29467,"
        + "\"ONCALL\":true,"
        + "\"ONBUSY\":true"
        + "}}";

    UPDATEPHONE phone = (UPDATEPHONE) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATEPHONE, phone.getCommand(), "COMMAND");
    assertEquals(RESPONSE, phone.getHeader(), "HEADER");
    assertEquals("12345678", phone.getUserid(), "userid");
    assertEquals("93.0.0.1", phone.getIp(), "ip");
    assertEquals(29467, phone.getPort(), "port");
    assertEquals(true, phone.isOnBusy(), "onbusy");

  }



  @Test
  public void decodeUPDATEPHONE_RESPONSE_OHNE_ONCALL() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"UPDATEPHONE\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ONBUSY\":true"
        + "}}";

    UPDATEPHONE phone = (UPDATEPHONE) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATEPHONE, phone.getCommand(), "COMMAND");
    assertEquals(RESPONSE, phone.getHeader(), "HEADER");
    assertEquals("12345678", phone.getUserid(), "userid");
    assertEquals(true, phone.isOnBusy(), "onbusy");

  }



  @Test
  public void decodeUPDATEPHONE_RESPONSE_OHNE_PORT() throws DecodeException {
    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"UPDATEPHONE\","
        + "\"DATASET\":{"
        + "\"USERID\":\"12345678\","
        + "\"ONCALL\":true,"
        + "\"ONBUSY\":true"
        + "}}";

    UPDATEPHONE phone = (UPDATEPHONE) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATEPHONE, phone.getCommand(), "COMMAND");
    assertEquals(RESPONSE, phone.getHeader(), "HEADER");
    assertEquals("12345678", phone.getUserid(), "userid");
    assertEquals(true, phone.isOnBusy(), "onbusy");

  }



  @Test
  public void decodeUPDATEPORT_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPDATEPORT\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"localhost\","
        + "\"PORT\":8443"
        + "}}";

    UPDATEPORT updateport = (UPDATEPORT) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATEPORT, updateport.getCommand(), "command");
    assertEquals(REQUEST, updateport.getHeader(), "header");
    assertEquals(Protocol.DATASET, updateport.getDataset(), "dataset");
    assertEquals("louisxiv", updateport.getUserid(), "userid");
    assertEquals("localhost", updateport.getIp(), "ip");
    assertEquals(8443, updateport.getPort(), "port");
  }



  @Test
  public void decodeUPDATEPORT_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"UPDATEPORT\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"localhost\","
        + "\"PORT\":8443"
        + "}}";

    UPDATEPORT updateport = (UPDATEPORT) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATEPORT, updateport.getCommand(), "command");
    assertEquals(RESPONSE, updateport.getHeader(), "header");
    assertEquals(Protocol.DATASET, updateport.getDataset(), "dataset");
    assertEquals("louisxiv", updateport.getUserid(), "userid");
    assertEquals("localhost", updateport.getIp(), "ip");
    assertEquals(8443, updateport.getPort(), "port");
  }



  @Test
  public void decodeUPDATEUSER_CONFIRM() {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"UPDATE_USER\","
        + "\"DATASET\":{"
        + "\"FOREGROUND_COLOR\":3,"
        + "\"BACKGROUND_COLOR\":4,"
        + "\"NICKNAME\":\"anaconda\","
        + "\"EMAIL\":\"llange@123\","
        + "\"USERID\":\"louisxiv\","
        + "\"PASSWORD\":\"geheim\""
        + "}}";

    UPDATEUSER updateuser;
    try {
      updateuser = (UPDATEUSER) messagedecoder.decode(encoded);
      fail("DecodeException wurde nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }

  }



  @Test
  public void decodeUPDATEUSER_CONFIRM_mit_IP() throws DecodeException {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"UPDATE_USER\","
        + "\"DATASET\":{"
        + "\"FOREGROUND_COLOR\":3,"
        + "\"BACKGROUND_COLOR\":4,"
        + "\"NICKNAME\":\"anaconda\","
        + "\"EMAIL\":\"llange@123\","
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"localhost\","
        + "\"PASSWORD\":\"geheim\","
        + "\"VOLUME\":68,"
        + "\"ONCALL\":true,"
        + "\"TELEFONPORT\":29475"
        + "}}";

    UPDATEUSER updateuser = (UPDATEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATE_USER, updateuser.getCommand(), "command");
    assertEquals(CONFIRM, updateuser.getHeader(), "header");
    assertEquals(Protocol.DATASET, updateuser.getDataset(), "dataset");
    assertEquals("louisxiv", updateuser.getUserid(), "userid");
    assertEquals(4, updateuser.getBackgroundColor(), "bgColor");
    assertEquals(3, updateuser.getForegroundColor(), "fgColor");
    assertEquals("llange@123", updateuser.getEmail(), "email");
    assertEquals("anaconda", updateuser.getNickname(), "nickname");
    assertEquals("geheim", updateuser.getPassword(), "password");
    assertEquals("localhost", updateuser.getIp(), "ip");

  }



  @Test
  public void decodeUPDATEUSER_CONFIRM_mit_IP_VOLUME_TELEFONPORT() throws DecodeException {
    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"UPDATE_USER\","
        + "\"DATASET\":{"
        + "\"FOREGROUND_COLOR\":3,"
        + "\"BACKGROUND_COLOR\":4,"
        + "\"NICKNAME\":\"anaconda\","
        + "\"EMAIL\":\"llange@123\","
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"localhost\","
        + "\"PASSWORD\":\"geheim\","
        + "\"VOLUME\":68,"
        + "\"ONCALL\":true,"
        + "\"TELEFONPORT\":29477"

        + "}}";

    UPDATEUSER updateuser = (UPDATEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATE_USER, updateuser.getCommand(), "command");
    assertEquals(CONFIRM, updateuser.getHeader(), "header");
    assertEquals(Protocol.DATASET, updateuser.getDataset(), "dataset");
    assertEquals("louisxiv", updateuser.getUserid(), "userid");
    assertEquals(4, updateuser.getBackgroundColor(), "bgColor");
    assertEquals(3, updateuser.getForegroundColor(), "fgColor");
    assertEquals("llange@123", updateuser.getEmail(), "email");
    assertEquals("anaconda", updateuser.getNickname(), "nickname");
    assertEquals("geheim", updateuser.getPassword(), "password");
    assertEquals("localhost", updateuser.getIp(), "ip");
    assertEquals(68, updateuser.getVolume(), "volume");
    assertEquals(true, updateuser.isOncall(), "oncall");
  }



  @Test
  public void decodeUPDATEUSER_ERROR() throws DecodeException {
    String encoded = "{\"HEADER\":\"ERROR\"," + "\"COMMAND\":\"UPDATE_USER\"" + "}";

    UPDATEUSER updateuser = (UPDATEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATE_USER, updateuser.getCommand(), "command");
    assertEquals(ERROR, updateuser.getHeader(), "header");
    assertNull(updateuser.getDataset());
  }



  @Test
  public void decodeUPDATEUSER_REQUEST() {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPDATE_USER\","
        + "\"DATASET\":{"
        + "\"FOREGROUND_COLOR\":3,"
        + "\"BACKGROUND_COLOR\":4,"
        + "\"NICKNAME\":\"anaconda\","
        + "\"EMAIL\":\"llange@123\","
        + "\"USERID\":\"louisxiv\","
        + "\"PASSWORD\":\"geheim\""
        + "}}";

    UPDATEUSER updateuser;
    try {
      updateuser = (UPDATEUSER) messagedecoder.decode(encoded);
      fail("DecodeException wurde nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeUPDATEUSER_REQUEST_mit_IP() throws DecodeException {
    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPDATE_USER\","
        + "\"DATASET\":{"
        + "\"FOREGROUND_COLOR\":3,"
        + "\"BACKGROUND_COLOR\":4,"
        + "\"NICKNAME\":\"anaconda\","
        + "\"EMAIL\":\"llange@123\","
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"localhost\","
        + "\"PASSWORD\":\"geheim\""
        + "}}";

    UPDATEUSER updateuser = (UPDATEUSER) messagedecoder.decode(encoded);

    assertEquals(Command.UPDATE_USER, updateuser.getCommand(), "command");
    assertEquals(REQUEST, updateuser.getHeader(), "header");
    assertEquals(Protocol.DATASET, updateuser.getDataset(), "dataset");
    assertEquals("louisxiv", updateuser.getUserid(), "userid");
    assertEquals(4, updateuser.getBackgroundColor(), "bgColor");
    assertEquals(3, updateuser.getForegroundColor(), "fgColor");
    assertEquals("llange@123", updateuser.getEmail(), "email");
    assertEquals("anaconda", updateuser.getNickname(), "nickname");
    assertEquals("geheim", updateuser.getPassword(), "password");
    assertEquals("localhost", updateuser.getIp(), "ip");

  }



  @Test
  public void decodeUPLOADFILES_REQUEST() {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPLOADFILES\","
        + "\"DATASET\":{"
        + "\"START\":true,"
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"192.168.1.41\","
        + "\"UPLOADFILES\":["
        + "{\"FILE\":\"abc.123\"},"
        + "{\"FILE\":\"hallo.456\"}]}}";

    try {
      UPLOADFILES uploadfiles = (UPLOADFILES) messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {}

  }



  @Test
  public void decodeUPLOADFILES_REQUEST_PORT() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"UPLOADFILES\","
        + "\"DATASET\":{"
        + "\"START\":true,"
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"192.168.1.41\","
        + "\"PORT\":8443,"
        + "\"UPLOADFILES\":["
        + "{\"FILENAME\":\"abc.txt\","
        + "\"FILESIZE\":1000},"
        + "{\"FILENAME\":\"zwei.txt\","
        + "\"FILESIZE\":1234}"
        + "]}}";

    UPLOADFILES uploadfiles = (UPLOADFILES) messagedecoder.decode(encoded);
    Uploadfile[] files = uploadfiles.getUploadfile();

    assertEquals(Command.UPLOADFILES, uploadfiles.getCommand(), "command");
    assertEquals(REQUEST, uploadfiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, uploadfiles.getDataset(), "dataset");
    assertEquals("louisxiv", uploadfiles.getUserid(), "userid");
    assertEquals("192.168.1.41", uploadfiles.getIp(), "ip");
    assertEquals("abc.txt", files[0].getFilename(), "filename");
    assertEquals(1000, files[0].getFilesize(), "filesize");
    assertEquals("zwei.txt", files[1].getFilename(), "filename2");
    assertEquals(1234, files[1].getFilesize(), "filesize2");
    assertEquals(true, uploadfiles.hasStarted(), "strat");
    assertEquals(8443, uploadfiles.getPort(), "port");
  }



  @Test
  public void decodeUPLOADFILES_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"UPLOADFILES\","
        + "\"DATASET\":{"
        + "\"START\":true,"
        + "\"USERID\":\"louisxiv\","
        + "\"IP\":\"192.168.1.41\""
        + "}}";

    UPLOADFILES uploadfiles = (UPLOADFILES) messagedecoder.decode(encoded);
    Uploadfile[] files = uploadfiles.getUploadfile();

    assertEquals(Command.UPLOADFILES, uploadfiles.getCommand(), "command");
    assertEquals(RESPONSE, uploadfiles.getHeader(), "header");
    assertEquals(Protocol.DATASET, uploadfiles.getDataset(), "dataset");
    assertEquals("louisxiv", uploadfiles.getUserid(), "userid");
    assertEquals("192.168.1.41", uploadfiles.getIp(), "ip");
    assertEquals(true, uploadfiles.hasStarted(), "strat");
  }



  @Test
  public void decodeUSERONLINELIST_REQUEST() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"USERONLINELIST\","
        + "\"DATASET\":{"
        + "\"USERID\":\"louisxiv\""
        + "}}";

    USERONLINELIST useronline = (USERONLINELIST) messagedecoder.decode(encoded);
    assertEquals(Command.USERONLINELIST, useronline.getCommand(), "command");
    assertEquals(REQUEST, useronline.getHeader(), "header");
    assertEquals(Protocol.DATASET, useronline.getDataset(), "dataset");
    assertEquals("louisxiv", useronline.getUserid(), "userid");

  }



  @Test
  public void decodeUSERONLINELIST_REQUEST_OHNE_USERID() throws DecodeException {

    String encoded = "{\"HEADER\":\"REQUEST\"," + "\"COMMAND\":\"USERONLINELIST\"," + "\"DATASET\":{" + "}}";

    USERONLINELIST useronline = (USERONLINELIST) messagedecoder.decode(encoded);
    assertEquals(Command.USERONLINELIST, useronline.getCommand(), "command");
    assertEquals(REQUEST, useronline.getHeader(), "header");
    assertEquals(Protocol.DATASET, useronline.getDataset(), "dataset");
    assertNull(useronline.getUserid(), "userid");

  }



  @Test
  public void decodeUSERONLINELIST_RESPONSE() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"USERONLINELIST\","
        + "\"DATASET\":{"
        + "\"USERID\":\"python\","
        + "\"USERONLINELIST\":["
        + "{\"IP\":\"localhost\","
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\"}]}}";

    USERONLINELIST useronline = (USERONLINELIST) messagedecoder.decode(encoded);

    assertEquals(Command.USERONLINELIST, useronline.getCommand(), "command");
    assertEquals(RESPONSE, useronline.getHeader(), "header");
    assertEquals(Protocol.DATASET, useronline.getDataset(), "dataset");
    UserOnline online[] = useronline.getUserOnline();
    assertEquals("localhost", online[0].getIp(), "ip");
    assertEquals("anaconda", online[0].getNickname(), "nickname");
    assertEquals("louisxiv", online[0].getUserid(), "userid");
    assertEquals("python", useronline.getUserid(), "sender");

  }



  @Test
  public void decodeUSERONLINELIST_RESPONSE_OHNE_USERID() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"USERONLINELIST\","
        + "\"DATASET\":{"
        + "\"USERONLINELIST\":["
        + "{\"IP\":\"localhost\","
        + "\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\"}]}}";

    USERONLINELIST useronline = (USERONLINELIST) messagedecoder.decode(encoded);

    assertEquals(Command.USERONLINELIST, useronline.getCommand(), "command");
    assertEquals(RESPONSE, useronline.getHeader(), "header");
    assertEquals(Protocol.DATASET, useronline.getDataset(), "dataset");
    UserOnline online[] = useronline.getUserOnline();
    assertEquals("localhost", online[0].getIp(), "ip");
    assertEquals("anaconda", online[0].getNickname(), "nickname");
    assertEquals("louisxiv", online[0].getUserid(), "userid");
    assertNull(useronline.getUserid(), "sender Userid");

  }



  @Test
  public void decodeUSERONLINELIST_RESPONSE_ohne_IP() throws DecodeException {

    String encoded = "{\"HEADER\":\"RESPONSE\"," + "\"COMMAND\":\"USERONLINELIST\","
        + "\"DATASET\":{"
        + "\"USERID\":\"python\","
        + "\"USERONLINELIST\":["
        + "{\"USERID\":\"louisxiv\","
        + "\"NICKNAME\":\"anaconda\"}]}}";

    USERONLINELIST useronline = (USERONLINELIST) messagedecoder.decode(encoded);

    assertEquals(Command.USERONLINELIST, useronline.getCommand(), "command");
    assertEquals(RESPONSE, useronline.getHeader(), "header");
    assertEquals(Protocol.DATASET, useronline.getDataset(), "dataset");
    UserOnline online[] = useronline.getUserOnline();
    assertNull(online[0].getIp(), "ip");
    assertEquals("anaconda", online[0].getNickname(), "nickname");
    assertEquals("louisxiv", online[0].getUserid(), "userid");
    assertEquals("python", useronline.getUserid(), "sender");

  }



  @Test
  public void decodeUSRLOGIN() {
    try {
      messagedecoder.decode(null);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void decodeUSRLOGIN_CONFIRM_Gutfall()
      throws DecodeException, JsonSyntaxException, UnsupportedEncodingException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"USRLOGIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"llange@javacomm.net\","
        + "\"USERID\":\"louisxiv\","
        + "\"SESSION\":\"ab\","
        + "\"PASSWORD\":\"PASSWORD\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":-16777216,"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"VOLUME\":100,"
        + "\"NICKNAME\":\"anaconda\"}}";

    USRLOGIN result = (USRLOGIN) messagedecoder.decodeUSRLOGIN(encoded);
    assertEquals("llange@javacomm.net", result.getEmail(), "eMail");
    assertEquals(Command.USRLOGIN, result.getCommand(), "command");
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Protocol.DATASET, result.getDataset(), "dataset");
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("PASSWORD", result.getPassword(), "password");
    assertEquals("anaconda", result.getNickname(), "nickname");
    assertEquals(-16777216, result.getForegroundColor(), "fgcolor");
    assertEquals(-1, result.getBackgroundColor(), "bgcolor");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(100, result.getVolume(), "volume");
    assertEquals("ab", result.getSession(), "SESSION");
  }



  @Test
  public void decodeUSRLOGIN_CONFIRM_Gutfall_ohne_Volume()
      throws DecodeException, JsonSyntaxException, UnsupportedEncodingException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"USRLOGIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"llange@javacomm.net\","
        + "\"USERID\":\"louisxiv\","
        + "\"SESSION\":\"da\","
        + "\"PASSWORD\":\"PASSWORD\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":-16777216,"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"NICKNAME\":\"anaconda\"}}";

    USRLOGIN result = (USRLOGIN) messagedecoder.decodeUSRLOGIN(encoded);
    assertEquals("llange@javacomm.net", result.getEmail(), "eMail");
    assertEquals(Command.USRLOGIN, result.getCommand(), "command");
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Protocol.DATASET, result.getDataset(), "dataset");
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("PASSWORD", result.getPassword(), "password");
    assertEquals("anaconda", result.getNickname(), "nickname");
    assertEquals(-16777216, result.getForegroundColor(), "fgcolor");
    assertEquals(-1, result.getBackgroundColor(), "bgcolor");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(68, result.getVolume(), "volume");
    assertEquals("da", result.getSession(), "SESSION");
  }



  @Test
  public void decodeUSRLOGIN_CONFIRM_Gutfall_ONCALL()
      throws DecodeException, JsonSyntaxException, UnsupportedEncodingException {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"USRLOGIN\","
        + "\"DATASET\":"
        + "{\"EMAIL\":\"llange@javacomm.net\","
        + "\"USERID\":\"louisxiv\","
        + "\"SESSION\":\"12\","
        + "\"PASSWORD\":\"PASSWORD\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":-16777216,"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"VOLUME\":100,"
        + "\"ONCALL\":true,"
        + "\"NICKNAME\":\"anaconda\"}}";

    USRLOGIN result = (USRLOGIN) messagedecoder.decodeUSRLOGIN(encoded);
    assertEquals("llange@javacomm.net", result.getEmail(), "eMail");
    assertEquals(Command.USRLOGIN, result.getCommand(), "command");
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Protocol.DATASET, result.getDataset(), "dataset");
    assertEquals("louisxiv", result.getUserid(), "userid");
    assertEquals("PASSWORD", result.getPassword(), "password");
    assertEquals("anaconda", result.getNickname(), "nickname");
    assertEquals(-16777216, result.getForegroundColor(), "fgcolor");
    assertEquals(-1, result.getBackgroundColor(), "bgcolor");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(true, result.isOncall(), "oncall");
    assertEquals(100, result.getVolume(), "volume");
    assertEquals("12", result.getSession(), "12");
  }



  @Test
  public void decodeUSRLOGIN_CONFIRM_Problem()
      throws DecodeException, JsonSyntaxException, UnsupportedEncodingException {

    String problem = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"USRLOGIN\","
        + "\"DATASET\":{\"EMAIL\":\"llange@javacomm.net\","
        + "\"USERID\":\"louisxiv\","
        + "\"SESSION\":\"ac\","
        + "\"PASSWORD\":\"123\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":-1,"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"VOLUME\":50,"
        + "\"NICKNAME\":\"python\"}}";

    USRLOGIN result = (USRLOGIN) messagedecoder.decodeUSRLOGIN(problem);
    assertEquals(CONFIRM, result.getHeader(), "header");
    assertEquals(Command.USRLOGIN, result.getCommand(), "command");
    assertEquals(Protocol.DATASET, result.getDataset(), "dataset");

    assertEquals("louisxiv", result.getUserid(), "usewrid");
    assertEquals("123", result.getPassword(), "password");
    assertEquals(-1, result.getForegroundColor(), "fg");
    assertEquals(-1, result.getBackgroundColor(), "bg");
    assertEquals("python", result.getNickname(), "nickname");
    assertEquals("localhost", result.getIp(), "ip");
    assertEquals(50, result.getVolume(), "volume");
    assertEquals("ac", result.getSession(), "ac");
  }



  @Test
  public void decodeUSRLOGIN_REQUEST_ohne_User_und_Mail() {

    String encoded = "{\"HEADER\":\"CONFIRM\"," + "\"COMMAND\":\"USRLOGIN\","
        + "\"DATASET\":"
        + "{\"PASSWORD\":\"PASSWORD\","
        + "\"IP\":\"localhost\","
        + "\"FOREGROUND_COLOR\":-16777216,"
        + "\"BACKGROUND_COLOR\":-1,"
        + "\"VOLUME\":100,"
        + "\"ONCALL\":true,"
        + "\"NICKNAME\":\"anaconda\"}}";

    try {
      messagedecoder.decodeUSRLOGIN(encoded);
      fail("DecoderException nicht geworfen, USERID und MAIL fehlen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void kein_COMMAND() {
    String encoded = "{\"HEADER\":\"REQUEST\",\"kein_COMMAND\":\"USRLOGIN\",\"DATASET\":{\"EMAIL\":\"team@javacomm.net\"}}";
    try {
      MESSAGE result = messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void ohne_COMMAND() {
    String encoded = "{\"HEADER\":\"REQUEST\":\"USRLOGIN\",\"DATASET\":{\"EMAIL\":\"team@javacomm.net\"}}";
    try {
      MESSAGE result = messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void ohne_COMMAND_ohne_REQUEST() {
    String encoded = "{\"HEADER\":\"DATASET\":{\"EMAIL\":\"team@javacomm.net\"}}";
    try {
      MESSAGE result = messagedecoder.decode(encoded);
      fail("DecodeException nicht geworfen");
    }
    catch (DecodeException e) {
      assertTrue(true);
    }
  }



  @Test
  public void patternJSON() {
    Matcher matcher1 = Protocol.PATTERN_JSON.matcher("{\"HEADER\"abcd\"COMMAND\" }");
    assertTrue(matcher1.matches());

    Matcher matcher2 = Protocol.PATTERN_JSON.matcher("{\"HEADER\"     \"COMMAND\"       }");
    assertTrue(matcher2.matches());

    Matcher matcher3 = Protocol.PATTERN_JSON.matcher("{");
    assertFalse(matcher3.matches());

    Matcher matcher4 = Protocol.PATTERN_JSON.matcher("}");
    assertFalse(matcher4.matches());

    Matcher matcher5 = Protocol.PATTERN_JSON.matcher("");
    assertFalse(matcher5.matches());

    String bla = "{\"HEADER\":REQUEST,\"COMMAND\":SIGNIN,DATASET:{EMAIL:team@javacomm.net}}";

    Matcher matcher7 = Protocol.PATTERN_JSON.matcher(bla);
    assertTrue(matcher7.matches());

    Matcher matcher6 = Protocol.PATTERN_JSON.matcher(
        "{\"HEADER\":\"REQUEST\",\"COMMAND\":\"SIGNIN\",\"DATASET\":{\"EMAIL\":\"team@javacomm.net\"}}"
    );
    assertTrue(matcher6.matches());

  }



  @BeforeEach
  public void setUp() {
    messagedecoder = new MessageDecoder();
    messagedecoder.init(null);
  }

}
